/* codec_agsm.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * GSM 06.10 lossy speech compression by
 * Copyright 1992 by Jutta Degener and Carsten Bormann,
 * Technische Universitaet Berlin.
 * http://kbs.cs.tu-berlin.de/~jutta/toast.html
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_agsm.c,v 1.11 2003/01/30 00:50:59 nj_humfrey Exp $
 */

#include <string.h>
#include <stdlib.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "libgsm/gsm.h"


/* Some static global variables */
static int Initialised = 0;           /* Initialising counter */

static const char   Signature[5]  = "agsm";
static const char	*Name = "GSM 06.10"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Jutta Degener and Carsten Bormann"; 
static const char   *Description  =
"The GSM 06.10 codec is a lossy speech compressor as defined by the ETSI. "
"It provides telephone quality speech at a rate of 13 kbit/sec. The only "
"supported audio format is single channel audio, with 16 bit samples "
"and 8kHz sample rate.";


#define		BLOCK_SIZE				33
#define		SAMPLES_PER_BLOCK		160

#define 	PARAM_COUNT				2
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "Fast", OQT_PARAMETER_BOOLEAN, "Sacrifice conformance to performance"},
	{ "LTPCut", OQT_PARAMETER_BOOLEAN, "Ignore most samples during LTP"},
};



typedef struct
{
	gsm gsm_data;
	int fast;
	int ltp_cut;
} Param;



static const oqt_codec_param_t*
list_params_agsm(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return (oqt_codec_param_t*)ListParams;
}


static int
set_param_agsm(oqt_t *file, int track,
		  const char* key, const void* value)
{
  Param *p = (Param*)(file->atracks[track].codec_private);

 
  if(!strcasecmp(key, "fast")) {
    p->fast = *(int*)value;
  } else if(!strcasecmp(key, "ltpcut")) {
  	if (*(int*)value)	p->ltp_cut = 100;
  	else				p->ltp_cut = 0;
  } else {
    fprintf(stderr,"set_param_agsm: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }

  return 0;
}

static int
get_param_agsm(oqt_t *file, int track,
		  const char* key, void *value)
{
  Param *p = (Param*)(file->atracks[track].codec_private);

  if(!strcasecmp(key, "fast")) {
    *(char*)value = p->fast;
    return OQT_PARAMETER_BOOLEAN;
  } else if(!strcasecmp(key, "ltpcut")) {
    if (p->ltp_cut < 50)	*(char*)value = 0;
    else					*(char*)value = 1;
    return OQT_PARAMETER_BOOLEAN;
  } else {
    fprintf(stderr,"get_param_agsm: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
  
  return 0;
}

static int
init_agsm(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p;

	atrack->codec_private = (void*)malloc(sizeof(Param));
	if (!atrack->codec_private) {
		fprintf(stderr, "init_agsm: malloc failed when creating codec_private.\n");
		return 0;
	}
	p = (Param*)atrack->codec_private;

	// Create the gsm data structure
	if (! (p->gsm_data = gsm_create())) {
		fprintf(stderr, "init_agsm: malloc failed when creating gsm_data.\n");
		return 0;
	}
	
	// Set options
	p->fast = 0;
	p->ltp_cut = 0;

	return ++Initialised;
}

static int
delete_agsm(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p = (Param*)(atrack->codec_private);

	if (p) {
		if(p->gsm_data) gsm_destroy(p->gsm_data);
		free(p);
		atrack->codec_private = NULL;
	}
	
	return --Initialised;
}

static
int frames_to_bytes_agsm(
	oqt_t *file,
	int track,
	int frames)
{
	int channels = oqt_get_audio_channels(file, track);
	
	return (frames/SAMPLES_PER_BLOCK)*BLOCK_SIZE*channels;
}

static int
decode_agsm(oqt_t *file, int track, 
	       oqt_uint32_t inputsize, 
	       oqt_uint8_t  *input,
	       oqt_uint32_t outputsize, 
	       oqt_uint8_t  *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
  	gsm_byte* in_ptr = (gsm_byte*)input;
  	gsm_signal* out_ptr = (gsm_signal*)output;
  	int i;
  	
  	for(i=0; i<inputsize; i+=BLOCK_SIZE) {
		gsm_decode (p->gsm_data, in_ptr, out_ptr);
		in_ptr += BLOCK_SIZE;
		out_ptr += SAMPLES_PER_BLOCK;
	}	
	
	return i-BLOCK_SIZE;
}


static int
prepare_encode_agsm(oqt_t *file, int track)
{
	oqt_trak_t *atrak = file->atracks[track].track;
	Param *p = (Param*)(file->atracks[track].codec_private);
	
	// Set libgsm options...
	gsm_option (p->gsm_data,  GSM_OPT_FAST, &p->fast) ;
	gsm_option (p->gsm_data,  GSM_OPT_LTP_CUT, &p->ltp_cut) ;

	// Check audio format
	if (oqt_get_audio_bits(file,track)!=16) {
		fprintf(stderr, "prepare_encode_agsm: error, this codec only works for 16bit audio.\n");
		return -1;
	}
	if (oqt_get_audio_channels(file,track)!=1) {
		/*** This might actually be possible in stereo, if two lots of state are used ***/
		fprintf(stderr, "prepare_encode_agsm: error, this codec only works for mono audio.\n");
		return -1;
	}
	if (oqt_get_audio_samplerate(file,track)!=8000) {
		fprintf(stderr, "prepare_encode_agsm: error, this codec only works for 8000Hz audio.\n");
		return -1;
	}
	

	// Create a SoundDescription Version 1
	atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65535;
	atrak->mdia.minf.stbl.stsd.table[0].version = 1;
	atrak->mdia.minf.stbl.stsd.table[0].samplesPerPacket = SAMPLES_PER_BLOCK;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerPacket = BLOCK_SIZE;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerFrame = BLOCK_SIZE;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerSample = 2;
	
	/* (8000/2)/SAMPLES_PER_BLOCK is about 26 blocks per chunk
	   which gives just over half a second per chunk */
	return 26*SAMPLES_PER_BLOCK;
}


static int
encode_agsm(oqt_t *file, int track,
	       oqt_uint32_t num_samples,
	       oqt_uint8_t *input,
	       oqt_uint32_t out_size,
	       oqt_uint8_t *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	
	// check audio sample sizes here as well ?
  	gsm_signal* in_ptr = (gsm_signal*)input;
  	gsm_byte* out_ptr = (gsm_byte*)output;
  	int i;
  	
  	for(i=0; i<num_samples; i+=SAMPLES_PER_BLOCK) {
		gsm_encode(p->gsm_data, in_ptr, out_ptr);
		in_ptr += SAMPLES_PER_BLOCK;
		out_ptr += BLOCK_SIZE;
	}
	
	return (i/SAMPLES_PER_BLOCK)*BLOCK_SIZE;
}

oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_audio_codec_t *codec;

	// Allocate memory for codec record
	codec = oqt_allocate_audio_codec();
	if (!codec) return NULL;
	
	// Store information about the codec
	strncpy(codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.description = Description;
	codec->info.interface_age = OQT_INTERFACE_AGE;

	// Set the callbacks
	codec->init_codec = init_agsm;
	codec->prepare_encoder = prepare_encode_agsm;
	codec->frames_to_bytes = frames_to_bytes_agsm;
	codec->encode = encode_agsm;
	codec->decode = decode_agsm;
	codec->delete_codec = delete_agsm;

	codec->list_params = list_params_agsm;
	codec->set_param = set_param_agsm;
	codec->get_param = get_param_agsm;

	return &(codec->info);
}
