/* codec_dvca.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_dvca.c,v 1.4 2003/05/15 13:28:17 nhumfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libdv/dv.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int Initialised   = 0;      
static const char Signature[5]  = "dvca"; 
static const char *Name = "Quasar DV audio"; 
static const char *Version  = "0.1"; 
static const char *Copyright  = "2003 OpenQuicktime Team"; 
static const char *ModuleAuthor  = "Jozef Hatala"; 
static const char *CodecAuthor  = "Charles Krasic and Erik Walthinsen"; 
  
typedef struct {

  dv_decoder_t *dec_handle;
  oqt_int16_t *audio_buffers[4];

} Param;  

static int decode_dvca(oqt_t *file, int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint32_t outputsize,
		       oqt_uint8_t  *output)
{
  int i, j, c, res;
  Param *p = (Param*)(file->atracks[track].codec_private);
  int outputpos = 0;

  while(inputsize > 0) {

    res = dv_parse_header(p->dec_handle, input);
    if(res < 0) {
      fprintf(stderr, "decode_dvca: dv_parse_header failed\n");
      return 1;
    }

    if(p->dec_handle->frame_size > inputsize) {
      fprintf(stderr, "decode_dvca: unexpected frame size %d, should be at least %d\n",
	      inputsize, (int)p->dec_handle->frame_size);
      return 1;
    }

    dv_parse_packs(p->dec_handle, input);

    dv_decode_full_audio(p->dec_handle, input, p->audio_buffers);

    if(p->dec_handle->audio->samples_this_frame > DV_AUDIO_MAX_SAMPLES) {
      fprintf(stderr, "decode_dvca: this frame contained more than DV_AUDIO_MAX_SAMPLES: %d > %d\n",
        p->dec_handle->audio->samples_this_frame, DV_AUDIO_MAX_SAMPLES);
      fprintf(stderr, "decode_dvca:   expect memory corruption\n");
    }

    if(p->dec_handle->audio->num_channels > 4) {
      fprintf(stderr, "decode_dvca: this frame contained %d channels, although the maximum is 4.\n", p->dec_handle->audio->num_channels);
      fprintf(stderr, "decode_dvca:   expect memory corruption\n");
    }

    j = 0;
    for(i=0; i < p->dec_handle->audio->samples_this_frame; i++) {
      for(c=0; c < p->dec_handle->audio->num_channels; c++) {
	if(j+outputpos >= outputsize) {
	  fprintf(stderr, "decode_dvca: would overflow output buffer\n");
	  return j+outputpos;
	}
	*((oqt_int16_t *)(output+outputpos+j)) = p->audio_buffers[c][i];
	j += sizeof(oqt_int16_t);
      }
    }

    outputpos += j;
    input += p->dec_handle->frame_size;
    inputsize -= p->dec_handle->frame_size;
  }

  return outputpos;
}



static int init_dvca(oqt_t *file, int track)
{
  int i;
  Param *p;
  oqt_audio_map_t *atrack = &file->atracks[track];
		
  atrack->codec_private = malloc(sizeof(Param));
  p = (Param*)atrack->codec_private;	

  p->dec_handle = dv_decoder_new(TRUE, FALSE, FALSE);
  if(!p->dec_handle) {
    fprintf(stderr, "init_dvca: dv_decoder_new failed\n");
    /* XXX how do we indicate failure? */
    return 0;
  }

  for(i=0; i < 4; i++) {
    // this prevents occasional overflows, I don't know why...
    p->audio_buffers[i] = malloc(4*DV_AUDIO_MAX_SAMPLES*sizeof(oqt_int16_t));
  }

  Initialised++;
  return Initialised;
}


static int delete_dvca(oqt_t *file, int track)
{
  int i;
  oqt_audio_map_t *atrack = &file->atracks[track];
  Param *p = (Param*)(atrack->codec_private);

  if(p) {
    dv_decoder_free(p->dec_handle);

    for(i=0; i < 4; i++) {
      free(p->audio_buffers[i]);
    }

    free(p);
    atrack->codec_private=NULL;
  }

  return --Initialised;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
  oqt_audio_codec_t *codec;

  // Allocate the memory for the codec record
  codec = oqt_allocate_audio_codec();
  if (!codec) return NULL;

  // Set the name and signature of the codec
  memcpy(codec->info.fourcc, Signature, 4);
  codec->info.name = Name;

  // Set the callbacks
  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;
	
  codec->delete_codec = delete_dvca;
  codec->decode = decode_dvca;
  codec->init_codec = init_dvca;

  codec->info.merit = 90;

  return &codec->info;
}
