/* codec_MP3.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_MP3.c,v 1.20 2002/12/21 16:38:22 nj_humfrey Exp $
 */


#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#ifdef USELAME
#include <lame/lame.h>
#endif

#include "mpg123/mpg123.h"
#include "mpg123/mpglib.h"



/* Some static global variables */
static int Initialised = 0;           /* Initialising counter */


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
#ifdef USELAME
	lame_global_flags *gf;
	int lame_prepared;		// Lame ready to start encoding ?
#endif
	int num_channels;		// Number of channels in MP3 data
	struct mpstr mp;
} Param;

/* In trak.c */
int oqt_trak_init_audio(oqt_t *file, oqt_trak_t *trak, int channels, int sample_rate, int bits, char *compressor);


static const char   Signature[5]  = ".mp3"; // Standard Codec Signature
static const char	*Name = "MPEG Audio"; 
static const char	*AlternateName = "MPEG Audio (CBR)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Antoine Mine and Nicholas Humfrey"; 
#ifdef USELAME
static const char   *CodecAuthor  = "lame and mpg123 teams"; 
#else
static const char   *CodecAuthor  = "mpg123 team"; 
#endif


/*	Antoine says: OK, we can have multiple instances of the codecs working at the
	same time, but the mpg123lib (at least) has some static buffers that are used
	in layer3.c, so we cannot have safely two different threads calling at the 
	same time different instances of the mp3 encoder/decoder.
	For the moment, I've simply added a mutex allowing only one thread to call
	encodeMP3/decodeMP3 at the same time...
*/
static pthread_mutex_t decodelock;
static pthread_mutex_t encodelock;


#define PARAM_COUNT 12
 
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "VBR", OQT_PARAMETER_BOOLEAN, "Enable/Disable VBR"},
	{ "Bitrate", OQT_PARAMETER_INTEGER, "Bitrate of stream"},
	{ "Quality", OQT_PARAMETER_INTEGER, "Adjusts quality (and speed) [0=best, 9=worst]"},
	{ "Mode", OQT_PARAMETER_STRING, "[Stereo,JStereo,Dual,Mono]"},
	
	{ "Padding", OQT_PARAMETER_BOOLEAN, "Enable/Disable Padding"},
	{ "Copyright", OQT_PARAMETER_BOOLEAN, "Enable/Disable Copyright Bit"},
	{ "Original", OQT_PARAMETER_BOOLEAN, "Enable/Disable Original Bit"},
	{ "Protection", OQT_PARAMETER_BOOLEAN, "Enable/Disable Error Protection"},
	
	{ "LowPass_Freq", OQT_PARAMETER_INTEGER, "Low Pass Frequency (Hz) [0=auto, -1=disable]"},
	{ "HighPass_Freq", OQT_PARAMETER_INTEGER, "High Pass Frequency (Hz) [0=auto, -1=disable]"},
	{ "LowPass_Width", OQT_PARAMETER_INTEGER, "Width Low Pass Frequency Filter (Hz)"},
	{ "HighPass_Width", OQT_PARAMETER_INTEGER, "Width High Pass Frequency Filter (Hz)"}
	
};



#ifdef USELAME

static const oqt_codec_param_t*
list_params_MP3(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return (oqt_codec_param_t*)ListParams;
}

static int
set_param_MP3(oqt_t *file, int track,
		  const char* key, const void* value)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	
	if(!strcasecmp(key, "bitrate"))
		lame_set_brate( p->gf, *(int*)value);
	else if(!strcasecmp(key, "vbr"))
		if (*(int*)value) {
			lame_set_VBR( p->gf, vbr_default);
		} else {
			lame_set_VBR( p->gf, vbr_off);
		}
	else if(!strcasecmp(key, "quality"))
		lame_set_quality( p->gf, *(int*)value);
	else if(!strcasecmp(key, "mode")) {
		char* str = (char*)value;
		if (!strcasecmp(str, "Stereo")) {
			lame_set_mode(p->gf, STEREO);
			p->num_channels = 2;
		} else if (!strcasecmp(str, "JStereo")) {
			lame_set_mode(p->gf, JOINT_STEREO);
			p->num_channels = 2;
		} else if (!strcasecmp(str, "Dual")) {
			lame_set_mode(p->gf, DUAL_CHANNEL);
			p->num_channels = 2;
		} else if (!strcasecmp(str, "Mono")) {
			lame_set_mode(p->gf, MONO);
			p->num_channels = 1;
		} else return OQT_PARAMETER_UNKNOWN;
	}
	else if(!strcasecmp(key, "padding"))
		if (*(int*)value)
				lame_set_padding_type( p->gf, PAD_ALL);
		else	lame_set_padding_type( p->gf, PAD_NO);
	else if(!strcasecmp(key, "copyright"))
		lame_set_copyright( p->gf, *(int*)value);
	else if(!strcasecmp(key, "original"))
		lame_set_original( p->gf, *(int*)value);
	else if(!strcasecmp(key, "protection"))
		lame_set_error_protection( p->gf, *(int*)value);
	else if(!strcasecmp(key, "lowpass_freq"))
		lame_set_lowpassfreq( p->gf, *(int*)value);
	else if(!strcasecmp(key, "highpass_freq"))
		lame_set_highpassfreq( p->gf, *(int*)value);
	else if(!strcasecmp(key, "lowpass_width"))
		lame_set_lowpasswidth( p->gf, *(int*)value);
	else if(!strcasecmp(key, "highpass_width"))
		lame_set_highpasswidth( p->gf, *(int*)value);
	else {
		fprintf(stderr,"set_param_MP3: unknown parameter named '%s'\n",key);
		return OQT_PARAMETER_UNKNOWN;
	}
	
	return 0;
}

static int
get_param_MP3(oqt_t *file, int track,
		  const char* key, void *value)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	
	if (!p) fprintf(stderr, "get_param_MP3: Argh ! private memory is null.\n");
	
	if(!strcasecmp(key, "bitrate")) {
		*(int*)value = lame_get_brate( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else if(!strcasecmp(key, "vbr")) {
		*(int*)value = (lame_get_VBR( p->gf )==vbr_off ? 0:1);
		return OQT_PARAMETER_BOOLEAN;
	} else if(!strcasecmp(key, "quality")) {
		*(int*)value = lame_get_quality( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else if(!strcasecmp(key, "mode")) {
		switch(lame_get_mode( p->gf ) ) {
			case STEREO:  		strcpy((char*)value, "Stereo"); break;
			case JOINT_STEREO:  strcpy((char*)value, "JStereo"); break;
			case DUAL_CHANNEL:  strcpy((char*)value, "Dual"); break;
			case MONO:  		strcpy((char*)value, "Mono"); break;
			default: 			strcpy((char*)value, "Error"); break;
		}
		return OQT_PARAMETER_STRING;
	} else if(!strcasecmp(key, "padding")) {
		*(int*)value = lame_get_padding_type( p->gf )==PAD_NO ? 0:1;
		return OQT_PARAMETER_BOOLEAN;
	} else if(!strcasecmp(key, "copyright")) {
		*(int*)value = lame_get_copyright( p->gf );
		return OQT_PARAMETER_BOOLEAN;
	} else if(!strcasecmp(key, "original")) {
		*(int*)value = lame_get_original( p->gf );
		return OQT_PARAMETER_BOOLEAN;
	} else if(!strcasecmp(key, "protection")) {
		*(int*)value = lame_get_error_protection( p->gf );
		return OQT_PARAMETER_BOOLEAN;
	} else if(!strcasecmp(key, "lowpass_freq")) {
		*(int*)value = lame_get_lowpassfreq( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else if(!strcasecmp(key, "highpass_freq")) {
		*(int*)value = lame_get_highpassfreq( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else if(!strcasecmp(key, "lowpass_width")) {
		*(int*)value = lame_get_lowpasswidth( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else if(!strcasecmp(key, "highpass_width")) {
		*(int*)value = lame_get_highpasswidth( p->gf );
		return OQT_PARAMETER_INTEGER;
	} else {
		fprintf(stderr,"get_param_MP3: unknown parameter named '%s'\n",key);
		return OQT_PARAMETER_UNKNOWN;
	}
	
	return 0;
}



#define MPEGAUDIO_BITS_SAMPLE			16

#define MPEGLAYER3_ID_UNKNOWN            0
#define MPEGLAYER3_ID_MPEG               1
#define MPEGLAYER3_ID_CONSTANTFRAMESIZE  2

#define MPEGLAYER3_FLAG_PADDING_ISO      0x00000000
#define MPEGLAYER3_FLAG_PADDING_ON       0x00000001
#define MPEGLAYER3_FLAG_PADDING_OFF      0x00000002



static int
prepare_encode_MP3(oqt_t *file, int track)
{
	oqt_trak_t *atrak = file->atracks[track].track;
	Param *p = (Param*)(file->atracks[track].codec_private);
	int bytes_per_packet, samples_per_chunk, blocks_per_chunk;
	if (p->lame_prepared) fprintf(stderr, "prepare_encode_MP3: Warning Re-preparing Lame.\n");
	
	
	// Setup input parameters
	lame_set_in_samplerate( p->gf, oqt_get_audio_samplerate(file, track));
	lame_set_num_channels( p->gf, oqt_get_audio_channels(file, track));
	
	
	// set more internal configuration based on data provided
	if (lame_init_params( p->gf )==-1) {
		fprintf(stderr, "prepare_encode_MP3: Failed to prepare Lame.\n");
		return -1;
	} else p->lame_prepared=1;
	
	// Re-configure the audio track to the format that Lame is outputting
	oqt_trak_init_audio(file, atrak, p->num_channels, lame_get_out_samplerate( p->gf),
						MPEGAUDIO_BITS_SAMPLE, file->atracks[track].codec->info.fourcc);
	//file->atracks[track].channels = p->num_channels;

	
	// Create a SoundDescription Version 1
	bytes_per_packet = (lame_get_brate(p->gf)*lame_get_framesize(p->gf)) / 
	                 (((float)lame_get_out_samplerate(p->gf)/1000.0)*8);
	//if (lame_get_padding_type(p->gf)!=PAD_NO)	++bytes_per_packet;
	if (lame_get_VBR(p->gf) != vbr_off)
			atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65534;
	else	atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65535;
	atrak->mdia.minf.stbl.stsd.table[0].version = 1;
	atrak->mdia.minf.stbl.stsd.table[0].samplesPerPacket = lame_get_framesize(p->gf)*p->num_channels;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerPacket = bytes_per_packet;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerFrame = bytes_per_packet*p->num_channels;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerSample = (MPEGAUDIO_BITS_SAMPLE/8);
	
	// Copy accross the signature for the codec
	memcpy(	atrak->mdia.minf.stbl.stsd.table[0].wave.fmtType, 
			file->atracks[track].codec->info.fourcc, 4);
	atrak->mdia.minf.stbl.stsd.table[0].wave.little_endian = 0;	// MP3s arn't little endian (!)

	// The Wave 'fmt ' header
	atrak->mdia.minf.stbl.stsd.table[0].wave.wFormatTag = OQT_STSDWAVE_MPEG;
	atrak->mdia.minf.stbl.stsd.table[0].wave.nChannels = p->num_channels;
	atrak->mdia.minf.stbl.stsd.table[0].wave.nSamplesPerSec = lame_get_out_samplerate( p->gf );
	atrak->mdia.minf.stbl.stsd.table[0].wave.nAvgBytesPerSec = (lame_get_brate(p->gf)*1000)/8;
	//atrak->mdia.minf.stbl.stsd.table[0].wave.nBlockAlign = (int)(144*lame_get_brate(p->gf)/lame_get_out_samplerate( p->gf ));	// Maybe just set to 1 ?
	atrak->mdia.minf.stbl.stsd.table[0].wave.nBlockAlign = 1;
	atrak->mdia.minf.stbl.stsd.table[0].wave.wBitsPerSample = 0;		// Always zero (unused)
	atrak->mdia.minf.stbl.stsd.table[0].wave.cbSize = 12;	// The length of the MPEG Audio extension to the fmt header

	// MPEG Audio extension
	atrak->mdia.minf.stbl.stsd.table[0].wave.wID = MPEGLAYER3_ID_MPEG;
	atrak->mdia.minf.stbl.stsd.table[0].wave.nBlockSize = bytes_per_packet;
	atrak->mdia.minf.stbl.stsd.table[0].wave.nFramesPerBlock = 1;	// Apple have this set to one
	atrak->mdia.minf.stbl.stsd.table[0].wave.nCodecDelay = 0;		// Apple have this set to zero
	

	/* 0=no padding, 1=always pad, 2=adjust padding */
	atrak->mdia.minf.stbl.stsd.table[0].wave.fdwFlags = 32;
	// VBR
	// MPEG 1/2
	// Padding
	// Protection
	// .... ?
	
	/*if (lame_get_padding_type(p->gf)==PAD_ALL)
			atrak->mdia.minf.stbl.stsd.table[0].wave.fdwFlags |= MPEGLAYER3_FLAG_PADDING_ON;
	else if (p->gf.padding_type==PAD_ADJUST)
		 atrak->mdia.minf.stbl.stsd.table[0].wave.fdwFlags |= MPEGLAYER3_FLAG_PADDING_ISO;
	else atrak->mdia.minf.stbl.stsd.table[0].wave.fdwFlags |= MPEGLAYER3_FLAG_PADDING_OFF;*/

	/* Update the sample rate and channels */
	 

	/*#if DEBUG
	fprintf(stderr, "---------------------\n");
	fprintf(stderr, " Bitrate: %d\n", lame_get_brate(p->gf));
	fprintf(stderr, " VBR: %d\n", lame_get_VBR(p-gf));
	fprintf(stderr, " Quality: %d\n", lame_get_quality(p->gf));
	
	fprintf(stderr, " Mode: ");
	switch(p->gf.mode) {
  		case STEREO:  fprintf(stderr, "Stereo\n"); break;
  		case JOINT_STEREO:  fprintf(stderr, "Joint Stereo\n"); break;
  		case DUAL_CHANNEL:  fprintf(stderr, "Dual\n"); break;
  		case MONO:  fprintf(stderr, "Mono\n"); break;
  		default: fprintf(stderr, "Error\n"); break;
  	}
  	
	fprintf(stderr, " Padding: %d\n", lame_get_padding_type(p->gf));
	fprintf(stderr, " Copyright: %d\n", lame_get_copyright(p->gf));
	fprintf(stderr, " Original: %d\n", lame_get_original(p->gf));
	fprintf(stderr, " Protection: %d\n", lame_get_error_protection(p->gf));
	fprintf(stderr, " Low Pass Freq: %d\n", lame_get_lowpassfreq(p->gf));
	fprintf(stderr, " High Pass Freq: %d\n", lame_set_highpassfreq(p->gf));
	fprintf(stderr, " Low Pass Width: %d\n", lame_get_lowpasswidth(p->gf));
	fprintf(stderr, " High Pass Width: %d\n", lame_get_highpasswidth(p->gf));
	fprintf(stderr, "---------------------\n");
	#endif*/
	
	// Display Lame's settings
	#if DEBUG
		lame_print_config( p->gf );
		lame_print_internals( p->gf );
	#endif


	/* How many blocks per chunk ?  - chunks should be approx 0.5s */
	blocks_per_chunk = ceil((double)(lame_get_out_samplerate( p->gf ) / 2)
	                   / (lame_get_framesize(p->gf)*p->num_channels));
	samples_per_chunk = blocks_per_chunk * lame_get_framesize(p->gf)*p->num_channels;
	
	
	return samples_per_chunk;
}


static int
encode_MP3(oqt_t *file, int track,
	       oqt_uint32_t num_samples,
	       oqt_uint8_t *input,
	       oqt_uint32_t out_size,
	       oqt_uint8_t *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	
	int result=0, size_written=0;
	int sample_size = oqt_get_audio_bits(file, track);
	
	if (!p->lame_prepared) {
		fprintf(stderr, "Error: lame has not been prepared for encoding yet.\n");
		return 0;
	}
	
	pthread_mutex_lock(&encodelock);

	if (sample_size == 16) {
		if (SIZEOF_SHORT != 2)	fprintf(stderr, "encode_MP3: warning, size of short is not 16 bits.\n");

		if(p->num_channels==1) {
			const short int* input_ptr = (const short int*)input;
			result = lame_encode_buffer(p->gf, input_ptr, input_ptr,
				num_samples, output, out_size);
		} else {
			short int* input_ptr = (short int*)input;
			result = lame_encode_buffer_interleaved(p->gf, input_ptr,
				num_samples, output, out_size);
		}
	} else if (sample_size == 32) {
		const long int* input_ptr = (const long int*)input;
		if (SIZEOF_LONG != 4)	fprintf(stderr, "encode_MP3: warning, size of long is not 32 bits.\n");
	
		if(p->num_channels==1) {
			result = lame_encode_buffer_long2(p->gf, input_ptr, input_ptr,
				num_samples, output, out_size);
		} else {
			fprintf(stderr, "encode_MP3: sorry don't currently support 32 bit stereo enocoding.\n");
			return 0;
		}
	} else {
		fprintf(stderr, "encode_MP3: Error, unsupported sample size (%d).\n", sample_size);
		return 0;
	}
		
	
	/* Check if there was an error */
	if (result<0) {
		fprintf(stderr, "encode_MP3: error lame_encode returned %d.\n", result);
		size_written=0;
	} else {
		int extra;

		/* Make sure all the data was returned */
		extra=lame_encode_flush( p->gf, output+size_written, out_size-size_written );
		fprintf(stderr, "lame_encode_flush( p->gf, 0x%x, %d ) = %d\n", (int)output+size_written, (int)out_size-size_written, extra);
		
		size_written = result + size_written;
	}
 
 	
	fprintf(stderr,"num_samples %u, size_written %u, out_size %u\n",
	                   num_samples, size_written, out_size);
	

	pthread_mutex_unlock(&encodelock);
	
	return size_written;
}

#endif	// USELAME


static 
int decode_MP3(oqt_t *file, int track, 
	       oqt_uint32_t inputsize, 
	       oqt_uint8_t  *input,
	       oqt_uint32_t outputsize, 
	       oqt_uint8_t  *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	static int first_time = 1;
	int size, offset=0;
	int last_result = MP3_NEED_MORE;
	
	pthread_mutex_lock(&decodelock);
	
	//printf("inputsize %lu outputsize %lu\n", inputsize, outputsize);
	last_result = decodeMP3(&(p->mp),(char*)input, inputsize,(char*)output,outputsize,&size);
	//printf("  offset=%i size=%i last_result=%i\n",offset,size,last_result);
	offset+=size;
	
	while(last_result== MP3_OK && offset<outputsize)
	{// OK, here are some
		last_result = decodeMP3(&(p->mp),NULL,0,(char*)output+offset,outputsize-offset,&size);
		//printf("  offset=%i size=%i last_result=%i\n",offset,size,last_result);
		offset+=size;
	}
	
	// If this is the first time round then transfer
	// information about the format of the stream
	if (first_time) {
		// **** Needs done ****
		//p->gf.brate = 
	
		first_time=0;
	}	
	
	pthread_mutex_unlock(&decodelock);
	
	return offset;
}


static int
init_MP3(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p;
	++Initialised;
	
	atrack->codec_private = malloc(sizeof(Param));
	if (!atrack->codec_private) {
		fprintf(stderr, "Error: malloc failed when creating codec_private in init_MP3\n");
		return 0;
	}
	
	// Initalise default parameter values
	p = (Param*)atrack->codec_private;
	
	// Init mpg123 */
	InitMP3(&(p->mp));
	
	
	// Init LAME */
	#ifdef USELAME
	{
		p->gf = lame_init();
		if (p->gf == NULL) {
			fprintf(stderr, "init_MP3: failed to initalise lame.\n");
		}

		/* Setup defaults for LAME */
		//lame_set_VBR_q( p->gf, atrack->track->mdia.minf.stbl.stsd.table[0].spatial_quality);
		//lame_set_brate( p->gf, (128000 + (p->gf.VBR_q * 32000)) / 125;
		
		lame_set_brate( p->gf, 128);
		lame_set_VBR( p->gf, vbr_off);	// VBR off by default
		
		p->num_channels = oqt_get_audio_channels(file, track);
		if(p->num_channels==1)		lame_set_mode( p->gf, MONO );			// Mono Mode
		else if(p->num_channels==2)	lame_set_mode( p->gf, JOINT_STEREO );	// Joint Stereo
		else fprintf(stderr, "init_MP3: MP3 codec can't enode more than two channels.\n");
		
		// Call lame_init_params again before encoding...
		lame_init_params( p->gf );
		p->lame_prepared=0;
	}
	#endif
	
	pthread_mutex_init(&decodelock,NULL);
	
	return Initialised;
}

static int
delete_MP3(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p = (Param*)(atrack->codec_private);
	
	if (p) {
		#ifdef USELAME
		if (p->gf) {
			// Close down lame
			lame_close( p->gf );
			p->lame_prepared = 0;
			p->gf = NULL;
		}
		#endif
		
		// Dispose of out private memory
		free(p);
		atrack->codec_private=NULL;
	}
	
	return --Initialised;
}



static void
set_standard_values(oqt_audio_codec_t *codec ) {

	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	//**** Do something with lame_version() (and MPG123) *** //
  
	codec->init_codec = init_MP3;
	codec->delete_codec = delete_MP3;
	codec->decode = decode_MP3;
	
#ifdef USELAME
	codec->prepare_encoder = prepare_encode_MP3;
	codec->encode = encode_MP3;
	codec->list_params = list_params_MP3;
	codec->set_param = set_param_MP3;
	codec->get_param = get_param_MP3;
#else
	// Decrease the merit if no encoding support
	codec->info.merit = 80;
#endif
}



oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_audio_codec_t *codec,*alt;

	// Allocate the memory for the codec record
	codec = oqt_allocate_audio_codec();
	alt = oqt_allocate_audio_codec();
	if (!codec || !alt) return NULL;

	// Set the name and signature of the codec
	strncpy(codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	
	// Set the alternate name and signature
	alt->info.fourcc[0]='m';
	alt->info.fourcc[1]='s';
	alt->info.fourcc[2]=0;
	alt->info.fourcc[3]=0x55;
	alt->info.name = AlternateName;
	
	alt->info.parent = &codec->info;
	codec->info.has_children = 1;
	
	// Set the callbacks
	set_standard_values(codec);
	set_standard_values(alt);
	
	
	// We only have to register the alternate codec
	if (oqt_register_codec(&alt->info)) {
		fprintf(stderr, "Failed to register alternate MP3 codec.\n");
	}
	

	return &codec->info;
}


