/* codec_ima4.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_ima4.c,v 1.9 2002/12/15 01:51:56 nj_humfrey Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"

/* Some static global variables */
static int Initialised = 0;           /* Initialising counter */

static const char   Signature[5]  = "ima4";
static const char	*Name = "IMA 4:1"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Adam Williams"; 


typedef struct
{
	// Used while interlacing/deinterlacting buffers
	oqt_int16_t *work_buffer;
	
	// Starting information for all channels during encoding. 
	int *last_samples, *last_indexes;
} Param;


static int ima4_step[89] = 
{
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

static int ima4_index[16] = 
{
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8
};

// Known by divine revelation 
#define BLOCK_SIZE 0x22
#define SAMPLES_PER_BLOCK 0x40



static void ima4_decode_sample(int *predictor, int *nibble, int *index, int *step)
{
	int difference, sign;

// Get new index value 
	*index += ima4_index[*nibble];

	if(*index < 0) *index = 0; 
	else 
	if(*index > 88) *index = 88;

// Get sign and magnitude from *nibble 
	sign = *nibble & 8;
	*nibble = *nibble & 7;

// Get difference 
	difference = *step >> 3;
	if(*nibble & 4) difference += *step;
	if(*nibble & 2) difference += *step >> 1;
	if(*nibble & 1) difference += *step >> 2;

// Predict value 
	if(sign) 
	*predictor -= difference;
	else 
	*predictor += difference;

	if(*predictor > 32767) *predictor = 32767;
	else
	if(*predictor < -32768) *predictor = -32768;

// Update the step value 
	*step = ima4_step[*index];
}

static void ima4_decode_block(oqt_int16_t *output, oqt_uint8_t *input)
{
	int predictor, index, step;
	int nibble, nibble_count;
	oqt_uint8_t *input_end = input + BLOCK_SIZE;

// Get the chunk header 
	predictor = *input++ << 8;
	predictor |= *input++;

	index = predictor & 0x7f;
	if(index > 88) index = 88;

	predictor &= 0xff80;
	if(predictor & 0x8000) predictor -= 0x10000;
	step = ima4_step[index];

// Read the input buffer sequentially, one nibble at a time 
	nibble_count = 0;
	while(input < input_end)
	{
		nibble = nibble_count ? (*input++  >> 4) & 0x0f : *input & 0x0f;

		ima4_decode_sample(&predictor, &nibble, &index, &step);
		*output++ = predictor;

		nibble_count ^= 1;
	}
}


static void ima4_encode_sample(int *last_sample, int *last_index, int *nibble, int next_sample)
{
	int difference, new_difference, mask, step;

	difference = next_sample - *last_sample;
	*nibble = 0;
	step = ima4_step[*last_index];
	new_difference = step >> 3;

	if(difference < 0)
	{
		*nibble = 8;
		difference = -difference;
	}

	mask = 4;
	while(mask)
	{
		if(difference >= step)
		{
			*nibble |= mask;
			difference -= step;
			new_difference += step;
		}

		step >>= 1;
		mask >>= 1;
	}

	if(*nibble & 8)
		*last_sample -= new_difference;
	else
		*last_sample += new_difference;

	if(*last_sample > 32767) *last_sample = 32767;
	else
	if(*last_sample < -32767) *last_sample = -32767;

	*last_index += ima4_index[*nibble];

	if(*last_index < 0) *last_index = 0;
	else
	if(*last_index > 88) *last_index= 88;
}

static void ima4_encode_block(Param *p, unsigned char *output, oqt_int16_t *input, int step, int channel)
{
	int i, nibble_count = 0, nibble, header;

// Get a fake starting sample 
	header = p->last_samples[channel];
// Force rounding. 
	if(header < 0x7fc0) header += 0x40;
	if(header < 0) header += 0x10000;
	header &= 0xff80;
	*output++ = (header & 0xff00) >> 8;
	*output++ = (header & 0x80) + (p->last_indexes[channel] & 0x7f);

	for(i = 0; i < SAMPLES_PER_BLOCK; i++)
	{
		ima4_encode_sample(&(p->last_samples[channel]), 
							&(p->last_indexes[channel]), 
							&nibble, 
							*input);

		if(nibble_count)
			*output++ |= (nibble << 4);
		else
			*output = nibble;

		input += step;
		nibble_count ^= 1;
	}
}


static int
init_ima4(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p;
	int i, channels = oqt_get_audio_channels(file, track);

	atrack->codec_private = malloc(sizeof(Param));
	if (!atrack->codec_private) {
		fprintf(stderr, "init_IMA4: malloc failed when creating codec_private.\n");
		return 0;
	}
	p = (Param*)atrack->codec_private;

	// Allocate a work buffer
	p->work_buffer = malloc(SAMPLES_PER_BLOCK * sizeof(oqt_int16_t));
	if (!p->work_buffer) {
		fprintf(stderr, "init_IMA4: malloc failed when creating work_buffer.\n");
		return 0;
	}

	// Allocate last sample buffer
	p->last_samples = malloc(sizeof(int) * channels);
	if (!p->last_samples) {
		fprintf(stderr, "init_IMA4: malloc failed when creating last_samples.\n");
		return 0;
	}
	for(i = 0; i < channels; i++) p->last_samples[i] = 0;

	// Allocate last index buffer
	p->last_indexes = malloc(sizeof(int) * channels);
	if (!p->last_indexes) {
		fprintf(stderr, "init_IMA4: malloc failed when creating last_indexes.\n");
		return 0;
	}
	for(i = 0; i < channels; i++) p->last_indexes[i] = 0;


	return ++Initialised;
}


static int
delete_ima4(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p = (Param*)(atrack->codec_private);

	if (p) {
		if(p->work_buffer) free(p->work_buffer);
		if(p->last_samples) free(p->last_samples);
		if(p->last_indexes) free(p->last_indexes);
		free(p);
		atrack->codec_private = NULL;
	}
	
	return --Initialised;
}


static
int frames_to_bytes_ima4(
	oqt_t *file,
	int track,
	int frames)
{
	int channels = oqt_get_audio_channels(file, track);
	
	return (frames/SAMPLES_PER_BLOCK)*BLOCK_SIZE*channels;
}


static int
decode_ima4(oqt_t *file, int track, 
	       oqt_uint32_t inputsize, 
	       oqt_uint8_t  *input,
	       oqt_uint32_t outputsize, 
	       oqt_uint8_t  *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	int channels = oqt_get_audio_channels(file,track);
  	int samples =  (inputsize * SAMPLES_PER_BLOCK) / BLOCK_SIZE;
	oqt_int16_t *output_ptr = (oqt_int16_t*)output;
	unsigned char* block_ptr = input;
	int i,j,k;

	for(i = 0; i < samples; i += (SAMPLES_PER_BLOCK*channels))
	{
		for(j=0; j < channels; j++) {
			ima4_decode_block(p->work_buffer, block_ptr);
			block_ptr += BLOCK_SIZE;
			
			for(k=0; k<SAMPLES_PER_BLOCK; k++) {
				output_ptr[i+(k*channels)+j] = p->work_buffer[k];
			}
		}
	}
	
	// Sanity check
	/*{
		int z = i*(oqt_get_audio_bits(file,track)/8);
		if (i != samples) fprintf(stderr, "decode_IMA4: Decoded %d of %d samples.\n", i, samples);
		if (z != outputsize) fprintf(stderr, "decode_IMA4: Decoded %d of %d bytes.\n", z, outputsize);
	}*/
	
	return i*(oqt_get_audio_bits(file,track)/8);
}

static int
prepare_encode_ima4(oqt_t *file, int track)
{
	oqt_trak_t *atrak = file->atracks[track].track;
	Param *p = (Param*)(file->atracks[track].codec_private);
	int channels = oqt_get_audio_channels(file, track);

	// Create a SoundDescription Version 1
	atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65535;
	atrak->mdia.minf.stbl.stsd.table[0].version = 1;
	atrak->mdia.minf.stbl.stsd.table[0].samplesPerPacket = SAMPLES_PER_BLOCK;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerPacket = BLOCK_SIZE;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerFrame = BLOCK_SIZE*channels;
	atrak->mdia.minf.stbl.stsd.table[0].bytesPerSample = 2;
	
	return 0;
}


static int
encode_ima4(oqt_t *file, int track,
	       oqt_uint32_t num_samples,
	       oqt_uint8_t *input,
	       oqt_uint32_t out_size,
	       oqt_uint8_t *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	int channels = oqt_get_audio_channels(file, track);
	oqt_int16_t *input_ptr = (oqt_int16_t*)input;
	unsigned char *output_ptr = output;
	int i,j,result=0;

	for(i = 0; 
		i + SAMPLES_PER_BLOCK <= num_samples; 
		i += SAMPLES_PER_BLOCK)
	{
		for(j = 0; j < channels; j++)
		{
			ima4_encode_block(p, output_ptr, input_ptr + j, channels, j);

			output_ptr += BLOCK_SIZE;
		}
		input_ptr += SAMPLES_PER_BLOCK * channels;
	}
	
	// How much data did we put in the output buffer ?
	result = (i/SAMPLES_PER_BLOCK)*channels*BLOCK_SIZE;
	
	// Sanity check
	/*{
		if (i != num_samples) fprintf(stderr, "encode_IMA4: Encoded %d of %d samples.\n", i, num_samples);
		if (result != out_size) fprintf(stderr, "encode_IMA4: Encoded %d of %d bytes.\n", result, out_size);
	}*/

	return result;
}

oqt_codec_info_t*
oqt_plugin_register_ima4( void )
{
	oqt_audio_codec_t *codec;

	// Allocate memory for codec record
	codec = oqt_allocate_audio_codec();
	if (!codec) return NULL;
	
	// Store information about the codec
	strncpy(codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;

	// Set the callbacks
	codec->init_codec = init_ima4;
	codec->prepare_encoder = prepare_encode_ima4;
	codec->frames_to_bytes = frames_to_bytes_ima4;
	codec->encode = encode_ima4;
	codec->decode = decode_ima4;
	codec->delete_codec = delete_ima4;

	return &(codec->info);
}
