/* codec_sowt.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_sowt.c,v 1.10 2002/12/15 01:51:56 nj_humfrey Exp $
 */

#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"


// Well this must be the simplest codec in OpenQuicktime
// It does absolutely nothing :-) - well apart from copy
// memory about a bit...

static const char   Signature[5]  = "sowt";
static const char	*Name = "Twos Complement - Little Endian PCM"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = ""; 
static int Initialised = 0;           /* Initialising counter */


#ifdef WORDS_BIGENDIAN
static int swap_bytes(char *buffer_in, char *buffer_out, long samples, int bits)
{
	long i = 0;

	switch(bits)
	{
		case 8:
			for(i=0;i < samples; i+=1)
			{
				buffer_out[i] = buffer_in[1];
			}
			break;

		case 16:
			for(i=0; i < samples * 2; i+=2)
			{
				buffer_out[i+1] = buffer_in[i];
				buffer_out[i] = buffer_in[i+1];
			}
			break;

		case 24:
			for(i=0; i < samples * 3; i+=3)
			{
				buffer_out[i+2] = buffer_in[i];
				buffer_out[i+1] = buffer_in[i+1];
				buffer_out[i] = buffer_in[i+2];
			}
			break;

		default:
			break;
	}
	return 0;
}
#endif

static int
init_sowt(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	return ++Initialised;
}

static int
delete_sowt(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	return --Initialised;
}

static
int frames_to_bytes_sowt(
	oqt_t *file,
	int track,
	int frames)
{
	int bits = oqt_get_audio_bits(file, track);
	int channels = oqt_get_audio_channels(file, track);
	
	return frames*(bits/8)*channels;
}


static 
int decode_sowt(oqt_t *file, int track, 
	       oqt_uint32_t inputsize, 
	       oqt_uint8_t  *input,
	       oqt_uint32_t outputsize, 
	       oqt_uint8_t  *output)
{
	int bits = oqt_get_audio_bits(file, track);

	// Shouldn't inputsize == outputsize anyway ?
	if (inputsize > outputsize) {
		fprintf(stderr, "decode_sowt: output buffer isn't big enough (%d>%d)\n",
						inputsize, outputsize);
		return 0;
	}
	
	#ifdef WORDS_BIGENDIAN
		// Swap the bytes and copy into output buffer
		swap_bytes(input, output, inputsize/(bits/8), bits);
	#else
		// OK just copy input to output
		memcpy(output, input, inputsize);
	#endif
	
	return inputsize;
}


static int
encode_sowt(oqt_t *file, int track,
	       oqt_uint32_t num_samples,
	       oqt_uint8_t *input,
	       oqt_uint32_t outputsize,
	       oqt_uint8_t *output)
{
	int bits = oqt_get_audio_bits(file, track);
	int channels = oqt_get_audio_channels(file, track);
	int total_samples = num_samples*channels;

	// Calculate the number of bytes in the input buffer
	oqt_uint32_t inputsize = (oqt_get_audio_bits(file, track)/8) * total_samples;
	
	// Shouldn't inputsize == outputsize anyway ?
	if (inputsize > outputsize) {
		fprintf(stderr, "encode_sowt: output buffer isn't big enough (%d>%d)\n",
						inputsize, outputsize);
		return 0;
	}
	
	#ifdef WORDS_BIGENDIAN
		// Swap the bytes and copy into output buffer
		swap_bytes(input, output, total_samples, bits);
	#else
		// OK just copy input to output
		memcpy(output, input, inputsize);
	#endif
	
	return inputsize;
}




oqt_codec_info_t*
oqt_plugin_register_sowt( void )
{
	oqt_audio_codec_t *codec;

	// Allocate memory for codec record
	codec = oqt_allocate_audio_codec();
	if (!codec) return NULL;
	
	// Store information about the codec
	strncpy(codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;

	// Set the callbacks
	codec->init_codec = init_sowt;
	//codec->prepare_encoder = prepare_encode_sowt;
	codec->frames_to_bytes = frames_to_bytes_sowt;
	codec->encode = encode_sowt;
	codec->decode = decode_sowt;
	codec->delete_codec = delete_sowt;

	return &(codec->info);
}
