/* codec_ulaw.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_ulaw.c,v 1.11 2003/01/23 14:31:25 nj_humfrey Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"

static int Initialised = 0;           /* Initialising counter */

static const char   Signature[5]  = "ulaw"; // Standard Codec Signature
static const char	*Name = "(mu)-Law 2:1"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Adam Williams"; 


// ==================================== private for ulaw 

#define uBIAS 0x84
#define uCLIP 32635

typedef struct
{
	oqt_int16_t *ulawtoint16_table;
	oqt_int16_t *ulawtoint16_ptr;
	unsigned char *int16toulaw_table;
	unsigned char *int16toulaw_ptr;
} Param;


static int
ulaw_init_ulawtoint16(oqt_audio_map_t *atrack)
{
	Param *p = (Param*)atrack->codec_private;
	int i;

	if(!p->ulawtoint16_table)
	{
    	static int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
    	int sign, exponent, mantissa, sample;
		unsigned char ulawbyte;

		p->ulawtoint16_table = malloc(sizeof(oqt_int16_t) * 256);
		p->ulawtoint16_ptr = p->ulawtoint16_table;
		for(i = 0; i < 256; i++)
		{
			ulawbyte = (unsigned char)i;
    		ulawbyte = ~ulawbyte;
    		sign = (ulawbyte & 0x80);
    		exponent = (ulawbyte >> 4) & 0x07;
    		mantissa = ulawbyte & 0x0F;
    		sample = exp_lut[exponent] + (mantissa << (exponent + 3));
    		if(sign != 0) sample = -sample;
			p->ulawtoint16_ptr[i] = sample;
		}
	}
	return 0;
}

static int
ulaw_init_int16toulaw(oqt_audio_map_t *atrack)
{
	Param *p = (Param*)atrack->codec_private;
	int i;

	if(!p->int16toulaw_table)
	{
    	int sign, exponent, mantissa;
    	unsigned char ulawbyte;
		int sample;
		int i;
    	int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
                               4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                               5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                               5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};

 		p->int16toulaw_table = malloc(65536);
		p->int16toulaw_ptr = p->int16toulaw_table + 32768;

		for(i = -32768; i < 32768; i++)
		{
			sample = i;
// Get the sample into sign-magnitude. 
    		sign = (sample >> 8) & 0x80;		// set aside the sign 
    		if(sign != 0) sample = -sample;		// get magnitude 
    		if(sample > uCLIP) sample = uCLIP;		// clip the magnitude 

// Convert from 16 bit linear to ulaw. 
    		sample = sample + uBIAS;
		    exponent = exp_lut[(sample >> 7) & 0xFF];
		    mantissa = (sample >> (exponent + 3)) & 0x0F;
		    ulawbyte = ~(sign | (exponent << 4) | mantissa);
#ifdef ZEROTRAP
		    if (ulawbyte == 0) ulawbyte = 0x02;	    // optional CCITT trap 
#endif

		    p->int16toulaw_ptr[i] = ulawbyte;
		}
	}
	return 0;
}

static oqt_int16_t ulaw_bytetoint16(Param *p, oqt_uint8_t input)
{
	return p->ulawtoint16_ptr[input];
}

static oqt_uint8_t ulaw_int16tobyte(Param *p, oqt_int16_t input)
{
	return p->int16toulaw_ptr[input];
}


static int
init_ulaw(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p;
	++Initialised;

	atrack->codec_private = malloc(sizeof(Param));
	if (!atrack->codec_private) {
		fprintf(stderr, "Error: malloc failed when creating codec_private in init_sowt\n");
		return 0;
	}
	
	// Initalise default private values
	p = (Param*)atrack->codec_private;

	p->ulawtoint16_table=0;
	p->ulawtoint16_ptr=0;
	p->int16toulaw_table=0;
	p->int16toulaw_ptr=0;
	
	// Initalise tables
	ulaw_init_ulawtoint16( atrack );
	ulaw_init_int16toulaw( atrack );
	
	return Initialised;
}

static int
delete_ulaw(oqt_t *file, int track)
{
	oqt_audio_map_t *atrack = &file->atracks[track];
	Param *p = (Param*)atrack->codec_private;

	if (p) {
		if(p->ulawtoint16_table)	free(p->ulawtoint16_table); 
		if(p->int16toulaw_table)	free(p->int16toulaw_table); 
		free(p);
		atrack->codec_private = NULL;
	}
		
	return --Initialised;
}

static
int frames_to_bytes_ulaw(
	oqt_t *file,
	int track,
	int frames)
{
	int channels = oqt_get_audio_channels(file, track);
	
	// ulaw is always 1 byte per sample
	return frames*channels;
}

static int
decode_ulaw(oqt_t *file, int track, 
	       oqt_uint32_t inputsize, 
	       oqt_uint8_t  *input,
	       oqt_uint32_t outputsize, 
	       oqt_uint8_t  *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	oqt_int16_t *output_ptr = (oqt_int16_t*)output;
	int i;
	
	if (oqt_get_audio_bits(file,track)!=16) {
		fprintf(stderr, "decode_ulaw: error, this codec only works for 16bit audio.\n");
		return 0;
	}
	
	for(i=0;i<inputsize;i++) {
		output_ptr[i] = ulaw_bytetoint16(p, input[i]);
	}		
	
	// outputsize == inputsize * 2
	return inputsize*2;
}



static int
encode_ulaw(oqt_t *file, int track,
	       oqt_uint32_t num_samples,
	       oqt_uint8_t *input,
	       oqt_uint32_t out_size,
	       oqt_uint8_t *output)
{
	Param *p = (Param*)(file->atracks[track].codec_private);
	int i,samples = num_samples * oqt_get_audio_channels(file,track);
	oqt_uint16_t* input_ptr = (oqt_uint16_t*)input;
	
	if (oqt_get_audio_bits(file,track)!=16) {
		fprintf(stderr, "encode_ulaw: error, this codec only works for 16bit audio.\n");
		return 0;
	}
	
	for(i=0;i<samples;i++) {
		output[i] = ulaw_int16tobyte(p, input_ptr[i]);
	}	
	
	return samples;	
}


oqt_codec_info_t*
oqt_plugin_register_ulaw( void )
{
	oqt_audio_codec_t *codec;

	// Allocate memory for codec record
	codec = oqt_allocate_audio_codec();
	if (!codec) return NULL;
	
	// Store information about the codec
	strncpy(codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;

	// Set the callbacks
	codec->init_codec = init_ulaw;
	//codec->prepare_encoder = prepare_encode_ulaw;
	codec->frames_to_bytes = frames_to_bytes_ulaw;
	codec->encode = encode_ulaw;
	codec->decode = decode_ulaw;
	codec->delete_codec = delete_ulaw;

	return &(codec->info);
}
