/* io_curl.c
 * Copyright (C) 2003 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: io_curl.c,v 1.9 2003/05/15 13:23:10 nhumfrey Exp $
 */
 
/* 
 * WWW I/O Routines
 */ 

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <stdio.h>
#include <unistd.h>
#include <time.h>

#include "openquicktime/structs.h"
#include "openquicktime/openquicktime.h"

#include "funcprotos.h"

#ifdef USECURL

#include <curl/curl.h>

#if (LIBCURL_VERSION_NUM < 0x070907)
#error "too old libcurl version, get the latest!"
#endif


typedef struct {
	CURL *curl;
	
	char buffer[CURL_MAX_WRITE_SIZE];
	char *readptr; 		// read from here
	char *writeptr; 	// write to here
	oqt_int64_t bytes;	// bytes available from read pointer

	oqt_int64_t pos;	// Position in the file
} stream_t, *stream_ptr;


// we use a global one for convenience
CURLM *oqt_curl_multi_handle = NULL;


static
size_t write_callback(char *buffer,
                      size_t size,
                      size_t nitems,
                      void *userp)
{
	stream_ptr stream = (stream_ptr)userp;
	fprintf(stderr, "write_callback (%d,%d) bytes=%lld/%d.\n",
		(int)size, (int)nitems,stream->bytes,CURL_MAX_WRITE_SIZE);
	//fwrite(stream->buffer, size, nitems, stdout);
	size *= nitems;
	
	memcpy(stream->readptr, buffer, size);
	stream->writeptr += size;
	stream->bytes += size;
	
	return size;
}




static int close_curl(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;

	fprintf(stderr, "close_curl: Closing connection.\n");

	/* make sure the easy handle is not in the multi handle anymore */
	curl_multi_remove_handle(oqt_curl_multi_handle, stream->curl);
	
	/* cleanup */
	curl_easy_cleanup(stream->curl);

	return 1;
} // Returns 0 if unsuccessful, 1 if successful



static int read_curl(oqt_t *file, char *data, oqt_int64_t size)
{
	stream_ptr stream = (stream_ptr)file->stream;
	fd_set fdread;
	fd_set fdwrite;
	fd_set fdexcep;
	int still_running, abort=0;
	struct timeval timeout;
	int rc, maxfd;
	
	fprintf(stderr, "read_curl: Starting read (%lld).\n", size);
	
	/* If we already have data then don't bother reading more */
	if (!stream->bytes) {
	
		/* Do a little work before the select */
		while(CURLM_CALL_MULTI_PERFORM ==
			curl_multi_perform(oqt_curl_multi_handle, &still_running)) {
			if(stream->bytes) {
			  fprintf(stderr, "read_curl: there is data available before select.\n");
			  break;        
			}
		}
	
	
		/* Get some data from the network */
		while (still_running && !abort && !stream->bytes) {
			fprintf(stderr, "Start of while loop.\n");
	
			FD_ZERO(&fdread);
			FD_ZERO(&fdwrite);
			FD_ZERO(&fdexcep);
			
			/* set a suitable timeout to fail on */
			timeout.tv_sec = 120; /* 2 minutes */
			timeout.tv_usec = 0;
			
			/* get file descriptors from the transfers */
			curl_multi_fdset(oqt_curl_multi_handle, &fdread, &fdwrite, &fdexcep, &maxfd);
			
			fprintf(stderr, "read_curl: Before select.\n");
			rc = select(maxfd+1, &fdread, &fdwrite, &fdexcep, &timeout);
			
			switch(rc) {
				case -1:
					/* select error */
					fprintf(stderr, "read_curl: Select Error.\n");
					abort = 1;
					break;
				case 0:
					/* Timeout */
					fprintf(stderr, "read_curl: Select Timeout.\n");
					abort = 1;
					break;
				default:
					/* readable/writable sockets */
					fprintf(stderr, "read_curl: Select Default.\n");
					
					while(CURLM_CALL_MULTI_PERFORM ==
						curl_multi_perform(oqt_curl_multi_handle, &still_running)) {
						if(stream->bytes) {
						  // we have received data, return that now 
						  break;        
						}
					}
					
				break;
			}
		}
	}


	/* Is there some data to return ? */ 
	if(stream->bytes) {
		/* data already available, return that */
		if(stream->bytes < size)
			size = stream->bytes;
		
		memcpy(data, stream->buffer, size);
		//stream->readptr += size;
		stream->bytes -= size;
		
		fprintf(stderr, "read_curl: return %lld bytes\n", size);
		
		return size;
	} else {
		fprintf(stderr, "read_curl: nothing available.\n");
		return 0; /* no data available to return */
	}

} // Returns 0 if unsuccessful, 1 if successful



static int write_curl(oqt_t *file, char *data, oqt_int64_t size)
{
	fprintf(stderr, "write_curl: Error writing not supported.\n");
	return 0;
} // Returns 0 if unsuccessful, 1 if successful


static int fseek_curl(oqt_t *file, oqt_int64_t offset)
{
	fprintf(stderr, "fseek_curl: Seeking to %lld.\n", offset);

	return 0;
} // Returns 0 if unsuccessful, 1 if successful


static oqt_int64_t ftell_curl(oqt_t *file)
{
	/* Work around a bug in glibc where ftello returns only 32 bits by maintaining
   our own position */
	//return file->ftell_position;
	fprintf(stderr, "ftell_curl: returning -1\n");
	
	return -1;
}
/* Returns -1 if unsuccessful, or position if successful */


static oqt_int64_t flen_curl(oqt_t *file)
{
	fprintf(stderr, "flen_curl: returning -1\n");

	// Moo !
	return -1;
}
/* Returns -1 if unsuccessful, or length if successful */





/*
	PUBLIC
*/

OPENQUICKTIMELIB_API oqt_t*
OQT_STDCALL oqt_open_curl(char *url)
{
	oqt_t *new_file = malloc(sizeof(oqt_t));
	stream_ptr stream = malloc(sizeof(stream_t));
	char useragent[128];
	int still_running = 0;

	// Check mallocs worked
	if (!new_file || !stream) {
		fprintf(stderr, "oqt_open_curl: Failed to allocate memory for structures.\n");
		if (new_file) free(new_file);
		if (stream) free(stream);
		return NULL;
	}
	
	
	// Initalise the file record
	oqt_init(new_file);
	new_file->stream = stream;
	new_file->stream_mode = OQT_MODE_READ;
	new_file->stream_type = strdup("curl");
	new_file->stream_reference = strdup(url);
	
	
	// Setup callbacks
	new_file->oqt_read = read_curl;
	new_file->oqt_fseek = fseek_curl;
	new_file->oqt_ftell = ftell_curl;
	new_file->oqt_flen = flen_curl;
	new_file->oqt_write = write_curl;
	new_file->oqt_close = close_curl;
	
	
	/* Initialise CURL library */
	stream->bytes = 0;
	stream->readptr = stream->buffer;
	stream->curl = curl_easy_init();
	curl_easy_setopt(stream->curl, CURLOPT_URL, url);
	curl_easy_setopt(stream->curl, CURLOPT_FILE, stream);
	curl_easy_setopt(stream->curl, CURLOPT_VERBOSE, FALSE);
	curl_easy_setopt(stream->curl, CURLOPT_WRITEFUNCTION, write_callback);
	
	/* Set user agent string */
	snprintf(useragent, sizeof(useragent),
             PACKAGE_NAME "/" PACKAGE_VERSION " (" OQT_HOST ") %s", curl_version());
    curl_easy_setopt(stream->curl, CURLOPT_USERAGENT, &useragent[0]);


	/* Initalise CURL Multi */
	if(!oqt_curl_multi_handle)
		oqt_curl_multi_handle = curl_multi_init();

	curl_multi_add_handle(oqt_curl_multi_handle, stream->curl);

	while(CURLM_CALL_MULTI_PERFORM ==
		curl_multi_perform(oqt_curl_multi_handle, &still_running));


	/* if still_running would be 0 now, we should return NULL */
	if (!still_running) {
		fprintf(stderr, "oqt_open_curl: failed.\n");
		if (new_file) free(new_file);
		if (stream) free(stream);
		return NULL;
	}

	return new_file;
}

#else // USECURL

OPENQUICKTIMELIB_API oqt_t*
OQT_STDCALL oqt_open_curl(char *url)
{
	fprintf(stderr, "OpenQuicktime was not compiled with libcurl support.\n");
	
	return NULL;
}

#endif // USECURL


