/* io_file.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: io_file.c,v 1.8 2003/02/08 04:23:19 jhatala Exp $
 */
 
/* 
 * Disk I/O Routines
 * Many of these routines should not be called directly
 */ 

#include <sys/stat.h>
#include <stdio.h>
#include <time.h>

#include "funcprotos.h"

#include "openquicktime/structs.h"
#include "openquicktime/openquicktime.h"





/* 64 bit file seeking support ? */
#if OQT_64SEEK
#define FTELL(f) ftello64((FILE*)f)
#define FSEEK(f,o,w) fseeko64((FILE*)f,o,w)
#define FOPEN(p,m) fopen64(p,m)
#else
#define FTELL(f) ftell((FILE*)f)
#define FSEEK(f,o,w) fseek((FILE*)f,o,w)
#define FOPEN(p,m) fopen(p,m)
#endif


typedef struct {
	FILE*	file_ptr;

	/* Work around a bug in glibc where ftello returns only 32 bits by maintaining
	   our own position - should only used by oqt_*_file functions */
	oqt_int64_t	ftell_position;
} stream_t, *stream_ptr;





/*
	PRIVATE
*/


static int close_file(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;
	int result;
	
	// Close the file
	result = fclose(stream->file_ptr);

	// Free the stream record
	free(file->stream);
	file->stream = NULL;
	
	if (result==0)	return 1;	// Success
	else return 0;				// Error is in errno
} // Returns 0 if unsuccessful, 1 if successful



static int read_file(oqt_t *file, char *data, oqt_int64_t size)
{
	stream_ptr stream = (stream_ptr)file->stream;
	int result;
	
	//fprintf(stderr,"pos=%6li size=%li\n",(oqt_int32_t)oqt_get_position(file),(oqt_int32_t)size);
	result = fread(data, (size_t)size, 1, stream->file_ptr);
	
	if (result) stream->ftell_position += size;
	return result;
} // Returns 0 if unsuccessful, 1 if successful



static int write_file(oqt_t *file, char *data, oqt_int64_t size)
{
	stream_ptr stream = (stream_ptr)file->stream;

	int result = fwrite(data, size, 1, stream->file_ptr);
	if (result) stream->ftell_position += size;
	return result;
} // Returns 0 if unsuccessful, 1 if successful


static int fseek_file(oqt_t *file, oqt_int64_t offset)
{
	stream_ptr stream = (stream_ptr)file->stream;

	// Already in position ?
	if (offset == stream->ftell_position) return 1;
	
	//printf("oqt_fseek: file-> offset=$%8.8llx pos=$%8.8llx ftell=$%8.8llx\n",
	//			offset, file->file_position, stream->ftell_position);

	stream->ftell_position = offset;
	if(FSEEK(stream->file_ptr, offset, SEEK_SET))
	{
		perror("fseek_file");
		return 0;
	}
	return 1;
} // Returns 0 if unsuccessful, 1 if successful


static oqt_int64_t ftell_file(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;

	/* Work around a bug in glibc where ftello returns only 32 bits by maintaining
   our own position */
	return stream->ftell_position;
}
/* Returns -1 if unsuccessful, or length if successful */

// get the length of a standard file
static oqt_int64_t flen_file(oqt_t *file)
{
#ifndef WIN32
	stream_ptr stream = (stream_ptr)file->stream;
    struct stat status;
    
	if(fstat(fileno(stream->file_ptr), &status))
		perror("oqt_flen_file fstat:");
	return status.st_size;
#else
	fprintf(stderr, "flen_file: not supported on Win32.\n");
	return -1;
#endif

/* Alternative implementation
	FSEEK(stream->file_ptr, 0, SEEK_END);
	len = ftell(stream->file_ptr);
	FSEEK(stream->file_ptr, stream->ftell_position, SEEK_SET);
*/
}
/* Returns -1 if unsuccessful, or length if successful */




/*
	PUBLIC
*/

OPENQUICKTIMELIB_API oqt_t*
OQT_STDCALL oqt_open_file(char *filepath, int mode)
{
	oqt_t *new_file = malloc(sizeof(oqt_t));
	stream_ptr stream = malloc(sizeof(stream_t));
	char flags[10];
	int exists = 0;

	// Check mallocs worked
	if (!new_file || !stream) {
		fprintf(stderr, "oqt_open_file: Failed to allocate memory for structures.\n");
		if (new_file) free(new_file);
		if (stream) free(stream);
		return NULL;
	}
	
	
	// Initalise the file record
	oqt_init(new_file);
	new_file->stream = stream;
	new_file->stream_mode = mode;
	new_file->stream_type = strdup("file");
	new_file->stream_reference = strdup(filepath);
	
	
	// Setup callbacks
	new_file->oqt_read = read_file;
	new_file->oqt_fseek = fseek_file;
	new_file->oqt_ftell = ftell_file;
	new_file->oqt_flen = flen_file;
	new_file->oqt_write = write_file;
	new_file->oqt_close = close_file;
	
	
	// Check if the file exists
	if((mode&OQT_MODE_READ) && (stream->file_ptr = FOPEN(filepath, "rb"))) {
		exists = 1; 
		fclose(stream->file_ptr); 
	}
	
	// Work out what mode to open the file with
	if(mode == OQT_MODE_READ) sprintf(flags, "rb");
	else if(mode == OQT_MODE_WRITE) sprintf(flags, "wb");
	else if(mode == OQT_MODE_READWRITE) {
		if(exists) sprintf(flags, "rb+");
		else       sprintf(flags, "wb+");
	}
	
	if(!(stream->file_ptr = FOPEN(filepath, flags))) {
		perror("oqt_open_file");
		free(stream);
		free(new_file);
		return NULL;
	}
	
	return new_file;
}

