/* io_memfile.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: io_memfile.c,v 1.5 2003/01/19 17:33:00 nj_humfrey Exp $
 */

/* 
 * Memory I/O Routines
 * Read/Write a psudeo quicktime file from memory buffer
 */ 

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"


typedef struct {
	oqt_int64_t	pos;
	oqt_int64_t	size;
	oqt_int64_t	max_len;
	char		we_malloced;
	char*		data;
} stream_t, *stream_ptr;






/*
	PRIVATE
*/

static int
read_mem(oqt_t *file, char *data, oqt_int64_t size)
{
	stream_ptr stream = (stream_ptr)file->stream;
	//fprintf(stderr,"read_mem: size=%lld pos=%lld len=%lld\n", size, stream->pos, stream->size);
	
	// Make sure it is within range
	if ((size+stream->pos)>stream->size) {
		fprintf(stderr, "read_mem: read to outside file size.\n");
		return 0;
	}
	
	// Copy accross the memory
	memcpy(data, &(stream->data[stream->pos]), size);
	stream->pos += size;
	
	return 1;
}

static int
fseek_mem(oqt_t *file, oqt_int64_t offset)
{
	stream_ptr stream = (stream_ptr)file->stream;

	//fprintf(stderr, "fseek_mem: moving offset from %lld to %lld.\n", stream->pos, offset);
	
	// Make sure it is within range
	if (offset>stream->size) {
		fprintf(stderr, "fseek_mem: seek to outside file size.\n");
		return 0;
	}
	stream->pos = offset;
	
	return 1;
}

static int
write_mem(oqt_t *file, char *data, oqt_int64_t size)
{
	stream_ptr stream = (stream_ptr)file->stream;
	//fprintf(stderr,"write_mem: size=%lld pos=%lld len=%lld data=0x%x\n", size, stream->pos, stream->size, data);
	
	// Make sure it is within range
	if ((size+stream->pos)>stream->size) {
		fprintf(stderr, "write_mem: write to outside file size.\n");
		return 0;
	}
	
	// Copy accross the memory
	memcpy(&(stream->data[stream->pos]), data, size);
	stream->pos += size;
	if (stream->pos > stream->max_len)
		stream->max_len = stream->pos;
	
	return 1;
}


// Returns length of file (not the buffer)
static oqt_int64_t
flen_mem(oqt_t *file) {
	stream_ptr stream = (stream_ptr)file->stream;
	return stream->max_len;
}
/* Returns -1 if unsuccessful, or length if successful */

static oqt_int64_t
ftell_mem(oqt_t *file) {
	stream_ptr stream = (stream_ptr)file->stream;
	return stream->pos;
}
/* Returns -1 if unsuccessful, or position if successful */


static int
close_mem(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;
	//fprintf(stderr, "close_mem: closing memory file - freeing memory.\n");
	
	if (stream) {
		if (stream->data && stream->we_malloced)
			free(stream->data);
		free(file->stream);
		file->stream = NULL;
	}
	return 1;
} // Returns 0 if unsuccessful, 1 if successful



/*
	PUBLIC
*/

OPENQUICKTIMELIB_API oqt_t*
OQT_STDCALL oqt_open_memfile(char* ptr, int size)
{
	oqt_t *new_file = malloc(sizeof(oqt_t));
	stream_ptr stream = malloc(sizeof(stream_t));
	char ref[256];

	// Check mallocs worked
	if (!new_file || !stream) {
		fprintf(stderr, "oqt_open_memfile: Failed to allocate memory for structures.\n");
		if (new_file) free(new_file);
		if (stream) free(stream);
		return NULL;
	}
	
	// Allocate memory for the memory file data
	if (ptr) {
		stream->data = ptr;
		stream->we_malloced = 0;
	} else {
		stream->data = malloc(size);
		stream->we_malloced = 1;
		if (!stream->data) {
			fprintf(stderr, "oqt_open_memfile: Failed to allocate memory for memory file.\n");
			free(stream);
			free(new_file);
			return NULL;
		}
	}	
	
	
	// Intialise the stream descriptor
	stream->pos = 0;
	stream->max_len = 0;
	stream->size = size;
		
	
	// Initalise the file record
	oqt_init(new_file);
	new_file->stream = stream;
	new_file->stream_mode = OQT_MODE_READWRITE;
	new_file->stream_type = strdup("memfile");
	sprintf(ref, "%lld bytes @ 0x%8.8lx", stream->size, (unsigned long)stream->data);
	new_file->stream_reference = strdup(ref);
	//printf("Ref: %s\n", new_file->stream_reference);
	
	// Setup callbacks
	new_file->oqt_read = read_mem;
	new_file->oqt_fseek = fseek_mem;
	new_file->oqt_ftell = ftell_mem;
	new_file->oqt_flen = flen_mem;
	new_file->oqt_write = write_mem;
	new_file->oqt_close = close_mem;
	
	
	return new_file;
}


OPENQUICKTIMELIB_API char*
OQT_STDCALL oqt_memfile_get_ptr(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;
	return stream->data;
}


OPENQUICKTIMELIB_API oqt_int64_t 
OQT_STDCALL oqt_memfile_get_len(oqt_t *file)
{
	stream_ptr stream = (stream_ptr)file->stream;
	return stream->size;
}

