/* openquicktime.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: openquicktime.c,v 1.36 2003/04/07 21:02:23 shitowax Exp $
 */

#ifdef __MACOS__
#include "macconfig.h"
#endif

#include "openquicktime/structs.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/openquicktime.h"

#include "funcprotos.h"

#include <string.h>

#ifdef WIN32
#include <windows.h>

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

#else

#endif

#ifndef PACKAGE_VERSION
# define PACKAGE_VERSION VERSION
#endif

static const char* oqt_vers = PACKAGE_VERSION;


OPENQUICKTIMELIB_API const char*
OQT_STDCALL oqt_version()
{
	return oqt_vers;
}





// ================================== Entry points =============================




// Guess I/O handler to use and open read only
OPENQUICKTIMELIB_API oqt_t*
OQT_STDCALL oqt_open(char* reference)
{
	// Reference could be URL / file path / RTP IP address etc...
	
#ifdef USECURL
	// for now all non-file URLs go to curl
	if(strstr(reference,"://") != NULL &&
	   strncmp(reference, "file:", 5) != 0) {
		// non-file URL 
		return oqt_open_curl(reference);

	} else {
#endif
		// file URL: skip protocol if present
		if(strncmp(reference, "file:", 5) == 0) {
			reference += 5;
		}
		// skip host if present
		if(strncmp(reference, "//", 2) == 0) {
			char *p = strchr(reference+2, '/');
			if(p) {
				reference = p;
			}
		}
		return oqt_open_file(reference, OQT_MODE_READ);
#ifdef USECURL
	}
#endif
}





OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_close(oqt_t *file)
{
	int i, result=0;
	
	// Close the stream
	result = file->oqt_close(file);
	if (!result) {
		perror("oqt_close");
	}

	// Close referenced external resources
	for(i = 0; i < file->ext_count; ++i) {
		if(file->ext_files[i]) {
			result = file->ext_files[i]->oqt_close(file->ext_files[i]);
			if (!result) {
				perror("oqt_close");
			}
			free(file->ext_files[i]);
			file->ext_files[i] = NULL;
		}
	}

	// Free string information about stream (strdup)
	file->stream_mode = 0x00;
	if (file->stream_type)	{
		free(file->stream_type);
		file->stream_type=NULL;
	}
	if (file->stream_reference) {
		free(file->stream_reference);
		file->stream_reference=NULL;
	}

	if (!file->oqt_delete_audio_map)
		fprintf(stderr, "oqt_close: Warning, oqt_delete_audio_map is not available.\n");
	if (!file->oqt_delete_video_map)
		fprintf(stderr, "oqt_close: Warning, oqt_delete_video_map is not available.\n");


	// Delete Audio Tracks
	if(file->total_atracks) 
	{
		if(file->oqt_delete_audio_map)
			for(i = 0; i < file->total_atracks; i++)
				file->oqt_delete_audio_map(file, i);
		free(file->atracks);
		file->atracks = NULL;
		file->total_atracks = 0;
	}

	// Delete Video Tracks
	if(file->total_vtracks)
	{
		if(file->oqt_delete_video_map)
			for(i = 0; i < file->total_vtracks; i++)
				file->oqt_delete_video_map(file, i);
		free(file->vtracks);
		file->vtracks = NULL;
		file->total_vtracks = 0;
	}

	// Delete preload buffer
	if(file->preload_size)
	{
		free(file->preload_buffer);
		file->preload_size = 0;
	}

	// Delete the atom trees
	oqt_moov_delete(&(file->moov));
	oqt_mdat_delete(&(file->mdat));

	free(file);
	return result;
}



OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_init(oqt_t* file)
{
	// Allocate memory for the file record
	if (!file) {
		fprintf(stderr, "oqt_init: error, passed in NULL pointer\n");
		return 1;
	}

	// Zero the memory
	memset(file, 0, sizeof(oqt_t));
	
	file->stream = NULL;
	file->stream_mode = 0x00;
	file->stream_type = NULL;
	file->stream_reference = NULL;

	file->decompressed_buffer_size = 0;
	file->decompressed_buffer = NULL;
	file->decompressed_position = 0;

	// default: data is in the same file
	file->data_file = file;

	// No audio/video tracks yet
	file->total_atracks = 0;
	file->total_vtracks = 0;
	file->atracks = NULL;
	file->vtracks = NULL;
	
	// Initilise the data structures
	oqt_mdat_init(&(file->mdat));
	oqt_moov_init(&(file->moov));


	// We configure the library for simple codecs
	//  - Can be set to something else later, if required
	file->oqt_init_audio_map = oqt_init_audio_map;
	file->oqt_init_video_map = oqt_init_video_map;
	file->oqt_delete_audio_map = oqt_delete_audio_map;
	file->oqt_delete_video_map = oqt_delete_video_map;
	
	
	// Defaults to stub methods
	file->oqt_read = oqt_read_stub;
	file->oqt_write = oqt_write_stub;
	file->oqt_fseek = oqt_fseek_stub;
	file->oqt_ftell = oqt_ftell_stub;
	file->oqt_flen = oqt_flen_stub;
	file->oqt_close = oqt_close_stub;

	return 0;
}


OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_dump(oqt_t *file)
{
  //	printf("oqt_dump\n");
	printf("movie data\n");
	printf(" size %lld\n", file->mdat.atom.size);
	printf(" start %lld\n", file->mdat.atom.start);
	oqt_moov_dump(&(file->moov));
	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_read_headers(oqt_t *file)
{
	int result = 0, found_moov = 0, found_mdat = 0;
	int i, track;
	oqt_int64_t orig_position = oqt_get_position(file);
	oqt_atom_t leaf_atom;

	if(!(file->stream_mode & OQT_MODE_READ))
	{
		fprintf(stderr, "oqt_read_headers: file is not open with read access.\n");
		return 0;
	}
	
	
	// Move to start of file
	oqt_set_position(file, 0/*LL*/);

	// Find the postions of the movie header and data
	do
	{
		result = oqt_atom_read_header(file, &leaf_atom);

		if(!result)
		{
			if(oqt_atom_is(&leaf_atom, "mdat")) 
			{
				oqt_read_mdat(file, &(file->mdat), &leaf_atom);
				found_mdat = 1;
			}
			else
			if(oqt_atom_is(&leaf_atom, "moov")) 
			{
				oqt_read_moov(file, &(file->moov), &leaf_atom);
				found_moov = 1;
			}
			else
			if(oqt_atom_is(&leaf_atom, "free") || oqt_atom_is(&leaf_atom, "skip")) 
			{
				// Free space in file
				oqt_atom_skip(file, &leaf_atom);
			}
			else {
				#ifdef DEBUG
					fprintf(stderr, "oqt_read_headers: ignoring root '%4.4s' atom.\n", leaf_atom.type);
				#endif
				oqt_atom_skip(file, &leaf_atom);
			}
		}
		//printf("oqt_read_headers: ftell 0x%llx length %lld\n", oqt_get_position(file), total_length);
		//	}while(!result && oqt_get_position(file) < total_length);
	}while(!result && (found_mdat + found_moov != 2));


	#ifdef DEBUG
		if (!found_moov) fprintf(stderr, "oqt_read_headers: Didn't find moov atom in file.\n");
		if (!found_mdat) fprintf(stderr, "oqt_read_headers: Didn't find mdat atom in file.\n");
	#endif


	// go back to the original position 
	oqt_set_position(file, orig_position);
	
	
	// Check functions are available
	if (!file->oqt_init_audio_map) {
		fprintf(stderr, "oqt_read_headers: oqt_init_audio_map is not available.\n");
		return -1;
	} if (!file->oqt_init_video_map) {
		fprintf(stderr, "oqt_read_headers: oqt_init_video_map is not available.\n");
		return -1;
	}


	if(found_moov)
	{
		// get tables for all the different tracks
		file->total_atracks = oqt_get_audio_track_count(file);
		if (file->total_atracks) {
			file->atracks = (oqt_audio_map_t*)calloc(1, sizeof(oqt_audio_map_t) * file->total_atracks);
	
			for(i = 0, track = 0; i < file->total_atracks; i++)
			{
				while(!file->moov.trak[track]->mdia.minf.is_audio)
					track++;
				file->oqt_init_audio_map(file, i, file->moov.trak[track]);
			}
		}
		
		file->total_vtracks = oqt_get_video_track_count(file);
		if (file->total_vtracks) {
			file->vtracks = (oqt_video_map_t*)calloc(1, sizeof(oqt_video_map_t) * file->total_vtracks);
	
			for(track = 0, i = 0; i < file->total_vtracks; i++)
			{
				while(!file->moov.trak[track]->mdia.minf.is_video)
					track++;
	
				file->oqt_init_video_map(file, i, file->moov.trak[track]);
			}
		}
	}

	return !found_moov;
}
// If no MOOV atom exists return 1 else return 0.



OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_write_moviedata_init(oqt_t* file)
{
	if(!(file->stream_mode & OQT_MODE_WRITE))
	{
		fprintf(stderr, "oqt_write_moviedata_init: can't call without write access.\n");
		return 1;
	}
	
	// Start the mdat atom
	oqt_atom_write_header(file, &file->mdat.atom, "mdat");
	
	return 0;	// Success
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_write_moviedata_end(oqt_t* file)
{
	if(!(file->stream_mode & OQT_MODE_WRITE))
	{
		fprintf(stderr, "oqt_write_moviedata_end: can't call without write privilages.\n");
		return 1;
	}

	// End the 'mdat' atom and write the 'moov' atom
	oqt_atom_write_footer(file, &file->mdat.atom);
	oqt_write_moov(file, &(file->moov));
	
	return 0;	// Success
}



// **** It is likely that this routine will not work for files
//      not created by OpenQuicktime - due to the arbitrary length of the 'moov' atom ****

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_make_streamable(oqt_t *in_file, oqt_t *out_file)
{
	oqt_int64_t buf_size = 1000000;	// Arbitrary value ?
	char *buffer;
	int result = 0;
	
	// Check that files have been opened with correct modes
	if(!(in_file->stream_mode & OQT_MODE_READ)) {
		fprintf(stderr, "oqt_make_streamable: input file was opened without read access.\n");
		return 1;
	}
	else if(!(out_file->stream_mode & OQT_MODE_WRITE)) {
		fprintf(stderr, "oqt_make_streamable: output file was opened without write access.\n");
		return 1;
	}
	
	
	// Read information about input file
	if(oqt_read_headers(in_file)) {
		printf("oqt_make_streamable: Could not read movie headers from input file.\n");
		return 1;
	}
	
	// Sanity check
	if (in_file->moov.atom.start == in_file->mdat.atom.start) {
		fprintf(stderr, "oqt_make_streamable: moov and mdat at same offset. Something went wrong.\n");
		return 1;
	}
	
	// Is 'moov' before 'mdat' already ?
	if (in_file->moov.atom.start < in_file->mdat.atom.start) {
		fprintf(stderr, "oqt_make_streamable: 'moov' atom already preceeds 'mdat' atom.\n");
		return 0;
	} else {
		#ifdef DEBUG
		fprintf(stderr, "oqt_make_streamable: 'moov' atom is after 'mdat' atom...\n");
		#endif
	}
	
	
	
	
	// Update the offsets, now that the 'moov' atom preceeds it
	oqt_shift_offsets(&(in_file->moov), in_file->moov.atom.size);


	// Write the moov at the start of the new file		
	oqt_set_position(out_file, 0);
	oqt_write_moov(out_file, &(in_file->moov));
	
	// Move to the start of the movie data in the old file
	oqt_set_position(in_file, in_file->mdat.atom.start + OQT_ATOM_HEAD_LENGTH);


	if(!(buffer = malloc(buf_size)))
	{
		fprintf(stderr, "oqt_make_streamable: out of memory\n");
		return 1;
	}
	else
	{
		oqt_int64_t mdat_end = in_file->mdat.atom.start + in_file->mdat.atom.size;
	
		// Start the 'mdat' atom
		oqt_atom_reset(&out_file->mdat.atom);
		oqt_atom_write_header(out_file, &out_file->mdat.atom, "mdat");
	
		// Copy across the movie data
		while(oqt_get_position(in_file) < mdat_end)
		{
			if(oqt_get_position(in_file) + buf_size > mdat_end)
				buf_size = mdat_end - oqt_get_position(in_file);

			if(!oqt_read_data(in_file, buffer, buf_size)) result = 1;
			if(!oqt_write_data(out_file, buffer, buf_size)) result = 1;
			if (result) break;
		}
		
		// End the 'mdat' atom
		oqt_atom_write_footer(out_file, &out_file->mdat.atom);
		
		free(buffer);
	}
	
	printf("old len: $%llx\n", in_file->mdat.atom.size);
	printf("new len: $%llx\n", out_file->mdat.atom.size);
	
	return result;
}


/* It's the same as the previous but works on a file opened in read/write mode ... ;) */
OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_make_streamable_revisited(oqt_t *file)
{
	unsigned long reallyRead, reallyWritten;
	unsigned long size = file->moov.atom.size < 1000000 ? 1000000 : file->moov.atom.size;
	char *buffer1 = (char*)malloc(size);
	char *buffer2 = (char*)malloc(size);
	oqt_int64_t readOffset = 0, writeOffset = 0, original_moov_start = file->moov.atom.start;

	if( size > original_moov_start) size = original_moov_start;

	// we read the first part of the mdat
	oqt_set_position(file, readOffset);
	reallyRead = oqt_read_data(file, buffer1, size);
	readOffset += reallyRead;

	// Update the offsets, now that the 'moov' atom preceeds it
	oqt_shift_offsets(&(file->moov), file->moov.atom.size);

	// Write the moov at the start of the new file		
	oqt_set_position(file, writeOffset);
	oqt_write_moov(file, &(file->moov));
	writeOffset += file->moov.atom.size;

	while(reallyRead == size && readOffset <= original_moov_start) {
		char * interm;

		oqt_set_position(file, readOffset);
		reallyRead = oqt_read_data(file, buffer2, size);
		if(reallyRead == size) {
			readOffset += reallyRead;
		}
		
		oqt_set_position(file, writeOffset);
		reallyWritten = oqt_write_data(file, buffer1, size);
		writeOffset += reallyWritten;
		interm = buffer1;
		buffer1 = buffer2;
		buffer2 = interm;
	}

	oqt_set_position(file, writeOffset);
	reallyWritten = oqt_write_data(file, buffer1, original_moov_start - readOffset);	
	
	/* FIXME: what do we return? */	
	return 0;
}


/* =============================== Optimization functions */

/*OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_cpus(oqt_t *file, int cpus)
{
	if(cpus > 0) file->cpus = cpus;
	return 0;
}*/

OPENQUICKTIMELIB_API void
OQT_STDCALL oqt_set_preload(oqt_t *file, oqt_int64_t preload)
{
	if(preload)
		if(!file->preload_size)
		{
			file->preload_size = preload;
			file->preload_buffer = calloc(1, preload);
			file->preload_start = 0;
			file->preload_end = 0;
			file->preload_ptr = 0;
		}
}



OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_video_track_count(oqt_t *file)
{
	int i, result = 0;
	for(i = 0; i < file->moov.total_tracks; i++)
	{
		if(file->moov.trak[i]->mdia.minf.is_video) result++;
	}
	return result;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_audio_track_count(oqt_t *file)
{
	int i, result = 0;
	for(i = 0; i < file->moov.total_tracks; i++)
	{
		if(file->moov.trak[i]->mdia.minf.is_audio) result++;
	}
	return result;
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_audio_length(oqt_t *file, int track)
{
	oqt_stts_t *stts = &(file->atracks[track].track->mdia.minf.stbl.stts);
	
	if(track>=0 && track < file->total_atracks) {
		/* Yann: I've modified this to handle samples with a duration 
		different from 1 ... needed for ".mp3" fourcc */
		
//		return stts->table[0].sample_duration*oqt_track_samples(file, file->atracks[track].track);
		return (stts->table[0].sample_duration*oqt_track_samples(file, file->atracks[track].track)
					*oqt_get_audio_samplerate(file, track)) / file->atracks[track].track->mdia.mdhd.time_scale;
	}
	
	return 0;
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_video_length(oqt_t *file, int track)
{
/*printf("oqt_video_length %d %lld\n", oqt_track_samples(file, file->vtracks[track].track), track); */
	if(track>=0 && track < file->total_vtracks)
		return oqt_track_samples(file, file->vtracks[track].track);
	return 0;
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_audio_position(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return file->atracks[track].current_sample;
	} else {
		fprintf(stderr, "oqt_get_audio_position: invalid track number (%d)\n", track);
	}
		
	return 0;
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_video_position(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks) {
		return file->vtracks[track].current_frame;
	} else {
		fprintf(stderr, "oqt_get_video_position: invalid track number (%d)\n", track);
	}
		
	return 0;
}



OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_audio_position(oqt_t *file, int track, oqt_int64_t sample)
{
	oqt_int64_t offset, chunk_sample;
	oqt_int32_t chunk;
	oqt_trak_t *trak;
	oqt_int32_t sd_id;


	if(track>=0 && track < file->total_atracks)
	{
		trak = file->atracks[track].track;
		file->atracks[track].current_sample = sample;
		//		printf("BEFORE  oqt_chunk_of_sample track %d sample %li\n", track, sample);
		oqt_chunk_of_sample(&chunk_sample, &chunk, trak, &sd_id, sample);
		// atrack->current_chunk is the chunk currently in the buffers,
		// don't change it until we really update the buffers
		//		printf("AFTER  oqt_chunk_of_sample chunk %d chunk_sample %d\n", chunk, chunk_sample);
		offset = oqt_sample_to_offset(trak, &sd_id, sample);
		//		printf("AFTER  oqt_sample_to_offset offset %li\n", offset);
		file->data_file = oqt_file_from_sd_id(file, trak, sd_id);
		oqt_set_position(file->data_file, offset);
	}

	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_video_position(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_int64_t offset, chunk_sample;
	oqt_int32_t chunk;
	oqt_trak_t *trak;
	oqt_int32_t sd_id;

	if(track>=0 && track < file->total_vtracks)
	{
		trak = file->vtracks[track].track;
		file->vtracks[track].current_frame = frame;
		oqt_chunk_of_sample(&chunk_sample, &chunk, trak, &sd_id, frame);
		// vtrack->current_chunk is the chunk currently in the buffers,
		// don't change it until we really update the buffers
		offset = oqt_sample_to_offset(trak, &sd_id, frame);
		//fprintf(stderr, "oqt_set_video_position: offset: frame %lld -> 0x%llx\n",frame,offset);
		file->data_file = oqt_file_from_sd_id(file, trak, sd_id);
		oqt_set_position(file->data_file, offset);
	} else {
		fprintf(stderr, "oqt_set_video_position: invalid track number (%d)\n", track);
	}

	return 0;
}


OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_get_audio_samplerate(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return file->atracks[track].track->mdia.minf.stbl.stsd.table[0].sample_rate;
	} else {
		fprintf(stderr, "oqt_get_audio_samplerate: invalid track number (%d)\n", track);
	}
		
	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_audio_bits(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return file->atracks[track].track->mdia.minf.stbl.stsd.table[0].sample_size;
	} else {
		fprintf(stderr, "oqt_get_audio_bits: invalid track number (%d)\n", track);
	}

	return 0;
}

OPENQUICKTIMELIB_API char*
OQT_STDCALL oqt_get_audio_compressor(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return file->atracks[track].track->mdia.minf.stbl.stsd.table[0].format;
	} else {
		fprintf(stderr, "oqt_get_audio_compressor: invalid track number (%d)\n", track);
	}
	
	return NULL;
}

OPENQUICKTIMELIB_API const oqt_codec_info_t* 
OQT_STDCALL oqt_get_audio_codec(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return (const oqt_codec_info_t*)file->atracks[track].codec;
	} else {
		fprintf(stderr, "oqt_get_audio_codec: invalid track number (%d)\n", track);
	}
	
	return NULL;
}


OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_audio_channels(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_atracks) {
		return file->atracks[track].track->mdia.minf.stbl.stsd.table[0].channels;
	} else {
		fprintf(stderr, "oqt_get_audio_channels: invalid track number (%d)\n", track);
	}

	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_audio_channel_loc(oqt_t *file, int *oqt_track, int *oqt_channel, int channel)
{
	int current_channel = 0, current_track = 0;
	*oqt_channel = 0;
	*oqt_track = 0;
	for(current_channel = 0, current_track = 0; current_track < file->total_atracks; )
	{
		if(channel >= current_channel)
		{
			*oqt_channel = channel - current_channel;
			*oqt_track = current_track;
		}

		current_channel += oqt_get_audio_channels(file,current_track);
		current_track++;
	}
	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_video_width(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return file->vtracks[track].track->tkhd.track_width;
	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_video_height(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return file->vtracks[track].track->tkhd.track_height;
	return 0;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_video_depth(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].depth;
	return 0;
}

OPENQUICKTIMELIB_API void
OQT_STDCALL oqt_set_video_depth(oqt_t *file, int track, int depth)
{
	if(track>=0 && track < file->total_vtracks)
		file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].depth = depth;
}


OPENQUICKTIMELIB_API oqt_float_t
OQT_STDCALL oqt_get_video_framerate(oqt_t *file, int track)
{
	oqt_stts_t *stts;
	oqt_int64_t samplecount = 0, duration = 0;
	int i;

	if(track>=0 && track < file->total_vtracks) {
//		return (float)file->vtracks[track].track->mdia.mdhd.time_scale / 
//			file->vtracks[track].track->mdia.minf.stbl.stts.table[0].sample_duration;

		stts = &(file->vtracks[track].track->mdia.minf.stbl.stts);

		// We're gonna return the framerate used the most, we simply returned the first one before
		for(i = 0; i < stts->total_entries; i++)
		{
			if(samplecount < stts->table[i].sample_count) {
				samplecount = stts->table[i].sample_count;
				duration = stts->table[i].sample_duration;
			}
		}
	}

	return (float)file->vtracks[track].track->mdia.mdhd.time_scale / duration;
}

OPENQUICKTIMELIB_API oqt_int64_t 
OQT_STDCALL oqt_get_video_frame_duration(oqt_t *file, int track, oqt_int64_t frame_number)
{
	// returns the duration of the frame in 1/10000000 second .... :)
	oqt_stts_t *stts;
	oqt_int64_t samplecount = 0, duration = 0;
	double sec_duration;
	int i;

	if(track >=0 && track < file->total_vtracks) {
		stts = &(file->vtracks[track].track->mdia.minf.stbl.stts);

		for(i = 0; i < stts->total_entries; i++)
		{
			samplecount += stts->table[i].sample_count;
			duration = stts->table[i].sample_duration;
			if(samplecount >= frame_number)
				break;
		}

		sec_duration = (double)duration/(double)file->vtracks[track].track->mdia.mdhd.time_scale;
		duration = sec_duration * 10000000;
	}

	return duration;
}

OPENQUICKTIMELIB_API oqt_int32_t 
OQT_STDCALL oqt_get_audio_frame_duration(oqt_t *file, int track, oqt_int64_t frame_number)
{
	// returns the duration of the frame in samples
	oqt_stts_t *stts;
	oqt_int64_t samplecount = 0, duration = 0;
	double sec_duration;
	int i;

	if(track >=0 && track < file->total_atracks) {
		stts = &(file->atracks[track].track->mdia.minf.stbl.stts);

		for(i = 0; i < stts->total_entries; i++)
		{
			samplecount += stts->table[i].sample_count;
			duration = stts->table[i].sample_duration;
			if(samplecount >= frame_number)
				break;
		}
	}

	return duration;
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_video_poster(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return (oqt_int64_t)file->moov.mvhd.poster_time /
		       ((oqt_float_t)file->vtracks[track].track->mdia.minf.stbl.stts.table[0].sample_duration *
		        ((oqt_float_t)file->vtracks[track].track->mdia.mdhd.time_scale / file->moov.mvhd.time_scale));

	return 0;
}

OPENQUICKTIMELIB_API void
OQT_STDCALL oqt_set_video_poster(oqt_t *file, int track, oqt_int64_t frame)
{
	if(track>=0 && track < file->total_vtracks)
		file->moov.mvhd.poster_time = (int)(frame * 
		       ((oqt_float_t)file->vtracks[track].track->mdia.minf.stbl.stts.table[0].sample_duration *
		        ((oqt_float_t)file->vtracks[track].track->mdia.mdhd.time_scale / file->moov.mvhd.time_scale)));
}


OPENQUICKTIMELIB_API void
OQT_STDCALL oqt_set_video_framerate(oqt_t *file, oqt_float_t framerate)
{
	int i;
	int new_time_scale, new_sample_duration;

	new_time_scale = oqt_get_timescale(framerate);
	new_sample_duration = (int)((oqt_float_t)new_time_scale / framerate + 0.5);

	for(i = 0; i < file->total_vtracks; i++)
	{
		file->vtracks[i].track->mdia.mdhd.time_scale = new_time_scale;
		file->vtracks[i].track->mdia.minf.stbl.stts.table[0].sample_duration = new_sample_duration;
	}
}

OPENQUICKTIMELIB_API char*
OQT_STDCALL oqt_get_video_compressor(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].format;

	return NULL;
}

OPENQUICKTIMELIB_API const oqt_codec_info_t* 
OQT_STDCALL oqt_get_video_codec(oqt_t *file, int track)
{
	if(track>=0 && track < file->total_vtracks)
		return (const oqt_codec_info_t*)file->vtracks[track].codec;
	
	return NULL;
}

/* stux: it was necessary to get audio sample sizes for framed audio */
OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_get_audio_framesize(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_int32_t bytes = 0;
	oqt_trak_t *trak;
	
	if(track>=0 && track < file->total_atracks)
		trak = file->atracks[track].track;
	else
		return 0;

	if( trak->mdia.minf.stbl.stsz.sample_size )
	{
		bytes = trak->mdia.minf.stbl.stsz.sample_size;
	}
	else
	{
		oqt_int32_t total_frames = oqt_track_samples(file, trak);
		
		if(frame < 0)
			frame = 0;
		else if(frame > total_frames - 1)
			frame = total_frames - 1;
			
		bytes = trak->mdia.minf.stbl.stsz.table[frame].size;
	}


	return bytes;
}

OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_get_video_framesize(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_int32_t bytes = 0;
	oqt_trak_t *trak;
	
	if(track>=0 && track < file->total_vtracks)
			trak = file->vtracks[track].track;
	else
		return 0;

	if(trak->mdia.minf.stbl.stsz.sample_size)
	{
		bytes = trak->mdia.minf.stbl.stsz.sample_size;
	}
	else
	{
		oqt_int32_t total_frames = oqt_track_samples(file, trak);
		
		if(frame < 0)
			frame = 0;
		else if(frame > total_frames - 1)
			frame = total_frames - 1;
			
		bytes = trak->mdia.minf.stbl.stsz.table[frame].size;
	}


	return bytes;
}


OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_get_audio_chunks(oqt_t *file, int track)
{
	return file->atracks[track].track->mdia.minf.stbl.stco.total_entries;
}

/* the length in bytes of a chunk
	chunk should be a number greater than 0 */
OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_get_audio_chunksize(
	oqt_t *file,
	int track,
	oqt_int32_t chunk)
{
	oqt_audio_map_t *atrack = &(file->atracks[track]);
	oqt_stsc_table_t *stsc_table = atrack->track->mdia.minf.stbl.stsc.table;
	oqt_stsz_table_t *stsz_table = atrack->track->mdia.minf.stbl.stsz.table;
	int stsc_entries = atrack->track->mdia.minf.stbl.stsc.total_entries;
	int stsz_entries = atrack->track->mdia.minf.stbl.stsz.total_entries;
	
	int sample_size = atrack->track->mdia.minf.stbl.stsz.sample_size;
	int chunk_bytes=0, first_sample=0;//, offset_bytes;
	int i, entry;
	
		
	for(entry=0,i=1;i<=chunk;i++) {
		if ((entry+1)<stsc_entries && stsc_table[entry+1].chunk==i) entry++;
		
		if (i==chunk) break;
		first_sample+=stsc_table[entry].samples;
		//fprintf(stdout, "chunk=%.2d    entry=%.2d   ec=%.2d   samples=%.2d   total_samples=%d\n",
		//			i, entry, stsc_table[entry].chunk, stsc_table[entry].samples, first_sample);
		if (entry>=stsc_entries) {
			fprintf(stderr, "oqt_get_audio_chunksize: ran out of stsc entries (%d).\n", entry);
			break;
		}
	
	}
	
	
	// Ok now we know what sample we have got to
	//fprintf(stdout, "%lld: upto=%ld   entry=%ld   this chunk=%ld\n", chunk, 
	//		first_sample, entry, stsc_table[entry].samples);
	

	if(sample_size)
	{
	
		if (sample_size == 1) {
			// Usually a sample size of 1 is meaningless and
			// we must calculate it ourselves...
			chunk_bytes = oqt_get_audio_frames_to_bytes(file, track, stsc_table[entry].samples);
			
			if (chunk_bytes<=0 && stsc_table[entry].samples>0) {
				fprintf(stderr, "oqt_get_audio_chunksize: No stsz table and codec does not support frames_to_bytes().\n");
				chunk_bytes = sample_size * stsc_table[entry].samples;	// Poor attempt at a guess
			}

		} else {
			chunk_bytes = sample_size * stsc_table[entry].samples;
		}
		
	// If sample_size is zero then look it up in the table
	} else {
		int last_sample = first_sample + stsc_table[entry].samples;
		
		for(i = first_sample; i < last_sample; ++i)
		{
			if (i>=stsz_entries) {
				fprintf(stderr, "oqt_get_audio_chunksize: ran out of stsz entries (%d).\n", i);
				break;
			}
			chunk_bytes += stsz_table[i].size;
		}
	}

	// Difference between the 2 sound chunk offsets 
	//offset_bytes =  oqt_chunk_to_offset(atrack->track, chunk+1) - 
	//			   oqt_chunk_to_offset(atrack->track, chunk);
	//fprintf(stdout, "Newly calced length of chunk is %d\n", chunk_bytes);
	//fprintf(stdout, "Offset Length of chunk %d is %d\n", chunk, offset_bytes);
	
	return chunk_bytes;
}

static oqt_int32_t oqt_sd_id_of_chunk(
	oqt_t *file,
	oqt_trak_t *trak,
	oqt_int32_t chunk)
{
	oqt_stsc_table_t *stsc_table = trak->mdia.minf.stbl.stsc.table;
	int stsc_entries = trak->mdia.minf.stbl.stsc.total_entries;
	int first_sample=0;
	int i, entry;


	for(entry=0,i=1;i<=chunk;i++) {
		if ((entry+1)<stsc_entries && stsc_table[entry+1].chunk==i) entry++;
		if (i==chunk) {
			return stsc_table[entry].id;
		}
		first_sample+=stsc_table[entry].samples;
		if (entry>=stsc_entries) {
			fprintf(stderr, "oqt_sd_id_of_chunk: ran out of stsc entries (%d).\n", entry);
			break;
		}
	}

	// failed for whatever reason
	// be on the safe side and just say 1 (should always be there)
	return 1;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_audio_position_to_chunk(oqt_t *file, int track, oqt_int32_t chunk)
{
	oqt_audio_map_t *atrack = &(file->atracks[track]);
	oqt_int32_t sd_id = oqt_sd_id_of_chunk(file, atrack->track, chunk);
	oqt_int64_t offset = oqt_chunk_to_offset(atrack->track, chunk);
	file->data_file = oqt_file_from_sd_id(file, atrack->track, sd_id);
	return oqt_set_position(file->data_file, offset);
}

OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_audio_chunk_samples(oqt_t *file, int track, oqt_int32_t chunk)
{
	oqt_audio_map_t *atrack = &(file->atracks[track]);
	return oqt_chunk_samples(atrack->track, chunk);
}


OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_add_audio_track(oqt_t *file, 
								int channels,
								oqt_int32_t sample_rate,
								int bits,
								char *compressor)
{
	oqt_trak_t *trak;
	
	if (!file->oqt_init_audio_map) {
		perror("Error: oqt_init_audio_map is not available.\n");
		return -1;
	}
	
	file->atracks = (oqt_audio_map_t*)realloc(file->atracks,
					sizeof(oqt_audio_map_t)*(file->total_atracks+1));
	if (!file->atracks) {
		perror("realloc failed for atracks in oqt_add_audio_track\n");
		return -1;
	}
		
	// Add a track to the file
	trak = oqt_add_trak(&(file->moov));
	if (!trak) {
		perror("oqt_add_trak failed in oqt_add_video_track\n");
		return -1;
	}
	
	// Increment the number of audio tracks
	file->total_atracks++;

	oqt_trak_init_audio(file, trak, channels, sample_rate, bits, compressor);
	file->oqt_init_audio_map(file, file->total_atracks-1, trak);
	
	return file->total_atracks-1;   // Return the number of the track created
}


OPENQUICKTIMELIB_API oqt_int32_t
OQT_STDCALL oqt_add_video_track(oqt_t *file, 
								int frame_w, 
								int frame_h,
								oqt_float_t frame_rate,
								char *compressor)
{
	oqt_trak_t *trak;
	
	if (!file->oqt_init_video_map) {
		perror("Error: oqt_init_video_map is not available.\n");
		return -1;
	}

	// *** Check to see this has already been inited ? ***
	oqt_mhvd_init_video(file, &(file->moov.mvhd), frame_rate);
	
	file->vtracks = (oqt_video_map_t*)realloc(file->vtracks,
					sizeof(oqt_video_map_t)*(file->total_vtracks+1));
	if (!file->vtracks) {
		perror("realloc failed for vtracks in oqt_add_video_track\n");
		return -1;
	}

	// Add a track to the file
	trak = oqt_add_trak(&(file->moov));
	if (!trak) {
		perror("oqt_add_trak failed in oqt_add_video_track\n");
		return -1;
	}

	// Increment the number of video tracks
	file->total_vtracks++;

	oqt_trak_init_video(file, trak, frame_w, frame_h, frame_rate, compressor);
	file->oqt_init_video_map(file, file->total_vtracks-1, trak);
	
	return file->total_vtracks-1;   // Return the number of the track created
}




OPENQUICKTIMELIB_API oqt_int32_t 
OQT_STDCALL oqt_get_audio_frames_to_bytes(
	oqt_t *file,
	int track,
	int frames)
{
	int channels = oqt_get_audio_channels(file, track);
	int bytesPerPacket = file->atracks[track].track->mdia.minf.stbl.stsd.table[0].bytesPerPacket;
	int samplesPerPacket = file->atracks[track].track->mdia.minf.stbl.stsd.table[0].samplesPerPacket;

#ifndef WIN32
	int bytes=file->atracks[track].codec->frames_to_bytes(file, track, frames);
#else
	/* FIXME: */
	int bytes=0;
#endif
	//fprintf(stderr, "oqt_get_audio_frames_to_bytes: codec=%d bytesPerPacket=%d samplesPerPacket=%d frames=%d channels=%d\n",
	//		 bytes, bytesPerPacket, samplesPerPacket, frames, channels);
	
	// Try using the value from codecs function first
	if (bytes>0) return bytes;
	
	// Now try calculating it from the stsd
	if (bytesPerPacket && samplesPerPacket)
		return ((oqt_float_t)frames/samplesPerPacket)*bytesPerPacket*channels;
		
	// Erm, can't calculate it - return an error
	return -1;
}



/* return -1 if there is NO keyframe before */
OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_video_keyframe_before(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_trak_t *trak = file->vtracks[track].track;
	oqt_stss_t *stss = &trak->mdia.minf.stbl.stss;
	int lo, hi;
	
	/* per spec: if this table is absent, every sample is a sync sample */
	if (!stss->table) return frame;
	
	lo = 0;
	hi = stss->total_entries-1;
	if (stss->table[lo].sample-1> frame) return -1;
	if (stss->table[hi].sample-1<=frame) return stss->table[hi].sample-1;
	while (hi>lo+1) {
	  /* here: stss->table[lo].sample-1<=frame
	           stss->table[hi].sample-1> frame */	       
	  int med = (lo+hi)/2;
	  if (stss->table[med].sample-1<=frame) lo = med; else hi = med;
	}
	/* here: hi=lo+1 */
	return stss->table[lo].sample-1;
}

/* return -1 if there is NO keyframe after */
OPENQUICKTIMELIB_API oqt_int64_t
OQT_STDCALL oqt_get_video_keyframe_after(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_trak_t *trak = file->vtracks[track].track;
	oqt_stss_t *stss = &trak->mdia.minf.stbl.stss;
	int lo, hi;

	/* per spec: if this table is absent, every sample is a sync sample */
	if (!stss->table) return frame;
	
	lo = 0;
	hi = stss->total_entries-1;
	if (stss->table[lo].sample-1>=frame) return stss->table[lo].sample-1;
	if (stss->table[hi].sample-1< frame) return -1;
	while (hi>lo+1) {
	  /* here: stss->table[lo].sample-1<frame
	           stss->table[hi].sample-1>=frame */	       
	  int med = (lo+hi)/2;
	  if (stss->table[med].sample-1<frame) lo = med; else hi = med;
	}
	/* here: hi=lo+1 */
	return stss->table[hi].sample-1;
}
	

OPENQUICKTIMELIB_API void
OQT_STDCALL oqt_insert_video_keyframe(oqt_t *file, int track, oqt_int64_t frame)
{
	oqt_trak_t *trak = file->vtracks[track].track;
	oqt_stss_t *stss = &trak->mdia.minf.stbl.stss;
	int i;

	//printf("Total Entries %lu stss->entries_allocated %lu\n", stss->total_entries, stss->entries_allocated);

// Get the keyframe greater or equal to new frame
        for(i = 0; i < stss->total_entries; i++)
        {
			if(stss->table[i].sample >= frame) break;
        }

// Expand table
	if (stss->entries_allocated <= stss->total_entries)
	{
		stss->entries_allocated = stss->entries_allocated*2+1;
		stss->table = realloc(stss->table, sizeof(oqt_stss_table_t) * stss->entries_allocated);
	}

// Insert before existing frame
	if(i < stss->total_entries)
	{
		if(stss->table[i].sample > frame)
		{
			int j, k;
			for(j = stss->total_entries, k = stss->total_entries - 1;
				k >= i;
				j--, k--)
			{
				stss->table[j] = stss->table[k];
			}
			stss->table[i].sample = frame;
		}
	}
	else
// Insert after last frame
		stss->table[i].sample = frame;
  stss->total_entries++;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_mp4_video_decoder_config(oqt_t *file, int track, u_char** ppBuf, int* pBufSize)
{
	oqt_esds_t* esds;

	if (!file->total_vtracks) {
		return 0;
	}
	esds = &file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].esds;
	return oqt_esds_get_decoder_config(esds, ppBuf, pBufSize);
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_mp4_video_decoder_config(oqt_t *file, int track, u_char* pBuf, int bufSize)
{
	oqt_esds_t* esds;

	if (!file->total_vtracks) {
		return 0;
	}
	esds = &file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].esds;
	return oqt_esds_set_decoder_config(esds, pBuf, bufSize);
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_mp4_audio_decoder_config(oqt_t *file, int track, u_char** ppBuf, int* pBufSize)
{
	oqt_esds_t* esds;

	if (!file->total_atracks) {
		return 0;
	}
	esds = &file->atracks[track].track->mdia.minf.stbl.stsd.table[0].esds;
	return oqt_esds_get_decoder_config(esds, ppBuf, pBufSize);
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_mp4_audio_decoder_config(oqt_t *file, int track, u_char* pBuf, int bufSize)
{
	oqt_esds_t* esds;

	if (!file->total_atracks) {
		return 0;
	}
	esds = &file->atracks[track].track->mdia.minf.stbl.stsd.table[0].esds;
	return oqt_esds_set_decoder_config(esds, pBuf, bufSize);
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_iod_audio_profile_level(oqt_t *file)
{
	return file->moov.iods.audioProfileId;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_iod_audio_profile_level(oqt_t *file, int id)
{
	file->use_mp4 = 1;
	return oqt_iods_set_audio_profile(&file->moov.iods, id);
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_get_iod_video_profile_level(oqt_t *file)
{
	return file->moov.iods.videoProfileId;
}

OPENQUICKTIMELIB_API int
OQT_STDCALL oqt_set_iod_video_profile_level(oqt_t *file, int id)
{
	file->use_mp4 = 1;
	return oqt_iods_set_video_profile(&file->moov.iods, id);
}
