/* codec_cinepak.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_cinepak.c,v 1.8 2002/12/30 18:32:35 antoinemine Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "cinepak.h"


#ifdef __cplusplus
extern "C" {
#endif


/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "cvid"; 
  /* Codec Signature */
  
static const char   *Name = "Radius Cinepak"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *License  = "GNU General Public License"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Dr Tim Ferguson"; 
static const char   *Description  =
"The Cinepak codec is a relatively old coding technique that is still infrequently used"
" today. Its advantage comes from computational simplicity at the decoder, rather than "
"bit rate versus quality performance. This codec is basically a vector quantiser with "
"adaptive vector density. Each frame is segmented into 4x4 pixel blocks, and each block "
"is coded using either 1 or 4 vectors. For more information visit the codec authors "
"homepage here: http://www.csse.monash.edu.au/~timf/videocodec.html";


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	void *context;
	char *buffer;
} Param;  


static int init_cvid(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));

	/* Initalise the Codec */
	p->context = decode_cinepak_init();
	p->buffer = NULL;
	
	return ++Initialised;
}



static int delete_cvid(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		if (p->context)	free(p->context);
		if (p->buffer)	free(p->buffer);
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}



/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_cvid(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int depth   = oqt_get_video_depth(file, track);
	int i;
	
    if(!p->buffer)
		p->buffer = malloc(width * height * 4);

	if (color_model == BC_NONE) {
#ifdef DEBUG
		fprintf(stderr, "decode_cvid: dropping frame.\n");
#endif
	} else if (color_model == BC_BGR888) {
		decode_cinepak(p->context, input, inputsize,
			 		   p->buffer, width, height, 24, width*4);
		for(i = 0; i < height; i++)
		{
			memcpy(output[i], &(p->buffer[i*width*4]), width*3);
		}
    }
    else if(color_model == BC_BGR8888) {
		decode_cinepak(p->context, input, inputsize,
			 			p->buffer, width, height, 32, width*4);
		for(i = 0; i < height; i++)
		{
			memcpy(output[i], &(p->buffer[i*width*4]), width*4);
		}
    } else {
    	fprintf(stderr, "decode_cvid: decoding of colourmodel=%d not supported.\n", 
    					color_model);
    }
	
	return 0;
}


static int reads_colormodel_cvid(oqt_t *file, 
		int track,
		int colormodel)
{
	return ((colormodel == BC_BGR888) ||
	        (colormodel == BC_BGR8888));
}

static int preferred_colormodel_cvid(oqt_t *file, 
		int track)
{
	return BC_BGR8888;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;
	
	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.description = Description;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.license = License;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_cvid;
	//codec->encode = encode_cvid;
	codec->decode = decode_cvid;
	codec->delete_codec = delete_cvid;
	codec->reads_colormodel = reads_colormodel_cvid;
	//codec->writes_colormodel = writes_colormodel_cvid;
	codec->preferred_colormodel = preferred_colormodel_cvid;
	
	return &codec->info;
}

