/* codec_dvc.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_dvc.c,v 1.9 2003/05/15 13:28:17 nhumfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libdv/dv.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int Initialised   = 0;      
static const char Signature_NTSC[5]  = "dvc "; 
static const char Signature_PAL[5]  = "dvcp"; 
static const char *Name_NTSC = "Quasar DV video (NTSC)"; 
static const char *Name_PAL = "Quasar DV video (PAL)"; 
static const char *Version  = "0.1"; 
static const char *Copyright  = "2002 OpenQuicktime Team"; 
static const char *ModuleAuthor  = "Jozef Hatala"; 
static const char *CodecAuthor  = "Charles Krasic and Erik Walthinsen"; 
  
typedef struct {

  dv_decoder_t *dec_handle;
  dv_encoder_t *enc_handle;
  oqt_uint8_t *yuy2;
  int yuy2_size;
} Param;  

static int reads_colormodel_dvc(oqt_t *file, 
				int track,
				int colormodel)
{
  return ( colormodel == BC_YUV422P || colormodel == BC_YUV420P );
}

static int preferred_colormodel_dvc(oqt_t *file, 
				    int track)
{
  return BC_YUV422P;
}



static int decode_dvc(oqt_t *file,
		      int track,
		      oqt_uint32_t inputsize,
		      oqt_uint8_t *input,
		      oqt_uint8_t **output,
		      int color_model)
{
  int res, i;
  Param *p = (Param*)(file->vtracks[track].codec_private);
  int width 	= oqt_get_video_width(file, track);
  int height 	= oqt_get_video_height(file, track);

  if(color_model == BC_NONE) {
    return 0;
  } else if(color_model != BC_YUV422P && color_model != BC_YUV420P) {
    fprintf(stderr, "decode_dvc: decoding of colourmodel=%d not supported.\n", 
	    color_model);
    return 1;
  }

  if(!p->dec_handle) {
    p->dec_handle = dv_decoder_new(FALSE, FALSE, FALSE);
    if(!p->dec_handle) {
      fprintf(stderr, "decode_dvc: dv_decoder_new failed\n");
      return 1;
    }
    p->dec_handle->quality = DV_QUALITY_BEST;
  }
	
  res = dv_parse_header(p->dec_handle, input);
  if(res < 0) {
    fprintf(stderr, "decode_dvc: dv_parse_header failed\n");
    return 1;
  }

  if(inputsize != p->dec_handle->frame_size) {
    fprintf(stderr, "decode_dvc: unexpected frame size %d, should be %d\n",
	    inputsize, (int)p->dec_handle->frame_size);
    return 1;
  }

  if(2 * p->dec_handle->width * p->dec_handle->height > p->yuy2_size) {
    if(p->yuy2) {
      free(p->yuy2);
    }
    p->yuy2_size = 2 * p->dec_handle->width * p->dec_handle->height;
    p->yuy2 = malloc(p->yuy2_size);
  }

  if(2*(width/2) != width) {
    fprintf(stderr, "decode_dvc: this will not work with _odd_ width %d\n", width);
    return 1;
  }

  dv_parse_packs(p->dec_handle, input);

  if (!p->dec_handle->prev_frame_decoded ||
      dv_frame_changed(p->dec_handle)) {

    int pitches[3] = {2*p->dec_handle->width, 0, 0};
    oqt_uint8_t *pixels[3] = {p->yuy2, NULL, NULL};

    dv_decode_full_frame(p->dec_handle, input,
                         e_dv_color_yuv,
                         pixels,
                         pitches);
    p->dec_handle->prev_frame_decoded = 1;

  } else {
    /* nothing, just repeat last frame */
  }

  // q4l produces movies with non-dv dimensions
  if(width <= p->dec_handle->width && height <= p->dec_handle->height) {
    oqt_uint8_t *y_ptr = output[0];
    oqt_uint8_t *u_ptr = output[1];
    oqt_uint8_t *v_ptr = output[2];
    int skip_pixels = p->dec_handle->width - width;
    oqt_uint8_t *p_ptr = p->yuy2;
    int rows, parity;
    oqt_uint8_t u_row[width/2], v_row[width/2];

    for(rows = height, parity = 0; rows > 0; --rows, parity = !parity) {
      if(color_model == BC_YUV422P) {
	for(i = width/2; i > 0; --i) {
	  *(y_ptr++) = *(p_ptr++);
	  *(u_ptr++) = *(p_ptr++);
	  *(y_ptr++) = *(p_ptr++);
	  *(v_ptr++) = *(p_ptr++);
	}

      } else if(color_model == BC_YUV420P) {
	oqt_uint8_t *ur_ptr = u_row;
	oqt_uint8_t *vr_ptr = v_row;

	for(i = width/2; i > 0; --i) {
	  oqt_uint16_t u, v;

	  *(y_ptr++) = *(p_ptr++);
	  u = *(p_ptr++);
	  *(y_ptr++) = *(p_ptr++);
	  v = *(p_ptr++);

	  // average between pair and impair lines
	  if(parity == 0) {
	    *(ur_ptr++) = u;
	    *(vr_ptr++) = v;
	  } else {
	    *(u_ptr++) = (u + *(ur_ptr++))/2;
	    *(v_ptr++) = (v + *(vr_ptr++))/2;
	  }
	}
      }

      p_ptr += 2*skip_pixels;
    }

  } else {
    fprintf(stderr, "decode_dvc: %dx%d is larger that stream %dx%d\n",
	    width, height, p->dec_handle->width, p->dec_handle->height);
    return 1;
  }

  return 0;
}


static int writes_colormodel_dvc(oqt_t *file, 
				 int track,
				 int colormodel)
{
  return ( colormodel == BC_YUV422P || colormodel == BC_YUV420P );
}


static int encode_dvc(oqt_t *file, 
		      int track,
		      oqt_uint8_t **input,
		      oqt_uint32_t out_size,
		      oqt_uint8_t *output,
		      int *IsAKeyFrame,
		      int color_model)
{
  int xerr;
  Param *p = (Param*)(file->vtracks[track].codec_private);
  int width = file->vtracks[track].track->tkhd.track_width;
  int height = file->vtracks[track].track->tkhd.track_height;
  oqt_uint8_t *pixels[3] = {NULL, NULL, NULL};
  int isPAL;

  if(color_model != BC_YUV422P && color_model != BC_YUV420P) {
    fprintf(stderr,"encode_dvc: unsupported color model\n");
    return 0;
  }

  if(width == 720 && height == 576) {
    isPAL = 1;
  } else if(width == 720 && height == 480) {
    isPAL = 0;
  } else {
    fprintf(stderr,"encode_dvc: unsupported size: %dx%d\n", width, height);
    return 0;
  }

  if((isPAL && out_size < 144000) || 
     (!isPAL && out_size < 120000)) {
    fprintf(stderr,"encode_dvc: output buffer too small: %d\n", out_size);
    return 0;
  }

  if(!p->enc_handle) {
    p->enc_handle = dv_encoder_new(FALSE, FALSE, FALSE);
    if(!p->enc_handle) {
      fprintf(stderr, "encode_dvc: dv_encoder_new failed\n");
      return 1;
    }
    /* Hmmm... */
    p->enc_handle->vlc_encode_passes = 3;
    p->enc_handle->static_qno = 0;
    p->enc_handle->force_dct = DV_DCT_AUTO;
    p->enc_handle->is16x9 = 0;
    p->enc_handle->isPAL = isPAL;
  }

  if(2 * width * height > p->yuy2_size) {
    if(p->yuy2) {
      free(p->yuy2);
    }
    p->yuy2_size = 2 * width * height;
    p->yuy2 = malloc(p->yuy2_size);
  }

  pixels[0] = p->yuy2;

  if(color_model == BC_YUV422P) {
    oqt_uint8_t *y_ptr = input[0];
    oqt_uint8_t *u_ptr = input[1];
    oqt_uint8_t *v_ptr = input[2];
    oqt_uint8_t *p_ptr = p->yuy2;
    int i;

    for(i = height*width/2; i > 0; --i) {
      *(p_ptr++) = *(y_ptr++);
      *(p_ptr++) = *(u_ptr++);
      *(p_ptr++) = *(y_ptr++);
      *(p_ptr++) = *(v_ptr++);
    }
    
  } else if(color_model == BC_YUV420P) {
    oqt_uint8_t *y_ptr = input[0];
    oqt_uint8_t *u_ptr = input[1];
    oqt_uint8_t *v_ptr = input[2];
    oqt_uint8_t *p_ptr = p->yuy2;
    int rows, i, parity;

    for(rows = height, parity=0; rows > 0; --rows, parity=!parity) {
      for(i = width/2; i > 0; --i) {
	*(p_ptr++) = *(y_ptr++);
	*(p_ptr++) = *(u_ptr++);
	*(p_ptr++) = *(y_ptr++);
	*(p_ptr++) = *(v_ptr++);	
      }
      // same U and V on pair and impair lines
      if(parity == 0) {
	u_ptr -= width/2;
	v_ptr -= width/2;
      }
    }
  }

  dv_encode_full_frame(p->enc_handle, pixels, e_dv_color_yuv, output);

  *IsAKeyFrame = 1;
  return isPAL ? 144000 : 120000;
}


static int init_dvc(oqt_t *file, int track)
{
  Param *p;
  oqt_video_map_t *vtrack = &file->vtracks[track];
  int width = oqt_get_video_width(file, track);
  int height = oqt_get_video_height(file, track);
		
  vtrack->codec_private = malloc(sizeof(Param));
  p = (Param*)vtrack->codec_private;	

  p->dec_handle = NULL;
  p->yuy2 = NULL;
  p->yuy2_size = 0;

  p->enc_handle = NULL;

  Initialised++;
  return Initialised;
}


static int delete_dvc(oqt_t *file, int track)
{
  oqt_video_map_t *vtrack = &file->vtracks[track];
  Param *p = (Param*)(vtrack->codec_private);

  if(p) {
    if(p->dec_handle)
      dv_decoder_free(p->dec_handle);
    if(p->enc_handle)
      dv_encoder_free(p->enc_handle);
    if(p->yuy2)
      free(p->yuy2);
    p->yuy2_size = 0;
    free(p);
    vtrack->codec_private=NULL;
  }

  return --Initialised;
}

static void fill_codec_struct(oqt_video_codec_t *codec)
{
  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;
	
  codec->init_codec = init_dvc;
  codec->decode = decode_dvc;
  codec->reads_colormodel = reads_colormodel_dvc;
  codec->preferred_colormodel = preferred_colormodel_dvc;
  codec->delete_codec = delete_dvc;

  codec->encode = encode_dvc;
  codec->writes_colormodel = writes_colormodel_dvc;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
  oqt_video_codec_t *codec;
  oqt_video_codec_t *alt;

  // Allocate the memory for the codec record
  codec = oqt_allocate_video_codec();
  if(!codec) return NULL;
  alt = oqt_allocate_video_codec();
  if(!alt) return NULL;

  dv_init(FALSE, FALSE);

  // Information about the codec
  codec->info.name = Name_NTSC;
  memcpy( &codec->info.fourcc, Signature_NTSC, 4); 
  fill_codec_struct(codec);
  
  alt->info.name = Name_PAL;
  memcpy( &alt->info.fourcc, Signature_PAL, 4); 
  fill_codec_struct(alt);

  // Set parentage
  alt->info.parent = &codec->info;
  codec->info.has_children = 1;

  if (oqt_register_codec(&alt->info)) {
    fprintf(stderr, "Failed to register alternate video codec for dvc.\n");
    return NULL;
  }

  return &codec->info;
}
