/* $Id: codec_mjpeg.c,v 1.2 2003/05/19 11:35:05 nhumfrey Exp $

   Motion-JPEG A codec
   Ciaran Wills <c.j.wills@acm.org>, Framestore-CFC, May-03

   Based on the photo_jpeg codec, not on the qt4linux MJPEG codec

   Motion JPEG splits each frame into two interlaced fields and
   encodes each as a separate JPEG image.  The order of the field
   dominance is controled by the filedDominance parameter.

   libjpeg is used to do the compression (as it is in the photo_jpeg
   and qt4linux codecs.  However the Quicktime spec orders the JPEG
   headers differently from the JFIF standard, and this appears to
   upset some hardware decoders (such as the Aurora Igniter).
   Therefore we mangle the JPEG headers after compression to make it
   look like 'proper' MJPEG-A Quicktime.  This is controlled by the
   fixHeaders parameter, and is on by default.

   TODO:
   I'm sure it could be more robust
   Base a Motion-JPEG B codec on this
   Color models other than RGB888
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <jpeglib.h>

// Fix for compiler warning
#ifdef HAVE_STDLIB_H
#undef HAVE_STDLIB_H
#endif

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"
#include "openquicktime/bswap.h"


static int Initialised   = 0;      
static const char Signature[5]  = "mjpa"; 
static const char *Name = "Motion-JPEG A"; 
static const char *Version  = "0.1"; 
static const char *Copyright  = "2003 OpenQuicktime Team"; 
static const char *ModuleAuthor  = "Ciaran Wills <c.j.wills@acm.org>"; 
static const char *CodecAuthor  = "libjpeg - Thomas G. Lane"; 

#define JPEG_DEFAULT_QUALITY 85
#define MJPA_DEFAULT_DOMINANCE 1
#define PARAM_COUNT 3

static unsigned char comment[] = {'A', 'p', 'p', 'l', 'e', 'M', 'a', 'r', 'k', 0x0a};
static unsigned int app1[10] = {0, 0x67706a6d, 0, 0, 0, 0, 0, 0, 0, 0};

static const oqt_codec_param_t ListParams[PARAM_COUNT] = {
  { "quality", OQT_PARAMETER_INTEGER, "Quality of JPEG Compression [0-100]"},
  { "fieldDominance", OQT_PARAMETER_INTEGER, "Field dominance [1-2]"},
  { "fixHeaders", OQT_PARAMETER_INTEGER, "Fix JPEG headers [0-1]"}
};

typedef struct {
  /* Error structure */
  struct my_error_mgr {
    struct jpeg_error_mgr pub;
    jmp_buf setjmp_buffer;
  } my_jerr;
  
  int quality;
  int dominance;
  int fixheaders;
  
  //int nativeOrder; // 0 - big endian, 1 - little endian
  
  /* decompression structure */
  struct jpeg_decompress_struct decinfo;
  struct jpeg_source_mgr jsrc;
  
  /* compression structure */
  struct jpeg_compress_struct encinfo;
  struct jpeg_destination_mgr jdest;
  
} Param;  


static const oqt_codec_param_t* list_params_MJPA(int *o_numparams) {
  *o_numparams = PARAM_COUNT;
  return ListParams;
}

static int set_param_MJPA(oqt_t* file, int track, const char* param,
  const void* data) {
  Param *p = (Param*)(file->vtracks[track].codec_private);
  
  if (!strcmp(param,"quality")) {
    p->quality = *(( int*)data);
    jpeg_set_quality(&p->encinfo, p->quality, TRUE);
    
    return 0; 
  }
  if (!strcmp(param,"fieldDominance")) {
    p->dominance = *(( int*)data);
    
    return 0; 
  }
  if (!strcmp(param,"fixHeaders")) {
    p->fixheaders = *(( int*)data);
    
    return 0; 
  }
  
  fprintf(stderr,"set_param_MJPA: unknown parameter named '%s'\n",param);
  return OQT_PARAMETER_UNKNOWN;
}

static int get_param_MJPA(oqt_t* file, int track, const char* param,
  void* data) {
  Param *p = (Param*)(file->vtracks[track].codec_private);
  
  if (!strcmp(param,"quality")) {
    *((int*)data) = p->quality;
    return OQT_PARAMETER_INTEGER;
  }
  if (!strcmp(param,"fieldDominance")) {
    *((int*)data) = p->dominance;
    return OQT_PARAMETER_INTEGER;
  }
  if (!strcmp(param,"fixHeaders")) {
    *((int*)data) = p->fixheaders;
    return OQT_PARAMETER_INTEGER;
  }
  
  fprintf(stderr,"set_param_MJPA: unknown parameter named '%s'\n",param);
  return OQT_PARAMETER_UNKNOWN;
}

static void jpeg_oqt_error_exit(j_common_ptr cinfo) {
  struct my_error_mgr * jerr = (struct my_error_mgr *) cinfo->err;
  fprintf(stderr, "jpeg_oqt_error_exit: got error.\n");
  
  /* Display the error message */
  (*cinfo->err->output_message) (cinfo);
  
  /* Abort the compression/decompression so we can resuse */
  jpeg_abort(cinfo);
  
  /* Go to jump point near start of encode/decode */	
  longjmp(jerr->setjmp_buffer, 1);
}

/* Verify that the format of the frame is correct */
static inline int verify_jpeg_format( oqt_t *file, int track, Param *p ) {
  
  /* Check width */
  if (p->decinfo.output_width != oqt_get_video_width(file, track)) {
    fprintf(stderr, "verify_jpeg_format: error, width of jpeg frame is not same as track.\n");
    jpeg_abort_decompress(&p->decinfo);
    return 1;
  }
  
  /* Check height */
  if (p->decinfo.output_height != oqt_get_video_height(file, track)) {
    fprintf(stderr, "verify_jpeg_format: error, height of jpeg frame is not same as track.\n");
    jpeg_abort_decompress(&p->decinfo);
    return 1;
  }
  
  /* Check the comonents */
  if (p->decinfo.output_components != 3) {
    fprintf(stderr, "verify_jpeg_format: error, number of colour components is not 3.\n");
    jpeg_abort_decompress(&p->decinfo);
    return 1;
  }
  
  /* Check component size */
  if (BITS_IN_JSAMPLE!=8) {
    fprintf(stderr, "verify_jpeg_format: error, bits per JSample is not 8.\n");
    jpeg_abort_decompress(&p->decinfo);
    return 1;
  }
    
  /* Success */
  return 0;
}

// Decode a frame
// Find the two JPEG encoded fields and pass them to libjpeg
static int decode_MJPA(oqt_t *file, int track, oqt_uint32_t inputsize,
  oqt_uint8_t  *input, oqt_uint8_t  **output, int color_model) {

  // Find the APP1 header
  // This will go wrong if there is no APP1 header
  int i = 0, nextfield, d[2] = {0, 1};
  Param *p;

  while (!(input[i] == 0xff && input[i+1] == 0xe1))
    i++;
  nextfield = *(unsigned int*)(input+i+20);

  #ifndef WORDS_BIGENDIAN
  nextfield = bswap_32(nextfield);
  #endif
  
  p = (Param*)(file->vtracks[track].codec_private);
  if(setjmp(p->my_jerr.setjmp_buffer)) {
    // exception: we longjmp'ed here
    fprintf(stderr,
      "decode_MJPA: caught error - failed to decode JPEG stream\n");
    return 1;
  }
  p->my_jerr.pub.error_exit = jpeg_oqt_error_exit;
  
  if (file->moov.trak[track]->mdia.minf.stbl.stsd.table->field_dominance % 2 == 0) {
    d[0] = 1; d[1] = 0;
  }

  p->jsrc.next_input_byte = input;
  p->jsrc.bytes_in_buffer = nextfield;
  jpeg_read_header(&p->decinfo, TRUE);
  
  p->decinfo.do_fancy_upsampling = FALSE;
  p->decinfo.do_block_smoothing	= FALSE;
  p->decinfo.dct_method = JDCT_IFAST;
  p->decinfo.dither_mode = JDITHER_ORDERED;

  p->decinfo.raw_data_out = FALSE;
  p->decinfo.out_color_space = JCS_RGB;
  
  jpeg_start_decompress(&p->decinfo);
  for (i = 0; i < p->decinfo.output_height; i++)
    jpeg_read_scanlines( &p->decinfo, &output[i*2+d[0]], 1);
  jpeg_finish_decompress(&p->decinfo);

  p->jsrc.next_input_byte = input + nextfield;
  p->jsrc.bytes_in_buffer = inputsize - nextfield;
  jpeg_read_header(&p->decinfo, TRUE);
  
  jpeg_start_decompress(&p->decinfo);
  for (i = 0; i < p->decinfo.output_height; i++)
    jpeg_read_scanlines( &p->decinfo, &output[i*2+d[1]], 1);
  jpeg_finish_decompress(&p->decinfo);
  
  return 0;
}

// Non-spec M-JPEG A
// Just leave the APP1 marker empty - many players can cope with this anyway
static int fixApp1(unsigned char *p, int field1) {
  int i = 0;

  // Find the end marker
  while (!(p[i] == 0xff && p[i+1] == 0xd9))
    i++;

  return i;
}

// Reorder the headers as per the Quicktime spec
// The JFIF data generated by libjpeg has two DQT and 4 DHT markers,
// Here we merge these, reorder all the markers, and fill in the
// offsets in the APP1 marker
static int fixheaders(unsigned char *p, int field1, const Param *param) {
  unsigned char sof[32], dqt[256], dht1[1024], dht2[1024];
  int sofsize, dqtsize, dhtsize1, dhtsize2;
  int offsets[8];
  int i = 0, j, size, size2, a, pad;
  unsigned int *p2;

  // Find APP1 marker for later
  while (!(p[i] == 0xff && p[i+1] == 0xe1))
    i++;
  offsets[0] = i;

  // Find first DQT markers
  while (!(p[i] == 0xff && p[i+1] == 0xdb))
    i++;
  offsets[1] = i;

  // DQT
  i += 2;
  size = (p[i] << 8) | p[i+1];
  memcpy(dqt, p+i+2, size-2);
  i += size+2;
  size2 = (p[i] << 8) | p[i+1];
  memcpy(dqt+size-2, p+i+2, size2-2);
  dqtsize = size+size2-4;
  i += size2;

  // SOF
  i += 2;
  size = (p[i] << 8) | p[i+1];
  memcpy(sof, p+i+2, size-2);
  sofsize = size-2;
  i += size;

  // DHT (4 of these)
  i += 2;
  dhtsize1 = 0; dhtsize2 = 0;

  size = (p[i] << 8) | p[i+1];
  memcpy(dht1 + dhtsize1, p+i+2, size-2);
  i += size+2;
  dhtsize1 += size-2;

  size = (p[i] << 8) | p[i+1];
  memcpy(dht2 + dhtsize2, p+i+2, size-2);
  i += size+2;
  dhtsize2 += size-2;

  size = (p[i] << 8) | p[i+1];
  memcpy(dht1 + dhtsize1, p+i+2, size-2);
  i += size+2;
  dhtsize1 += size-2;

  size = (p[i] << 8) | p[i+1];
  memcpy(dht2 + dhtsize2, p+i+2, size-2);
  i += size+2;
  dhtsize2 += size-2;
  i -= 2;
  offsets[2] = i;

  // Spit them back out
  i = offsets[1];

  // DQT
  p[i++] = 0xff; p[i++] = 0xdb;
  p[i++] = (dqtsize+2) >> 8;
  p[i++] = (dqtsize+2) & 0xff;
  memcpy(p+i, dqt, dqtsize);
  i += dqtsize;

  // DHT
  size = dhtsize1 + dhtsize2;
  offsets[4] = i;
  p[i++] = 0xff; p[i++] = 0xc4;
  p[i++] = (size+2) >> 8;
  p[i++] = (size+2) & 0xff;
  memcpy(p+i, dht1, dhtsize1);
  memcpy(p+i+dhtsize1, dht2, dhtsize2);
  i += size;

  // SOF
  offsets[5] = i;
  p[i++] = 0xff; p[i++] = 0xc0;
  p[i++] = (sofsize+2) >> 8;
  p[i++] = (sofsize+2) & 0xff;
  memcpy(p+i, sof, sofsize);
  i += sofsize;
  offsets[3] = i;

  // Find the end of the data
  i = offsets[2];
  while (!(p[i] == 0xff && p[i+1] == 0xd9))
    i++;
  size = i - offsets[2] + 2;

  memmove(p+offsets[3], p+offsets[2], size);
  size = size + offsets[3];

  // Pad to a 4 byte boundary
  a = size % 4;
  size2 = size;
  pad = 0;
  if (a) {
    pad = 4 - a;
    for (i = 0; i < pad; i++)
      p[size+i] = 0xff;
    size2 = size + pad;
  }

  // Put the offsets in the APP1 header
  p2 = (unsigned int*)(p + offsets[0]);

  
  #ifdef WORDS_BIGENDIAN

    p2[3] = size;
    p2[4] = size2;
    if (field1)
      p2[5] = p2[4];
    else
      p2[5] = 0;
    p2[6] = offsets[1];
    p2[7] = offsets[4];
    p2[8] = offsets[5];
    p2[9] = 0x028b;
    p2[10] = 0x0299;
    
  #else
  
	  p2[3] = bswap_32(size);
	  p2[4] = bswap_32(size2);
	  if (field1)
		p2[5] = p2[4];
	  else
		p2[5] = 0;
	  p2[6] = bswap_32(offsets[1]);
	  p2[7] = bswap_32(offsets[4]);
	  p2[8] = bswap_32(offsets[5]);
	  p2[9] = bswap_32(0x028b);
	  p2[10] = bswap_32(0x0299);

  #endif  

  return size2;
}

static int encode_MJPA(oqt_t *file, int track, oqt_uint8_t **input,
  oqt_uint32_t out_size, oqt_uint8_t *output, int *IsAKeyFrame,
  int color_model) {

  oqt_video_map_t *vtrack = &(file->vtracks[track]);
  Param *p = (Param*)(vtrack->codec_private);
  int width 	= oqt_get_video_width(file, track);
  int height 	= oqt_get_video_height(file, track);
  int i, used, used2;

  int d[2] = {0, 1};
  if (p->dominance % 2 == 0) {
    d[0] = 1; d[1] = 0;
  }

  if(setjmp(p->my_jerr.setjmp_buffer)) {
    // exception: we longjmp'ed here
    fprintf(stderr, "encode_MJPA: caught error - libjpeg failed.\n");
    return 0;
  }
  p->my_jerr.pub.error_exit = jpeg_oqt_error_exit;

  p->encinfo.in_color_space = JCS_RGB;  

  p->jdest.next_output_byte = output;
  p->jdest.free_in_buffer = out_size;	
  
  p->encinfo.image_width = width;
  p->encinfo.image_height = height/2;
  p->encinfo.input_components = 3;	/* RGB/YUV */
  jpeg_set_defaults(&p->encinfo);
  p->encinfo.dct_method = JDCT_FASTEST;
  jpeg_set_quality(&p->encinfo, p->quality, TRUE);
  p->encinfo.comp_info[0].v_samp_factor = 1;

  jpeg_start_compress(&p->encinfo, TRUE);
  if (p->fixheaders)
    jpeg_write_marker(&p->encinfo, JPEG_COM, (JOCTET *)comment, 10);
  jpeg_write_marker(&p->encinfo, JPEG_APP0+1, (JOCTET *)app1, 40);
  for (i = 0; i < height/2; i++) {
    jpeg_write_scanlines( &p->encinfo, &input[i*2+d[0]], 1);
  }
  jpeg_finish_compress(&p->encinfo);

  if (p->fixheaders)
    used = fixheaders(output, 1, p);
  else
    used = fixApp1(output, 1);

  // Do the second field
  p->jdest.next_output_byte = output + used;
  p->jdest.free_in_buffer = out_size - used;	

  p->encinfo.image_width = width;
  p->encinfo.image_height = height/2;
  p->encinfo.input_components = 3;	/* RGB/YUV */
  jpeg_set_defaults(&p->encinfo);
  p->encinfo.dct_method = JDCT_FASTEST;
  jpeg_set_quality(&p->encinfo, p->quality, TRUE);
  p->encinfo.comp_info[0].v_samp_factor = 1;

  jpeg_start_compress(&p->encinfo, TRUE);
  if (p->fixheaders)
    jpeg_write_marker(&p->encinfo, JPEG_COM, (JOCTET *)comment, 10);
  jpeg_write_marker(&p->encinfo, JPEG_APP0+1, (JOCTET *)app1, 40);
  for (i = 0; i < height/2; i++) {
    jpeg_write_scanlines( &p->encinfo, &input[i*2+d[1]], 1);
  }
  jpeg_finish_compress(&p->encinfo);

  *IsAKeyFrame = 1;

  if (p->fixheaders)
    used2 = fixheaders(output + used, 1, p);
  else
    used2 = fixApp1(output + used, 1);

  file->moov.trak[track]->mdia.minf.stbl.stsd.table->fields = 2;
  file->moov.trak[track]->mdia.minf.stbl.stsd.table->field_dominance
    = p->dominance == 1 ? 1 : 6;

  return used + used2;
}

static int reads_colormodel_MJPA(oqt_t *file, int track, int colormodel) {
  return colormodel == BC_RGB888;
}

static int writes_colormodel_MJPA(oqt_t *file, int track, int colormodel) { 
  return colormodel == BC_RGB888;
}

static int preferred_colormodel_MJPA(oqt_t *file, int track) {
  return BC_RGB888;
}

/* Dummy decoding functions */
static void jpegdec_init_source (j_decompress_ptr cinfo){}
static int jpegdec_fill_input_buffer (j_decompress_ptr cinfo){
  /* There is no more data than we placed in
     the buffer in the first place */
  fprintf(stderr,
    "jpegdec_fill_input_buffer: error this should not be called.\n");
  return 0;	
}
static void jpegdec_skip_input_data (j_decompress_ptr cinfo, long num_bytes){
  if (num_bytes <=0) return;
  cinfo->src->next_input_byte += num_bytes;
  cinfo->src->bytes_in_buffer -= num_bytes;
  if (cinfo->src->bytes_in_buffer<0)
    fprintf(stderr, "jpegdec_skip_input_data: error, ran out of data.\n");
}
static void jpegdec_term_source (j_decompress_ptr cinfo){}

/* Dummy encoding functions */
static void jpegenc_init_destination (j_compress_ptr cinfo){}
static int jpegenc_flush_destination (j_compress_ptr cinfo){return 1;}
static void jpegenc_term_destination (j_compress_ptr cinfo){}

static int init_MJPA(oqt_t *file, int track) {
  /* Here we go, now we support as many initializatin as you want ;)  */
  Param *p;
  oqt_video_map_t *vtrack = &file->vtracks[track];
  
  /* Init params, should be called for every instance */
  vtrack->codec_private = malloc(sizeof(Param));
  if (!vtrack->codec_private) {
    fprintf(stderr, "init_MJPA: failed to allocate memory for codec.\n");
    return 0;
  }
  
  /* Setup jpeglib */
  p = (Param*)vtrack->codec_private;
  memset(p, 0, sizeof(Param));
  p->decinfo.err = p->encinfo.err = jpeg_std_error(&p->my_jerr.pub);
  
  /* Decoder init */
  jpeg_create_decompress(&p->decinfo);
  p->jsrc.init_source       = jpegdec_init_source;
  p->jsrc.fill_input_buffer = jpegdec_fill_input_buffer;
  p->jsrc.skip_input_data   = jpegdec_skip_input_data;
  p->jsrc.term_source       = jpegdec_term_source;
  p->decinfo.src = &p->jsrc;
    
  /* Encoder init */
  jpeg_create_compress(&p->encinfo);
  p->quality = JPEG_DEFAULT_QUALITY;
  p->dominance = MJPA_DEFAULT_DOMINANCE;
  p->fixheaders = 1;

  p->jdest.init_destination    = jpegenc_init_destination;
  p->jdest.empty_output_buffer = jpegenc_flush_destination;
  p->jdest.term_destination    = jpegenc_term_destination;
  p->encinfo.dest              = &p->jdest;
  
  return ++Initialised;
}

static int delete_MJPA(oqt_t *file, int track) {
  oqt_video_map_t *vtrack = &file->vtracks[track];
  Param *p = (Param*)(vtrack->codec_private);
  
  if(p) {
    // Free memory allocated by libjpeg
    jpeg_destroy_decompress(&p->decinfo);
    jpeg_destroy_compress(&p->encinfo);
    
    free(p);
    vtrack->codec_private=NULL;
  }
  
  return --Initialised;
}

oqt_codec_info_t* oqt_plugin_register(void) {
  oqt_video_codec_t *codec;
  
  // Allocate the memory for the codec record
  codec = oqt_allocate_video_codec();
  if (!codec) return NULL;
  
  // Information about the codec
  memcpy( &codec->info.fourcc, Signature, 4); 
  codec->info.name = Name;
  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;
  
  codec->init_codec = init_MJPA;
  codec->list_params = list_params_MJPA;
  codec->encode = encode_MJPA;
  codec->decode = decode_MJPA;
  codec->delete_codec = delete_MJPA;
  codec->reads_colormodel = reads_colormodel_MJPA;
  codec->writes_colormodel = writes_colormodel_MJPA;
  codec->preferred_colormodel = preferred_colormodel_MJPA;
  codec->set_param = set_param_MJPA;
  codec->get_param = get_param_MJPA;
  
  return &codec->info;
}

