/* codec_jpeg.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_jpeg.c,v 1.19 2002/12/30 17:34:08 antoinemine Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <jpeglib.h>

// Fix for compiler warning
#ifdef HAVE_STDLIB_H
#undef HAVE_STDLIB_H
#endif

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#ifdef __cplusplus
extern "C" {
#endif

/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "jpeg"; 
  /* Codec Signature */
  
static const char   *Name = "Photo JPEG"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 QT4Linux and OpenQuicktime Teams"; 
static const char   *ModuleAuthor  = ""; 
static const char   *CodecAuthor  = "libjpeg - Thomas G. Lane"; 
  
#define JPEG_DEFAULT_QUALITY 85
#define PARAM_COUNT 1
 
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "quality", OQT_PARAMETER_INTEGER, "Quality of JPEG Compression [0-100]"}
};


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	/* Error structure */
	struct my_error_mgr {
		struct jpeg_error_mgr pub;
		jmp_buf setjmp_buffer;
	} my_jerr;
	
	/* YUV Working buffer */
	oqt_uint8_t **yuv_line[3];
	
	/* quality */
	int quality;
	
	/* decompression structure */
	struct jpeg_decompress_struct decinfo;
	struct jpeg_source_mgr jsrc;
	
	/* compression structure */
	struct jpeg_compress_struct encinfo;
	struct jpeg_destination_mgr jdest;

} Param;  

static const oqt_codec_param_t* list_params_JPEG(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return ListParams;
}


static int set_param_JPEG (oqt_t* file,
		    int          track,
		    const char*  param,
		    const void*  data)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if (!strcmp(param,"quality")) {
		p->quality = *(( int*)data);
		jpeg_set_quality(&p->encinfo, p->quality, TRUE);
		
		return 0; 
	}
	
	fprintf(stderr,"set_param_JPEG: unknown parameter named '%s'\n",param);
	return OQT_PARAMETER_UNKNOWN;
}

static int get_param_JPEG (oqt_t* file,
		    int          track,
		    const char*  param,
		    void*        data)
{
  Param *p = (Param*)(file->vtracks[track].codec_private);

  if (!strcmp(param,"quality")) { *((int*)data) = p->quality; return OQT_PARAMETER_INTEGER; }
  
  fprintf(stderr,"set_param_JPEG: unknown parameter named '%s'\n",param);
  return OQT_PARAMETER_UNKNOWN;
}

static void jpeg_oqt_error_exit(j_common_ptr cinfo)
{
	struct my_error_mgr * jerr = (struct my_error_mgr *) cinfo->err;
	fprintf(stderr, "jpeg_oqt_error_exit: got error.\n");
	
	/* Display the error message */
	(*cinfo->err->output_message) (cinfo);
	
	/* Abort the compression/decompression so we can resuse */
	jpeg_abort(cinfo);

	/* Go to jump point near start of encode/decode */	
	longjmp(jerr->setjmp_buffer, 1);
}


/* Verify that the format of the frame is correct */
static inline int verify_jpeg_format( oqt_t *file, int track, Param *p )
{

	/* Check width */
	if (p->decinfo.output_width != oqt_get_video_width(file, track)) {
		fprintf(stderr, "verify_jpeg_format: error, width of jpeg frame is not same as track.\n");
		jpeg_abort_decompress(&p->decinfo);
		return 1;
	}
	
	/* Check height */
	if (p->decinfo.output_height != oqt_get_video_height(file, track)) {
		fprintf(stderr, "verify_jpeg_format: error, height of jpeg frame is not same as track.\n");
		jpeg_abort_decompress(&p->decinfo);
		return 1;
	}
	
	/* Check the comonents */
	if (p->decinfo.output_components != 3) {
		fprintf(stderr, "verify_jpeg_format: error, number of colour components is not 3.\n");
		jpeg_abort_decompress(&p->decinfo);
		return 1;
	}
	
	/* Check component size */
	if (BITS_IN_JSAMPLE!=8) {
		fprintf(stderr, "verify_jpeg_format: error, bits per JSample is not 8.\n");
		jpeg_abort_decompress(&p->decinfo);
		return 1;
	}

	
	/* Success */
	return 0;
}

static int allocate_jpeg_yuv( oqt_t *file, int track, Param *p )
{
	int height = oqt_get_video_height(file, track);

	/* Check not allocated */
	if (p->yuv_line[0] || p->yuv_line[1] || p->yuv_line[2]) {
		fprintf(stderr, "allocate_jpeg_yuv: buffers already allocated.\n");
		return 0;
	}

	/* Allocate memory for lines of YUV */
	p->yuv_line[0] = malloc(height*sizeof(oqt_uint8_t*));
	p->yuv_line[1] = malloc(height*sizeof(oqt_uint8_t*)/2);
	p->yuv_line[2] = malloc(height*sizeof(oqt_uint8_t*)/2);
	
	if(!p->yuv_line[0] || !p->yuv_line[1] || !p->yuv_line[2]) {
		fprintf(stderr, "allocate_jpeg_yuv: failed to allocate memory for YUV buffers.\n");
		return 1;
	}
	
	/* Success */
	return 0;
}


/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_JPEG(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if(color_model == BC_NONE) {
		return 0;
	}


	/* Setup return point for errors */
	if(setjmp(p->my_jerr.setjmp_buffer)) {
		// exception: we longjmp'ed here
		fprintf(stderr, "decode_JPEG: caught error - failed to decode JPEG stream\n");
		return 1;
	}
	p->my_jerr.pub.error_exit = jpeg_oqt_error_exit;
	
	
	/* Read in the JPEG headers */
	p->jsrc.next_input_byte = input;
	p->jsrc.bytes_in_buffer = inputsize;
	jpeg_read_header(&p->decinfo, TRUE);


	/* Configure Decoder */
	p->decinfo.do_fancy_upsampling	= FALSE;
	p->decinfo.do_block_smoothing	= FALSE;
	p->decinfo.dct_method       	= JDCT_IFAST;
   	p->decinfo.dither_mode			= JDITHER_ORDERED;


	switch(color_model) {
		case BC_RGB888: {
			int i;
			
			p->decinfo.raw_data_out = FALSE;
			p->decinfo.out_color_space = JCS_RGB;

			jpeg_start_decompress(&p->decinfo);
			
			/* Check format is correct */
			if (verify_jpeg_format(file, track, p)) break;
			
			/* Read scanlines one by one into output buffer */
			for (i = 0; i < p->decinfo.output_height; i++) {
				jpeg_read_scanlines( &p->decinfo, &output[i], 1);
			}
			
			jpeg_finish_decompress(&p->decinfo);
			break;
		}
		
		case BC_YUV420P: {
			int width, height;
			int width2, i, j, k, r_v;
			oqt_uint8_t *base[3];
			base[0] = output[0];
			base[1] = output[1];
			base[2] = output[2];

			/* get read for raw data output */
			r_v = p->decinfo.cur_comp_info[0]->v_samp_factor;
			p->decinfo.raw_data_out = TRUE;
			p->decinfo.out_color_space = JCS_YCbCr;

			/* Start decompressing JPEG */
			jpeg_start_decompress(&p->decinfo);
			width = p->decinfo.output_width;
			height = p->decinfo.output_height;
			width2 = width >> 1;
			
			/* Check format is correct */
			if (verify_jpeg_format(file, track, p)) break;

			/* Make sure YUV buffers have been allocated */
			if (!p->yuv_line[0] && allocate_jpeg_yuv(file, track, p)) break;

			/* Copy raw data into output buffer */
			for (i = 0; i < height; i += r_v*DCTSIZE) {
				for (j=0, k=0; j< (r_v*DCTSIZE); j += r_v, k++) {
					p->yuv_line[0][j]   = base[0]; base[0] += width;
					if (r_v == 2) {
						p->yuv_line[0][j+1] = base[0]; base[0] += width;
					}
					p->yuv_line[1][k]   = base[1]; 
					p->yuv_line[2][k]   = base[2];
					if (r_v == 2 || k&1) {
						base[1] += width2; base[2] += width2;
					}
				}
				jpeg_read_raw_data(&p->decinfo, p->yuv_line, r_v*DCTSIZE);
			}
			
			jpeg_finish_decompress(&p->decinfo);
			break;
		}
		
		default:
			fprintf(stderr, "decode_JPEG: decoding of colourmodel=%d not supported.\n", 
							color_model);
		break;
	}
	
	return 0;
}


/*************************************************************/
/* Encoding function                                         */
/*************************************************************/
static int encode_JPEG(oqt_t *file, 
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
	Param *p = (Param*)(vtrack->codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	
	
	/* Setup return point for errors */
	if(setjmp(p->my_jerr.setjmp_buffer)) {
		// exception: we longjmp'ed here
		fprintf(stderr, "encode_JPEG: caught error - libjpeg failed.\n");
		return 0;
	}
	p->my_jerr.pub.error_exit = jpeg_oqt_error_exit;


	/* Set the input colour model */
	switch(color_model) {
		case BC_YUV420P: p->encinfo.in_color_space = JCS_YCbCr; break;
		case BC_RGB888:	p->encinfo.in_color_space = JCS_RGB; break;
		default:
			fprintf(stderr, "encode_JPEG: encoding of colourmodel=%d not supported.\n", 
							color_model);
			return 0;
		break;
	}

	/* Setup the output buffer */
	p->jdest.next_output_byte = output;
	p->jdest.free_in_buffer = out_size;	
	
	/* Setup the compressor */
	p->encinfo.image_width = width;
	p->encinfo.image_height = height;
	p->encinfo.input_components = 3;	/* RGB/YUV */
	jpeg_set_defaults(&p->encinfo);
	p->encinfo.dct_method = JDCT_FASTEST;
	jpeg_set_quality(&p->encinfo, p->quality, TRUE);
	jpeg_suppress_tables(&p->encinfo, TRUE);
	
	
	/* Compress the frame */
	switch(color_model) {
		case BC_RGB888: {
			int i;
			
			jpeg_start_compress(&p->encinfo, TRUE);
			for (i = 0; i < height; i++) {
				jpeg_write_scanlines( &p->encinfo, &input[i], 1);
			}
			jpeg_finish_compress(&p->encinfo);
			
			break;
		}
		
		case BC_YUV420P: {
			int width2, i, j, k;
			oqt_uint8_t *base[3];
			base[0] = input[0];
			base[1] = input[1];
			base[2] = input[2];

			p->encinfo.raw_data_in = TRUE;
			p->encinfo.comp_info[0].h_samp_factor = 2;
			p->encinfo.comp_info[0].v_samp_factor = 2;
			p->encinfo.comp_info[1].h_samp_factor = 1;
			p->encinfo.comp_info[1].v_samp_factor = 1;
			p->encinfo.comp_info[2].h_samp_factor = 1;
			p->encinfo.comp_info[2].v_samp_factor = 1;
					
			jpeg_start_compress(&p->encinfo, TRUE);
			width2 = width>>1;
			
			/* Make sure YUV buffers have been allocated */
			if (!p->yuv_line[0] && allocate_jpeg_yuv(file, track, p)) break;
			
			for (i = 0; i < height; i += 2*DCTSIZE) {
				for (j=0, k=0; j<2*DCTSIZE;j+=2, k++) {
					p->yuv_line[0][j]   = base[0]; base[0] += width;
					p->yuv_line[0][j+1] = base[0]; base[0] += width;
					p->yuv_line[1][k]   = base[1]; base[1] += width2;
					p->yuv_line[2][k]   = base[2]; base[2] += width2;
				}
				jpeg_write_raw_data(&p->encinfo, p->yuv_line, 2*DCTSIZE);
			}
			jpeg_finish_compress(&p->encinfo);
			break;
		}
	}
	
	
	*IsAKeyFrame = 1;
	
	return (((out_size - p->jdest.free_in_buffer)+3)&~3);
}

static int reads_colormodel_JPEG(oqt_t *file, 
		int track,
		int colormodel)
{
	return ( colormodel == BC_YUV420P ||
	         colormodel == BC_RGB888 );
}

static int writes_colormodel_JPEG(oqt_t *file, 
		int track,
		int colormodel)
{ 
	return ( colormodel == BC_YUV420P ||
	         colormodel == BC_RGB888 );
}

static int preferred_colormodel_JPEG(oqt_t *file, 
		int track)
{
	return BC_YUV420P;
}


/* Dummy decoding functions */
static void jpegdec_init_source (j_decompress_ptr cinfo){}
static int jpegdec_fill_input_buffer (j_decompress_ptr cinfo){
	/* There is no more data than we placed in
	   the buffer in the first place */
	fprintf(stderr, "jpegdec_fill_input_buffer: error this should not be called.\n");
	return 0;	
}
static void jpegdec_skip_input_data (j_decompress_ptr cinfo, long num_bytes){
	if (num_bytes <=0) return;
	cinfo->src->next_input_byte += num_bytes;
	cinfo->src->bytes_in_buffer -= num_bytes;
	if (cinfo->src->bytes_in_buffer<0)
	fprintf(stderr, "jpegdec_skip_input_data: error, ran out of data.\n");
}
static void jpegdec_term_source (j_decompress_ptr cinfo){}

/* Dummy encoding functions */
static void jpegenc_init_destination (j_compress_ptr cinfo){}
static int jpegenc_flush_destination (j_compress_ptr cinfo){return 1;}
static void jpegenc_term_destination (j_compress_ptr cinfo){}


static int init_JPEG(oqt_t *file, int track)
{
	/* Here we go, now we support as many initializatin as you want ;)  */
	Param *p;
	oqt_video_map_t *vtrack = &file->vtracks[track];
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_JPEG: failed to allocate memory for codec.\n");
		return 0;
	}
	
	/* Setup jpeglib */
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	p->decinfo.err = p->encinfo.err = jpeg_std_error(&p->my_jerr.pub);
	
	/* Set YUV buffers pointers to null */
	p->yuv_line[0] = NULL;
	p->yuv_line[1] = NULL;
	p->yuv_line[2] = NULL;

	/* Decoder init */
	jpeg_create_decompress(&p->decinfo);
	p->jsrc.init_source       = jpegdec_init_source;
	p->jsrc.fill_input_buffer = jpegdec_fill_input_buffer;
	p->jsrc.skip_input_data   = jpegdec_skip_input_data;
	p->jsrc.term_source       = jpegdec_term_source;
	p->decinfo.src = &p->jsrc;
	
	
	/* Encoder init */
	jpeg_create_compress(&p->encinfo);
	p->quality = JPEG_DEFAULT_QUALITY;
		
	p->jdest.init_destination    = jpegenc_init_destination;
	p->jdest.empty_output_buffer = jpegenc_flush_destination;
	p->jdest.term_destination    = jpegenc_term_destination;
	p->encinfo.dest              = &p->jdest;
	
	return ++Initialised;
}


static int delete_JPEG(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		if (p->yuv_line[0]) { free(p->yuv_line[0]); p->yuv_line[0] = NULL; }
		if (p->yuv_line[1]) { free(p->yuv_line[1]); p->yuv_line[1] = NULL; }
		if (p->yuv_line[2]) { free(p->yuv_line[2]); p->yuv_line[2] = NULL; }
		
		// Free memory allocated by libjpeg
		jpeg_destroy_decompress(&p->decinfo);
	  	jpeg_destroy_compress(&p->encinfo);

		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;

	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;

	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_JPEG;
	codec->list_params = list_params_JPEG;
	codec->encode = encode_JPEG;
	codec->decode = decode_JPEG;
	codec->delete_codec = delete_JPEG;
	codec->reads_colormodel = reads_colormodel_JPEG;
	codec->writes_colormodel = writes_colormodel_JPEG;
	codec->preferred_colormodel = preferred_colormodel_JPEG;
	codec->set_param = set_param_JPEG;
	codec->get_param = get_param_JPEG;
	
	return &codec->info;
}

