/* codec_png.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_png.c,v 1.6 2002/12/15 01:52:00 nj_humfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <png.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#ifdef __cplusplus
extern "C" {
#endif


/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "png "; 
  /* Codec Signature */
  
static const char   *Name = "Portable Network Graphics (PNG)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 QT4Linux and OpenQuicktime Teams"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "libpng"; 
  
#define PARAM_COUNT 1
 
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "compression", OQT_PARAMETER_INTEGER, "PNG Compression (0-9), Default 6"}
};


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	int compression_level;
	
	oqt_uint32_t buffer_size;
	oqt_uint32_t buffer_position;
	oqt_uint32_t buffer_allocated;
	oqt_uint8_t  *buffer;
	
} Param;  

static const oqt_codec_param_t* list_params_png(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return ListParams;
}


static int set_param_png (oqt_t* file,
		    int          track,
		    const char*  param,
		    const void*  data)
{
  Param *p = (Param*)(file->vtracks[track].codec_private);

  if (!strcmp(param,"compression")) {
    p->compression_level = *(( int*)data);
    //jpeg_set_quality(&p->encinfo, p->compression_level, TRUE);

    return 0; 
  }

  fprintf(stderr,"set_param_png: unknown parameter named '%s'\n",param);
  return OQT_PARAMETER_UNKNOWN;
}

static int get_param_png (oqt_t* file,
		    int          track,
		    const char*  param,
		    void*        data)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if (!strcmp(param,"compression")) {
		*((int*)data) = p->compression_level;
		return OQT_PARAMETER_INTEGER;
	}
	
	fprintf(stderr,"set_param_png: unknown parameter named '%s'\n",param);
	return OQT_PARAMETER_UNKNOWN;
}




/*************************************************************/
/* Callbacks                                                 */
/*************************************************************/

static void read_function(png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	Param *p = png_get_io_ptr(png_ptr);
	
	if(length + p->buffer_position <= p->buffer_size)
	{
		memcpy(data, p->buffer + p->buffer_position, length);
		p->buffer_position += length;
	} else {
		fprintf(stderr, "codec_png: ran out of buffer space in read_function.\n");
	}
}

static void write_function(png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	Param *p = png_get_io_ptr(png_ptr);

	if(length + p->buffer_position <= p->buffer_size)
	{
		memcpy(p->buffer + p->buffer_position, data, length);
		p->buffer_position += length;
	} else {
		fprintf(stderr, "codec_png: ran out of buffer space in write_function.\n");
	}
}

static void flush_function(png_structp png_ptr)
{
	;
}


static void reset_buffer_png(Param *p) {
	p->buffer_size = 0;
	p->buffer_position = 0;
	p->buffer_allocated = 0;
	p->buffer = NULL;
}


/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_png(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	png_structp png_ptr;
	png_infop info_ptr;
	png_infop end_info = 0;	
	int png_depth, color_type;
	
	if(color_model == BC_NONE) {
		return 0;
	}
	
	reset_buffer_png( p );
	p->buffer_size = inputsize;
	p->buffer = input;
	
	/* Prepare libpng and read info */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	info_ptr = png_create_info_struct(png_ptr);
	png_set_read_fn(png_ptr, p, (png_rw_ptr)read_function);
	png_read_info(png_ptr, info_ptr);
	
	/* Check that png is the right format */
	color_type = png_get_color_type(png_ptr, info_ptr);

	if (png_get_bit_depth(png_ptr,info_ptr) != 8) {
		fprintf(stderr, "decode_png: unsupported PNG channel depth.\n");
	} else if (oqt_get_video_depth(file, track) == 24 && color_type != PNG_COLOR_TYPE_RGB) {
		fprintf(stderr, "decode_png: invalid PNG colour type for 24 bit movie.\n");
	} else if (oqt_get_video_depth(file, track) == 32 && color_type != PNG_COLOR_TYPE_RGB_ALPHA) {
		fprintf(stderr, "decode_png: invalid PNG colour type for 32 bit movie.\n");
	} else if (png_get_image_width(png_ptr,info_ptr) != oqt_get_video_width(file, track)) {
		fprintf(stderr, "decode_png: width of png does not match that of movie.\n");
	} else if (png_get_image_height(png_ptr,info_ptr) != oqt_get_video_height(file, track)) {
		fprintf(stderr, "decode_png: height of png does not match that of movie.\n");
	} else {
	
		/* go ahead and read the image */
		png_read_image(png_ptr, output);
		
	}

	/* Clean up */
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	reset_buffer_png( p );
	
	return 0;
}


/*************************************************************/
/* Encoding function                                         */
/*************************************************************/
static int encode_png(oqt_t *file, 
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
	Param *p = (Param*)(vtrack->codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int depth 	= oqt_get_video_depth(file, track);
	int frame_size = 0;
	png_structp png_ptr;
	png_infop info_ptr;
 	
 	reset_buffer_png( p );
	p->buffer = output;
	p->buffer_size = out_size;
 	
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	info_ptr = png_create_info_struct(png_ptr);
	png_set_write_fn(png_ptr, p,
  			   (png_rw_ptr)write_function,
               (png_flush_ptr)flush_function);
	png_set_compression_level(png_ptr, p->compression_level);

	png_set_IHDR(png_ptr, 
		info_ptr, width, height, 8, 
		color_model == BC_RGB888 ? 
		  PNG_COLOR_TYPE_RGB : 
		  PNG_COLOR_TYPE_RGB_ALPHA, 
		PNG_INTERLACE_NONE, 
		PNG_COMPRESSION_TYPE_DEFAULT, 
		PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png_ptr, info_ptr);
	png_write_image(png_ptr, input);
	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	*IsAKeyFrame = 1;
	frame_size = p->buffer_position;
	reset_buffer_png( p );
	
	
	return frame_size;
}

static int writes_colormodel_png(oqt_t *file, 
		int track,
		int colormodel)
{ 
	return ( colormodel == BC_RGB888 ||
			 colormodel == BC_RGBA8888 );
}

static int reads_colormodel_png(oqt_t *file, 
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	
	//if (depth == 8)		return (colormodel == BC_RGB8);
	//else if (depth == 16)	return (colormodel == BC_RGB565);
	if (depth == 24)		return (colormodel == BC_RGB888);
	else if (depth == 32)	return (colormodel == BC_RGBA8888);
	
	return 0;
}

static int preferred_colormodel_png(oqt_t *file, 
		int track)
{
	int depth = oqt_get_video_depth(file, track);
	if (depth == 24)		return BC_RGB888;
	else if (depth == 32)	return BC_RGBA8888;
	else return BC_UNSUPPORTED;
}



static int init_png(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_png: failed to allocate memory for codec.\n");
		return 0;
	}
	
	/* setup libpng */
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	reset_buffer_png( p );
	p->compression_level = 6;
	
	return ++Initialised;
}


static int delete_png(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;

	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;

	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_png;
	codec->list_params = list_params_png;
	codec->encode = encode_png;
	codec->decode = decode_png;
	codec->delete_codec = delete_png;
	codec->reads_colormodel = reads_colormodel_png;
	codec->writes_colormodel = writes_colormodel_png;
	codec->preferred_colormodel = preferred_colormodel_png;
	codec->set_param = set_param_png;
	codec->get_param = get_param_png;
	
	return &codec->info;
}


