/* codec_rle.c
 * Copyright (C) 2002 OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_rle.c,v 1.12 2002/12/28 18:50:59 nj_humfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#ifdef WORDS_BIGENDIAN
// big endian
#define RGBA_TO_RGBA_NATIVE(r,g,b,a) (r<<24|g<<16|b<<8|a)
#define RGBA_TO_ARGB_NATIVE(r,g,b,a) (a<<24|r<<16|g<<8|b)
#define RGBA_TO_ABGR_NATIVE(r,g,b,a) (a<<24|b<<16|g<<8|r)
#define RGBA_TO_BGRA_NATIVE(r,g,b,a) (b<<24|g<<16|r<<8|a)
#else
// little endian
#define RGBA_TO_RGBA_NATIVE(r,g,b,a) (a<<24|b<<16|g<<8|r)
#define RGBA_TO_ARGB_NATIVE(r,g,b,a) (b<<24|g<<16|r<<8|a)
#define RGBA_TO_ABGR_NATIVE(r,g,b,a) (r<<24|g<<16|b<<8|a)
#define RGBA_TO_BGRA_NATIVE(r,g,b,a) (a<<24|r<<16|g<<8|b)
#endif

static int          Initialised   = 0;      
static const char   Signature[5]  = "rle ";
static const char   *Name = "Apple Animation (RLE)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team";
static const char   *ModuleAuthor  = "Alexander Lechner"; 
static const char   *CodecAuthor  = "Alexander Lechner / Mark Podlipec"; 

static const oqt_codec_param_t params[] =
{
	{ "warn_conversion", OQT_PARAMETER_INTEGER, "Warn about slow conversion routines"},
};

#define PARAM_COUNT (sizeof(params)/sizeof(params[0]))

typedef struct {
	unsigned char  *mTempData;
	long            mTempSize;
	int             mWarnConversion;
} Private;

static void decodeRLE16_RGB565(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);
static void decodeRLE16_RGB888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);
static void decodeRLE24_RGB888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);
static void decodeRLE24_BGR888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);
static void decodeRLE32_RGBA8888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);
static void decodeRLE32_ARGB8888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output);

/* allocate one a temp buffer for RLE */
static int check_temp_buffer_rle(oqt_t *file, int track, long size)
{
	Private *p = (Private*)(file->vtracks[track].codec_private);

	if (!p->mTempData) {
		p->mTempSize=size;
		p->mTempData=(unsigned char*)malloc(p->mTempSize);
	} else if (p->mTempSize<size) {
		p->mTempSize=size;
		p->mTempData=(unsigned char*)realloc(p->mTempData,p->mTempSize);
	}
	if (!p->mTempData) {
		fprintf(stderr,"check_temp_buffer_rle: can't (re-)allocate temp buffer (%ld bytes)\n",size);
		return -1;
	}
	return 0;
}

static const oqt_codec_param_t* list_params_rle(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	
	return params;
}

static int set_param_rle(oqt_t* file,
			  int          track,
			  const char*  key,
			  const void*  value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "warn_conversion"))
    p->mWarnConversion = *(int*)value;
  else {
    fprintf(stderr,"set_param_rle: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
  return 0;
}

static int get_param_rle(oqt_t* file,
			  int          track,
			  const char*  key,
			  void*        value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "warn_conversion")) {
    *(int*)value = p->mWarnConversion;
    return OQT_PARAMETER_INTEGER;
  } else {
    fprintf(stderr,"get_param_rle: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
}



/* see if row_pointers are consecutive 
   - is this acutally a performance gain ?
*/
/*static int rows_consecutive_rle(unsigned char **row_pointers, int w, int h, int depth)
{
	int i, result;
	int bpl = w*depth;
	for(i = 1, result = 1; i < h; i++)
	{
		if(row_pointers[i] - row_pointers[i - 1] != bpl) result = 0;
	}
	return result;
}*/

/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_rle(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t *input,
		       oqt_uint8_t **output,
		       int color_model)
{
	Private *p = (Private*)(file->vtracks[track].codec_private);
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int depth   = oqt_get_video_depth(file, track);
	int y, bpl, frame_size;
	oqt_uint8_t* src;

	if (color_model==BC_NONE) return 0;
	
	/* allocate temp buffer or resize buffer if needed */
	frame_size = oqt_cmodel_calculate_framesize(width, height, color_model, -1);
	if (check_temp_buffer_rle(file, track, frame_size)) return -1;

	/* color_model is desired output format! */
	switch (depth) {
		case 32:
			switch(color_model) {
			case BC_ARGB8888:
				decodeRLE32_ARGB8888(input,inputsize,width,height,p->mTempData);
				break;
			case BC_RGBA8888:
				decodeRLE32_RGBA8888(input,inputsize,width,height,p->mTempData);
				break;
			default:
				fprintf(stderr,"decode_rle: decoding of depth=%d colormodel=%d not supported!\n",depth,color_model);
				return -1;
			}
			break;
		case 24:
			switch(color_model) {
			case BC_RGB888:
				decodeRLE24_RGB888(input,inputsize,width,height,p->mTempData);
				break;
			case BC_BGR888:
				decodeRLE24_BGR888(input,inputsize,width,height,p->mTempData);
				break;
			default:
				fprintf(stderr,"decode_rle: decoding of depth=%d colormodel=%d not supported!\n",depth,color_model);
				return -1;
			}
			break;
		case 16:
			switch(color_model) {
			case BC_RGB565:
				decodeRLE16_RGB565(input,inputsize,width,height,p->mTempData);
				break;
			case BC_RGB888:
				decodeRLE16_RGB888(input,inputsize,width,height,p->mTempData);
				break;
			default:
				fprintf(stderr,"decode_rle: decoding of depth=%d colormodel=%d not supported!\n",depth,color_model);
				return -1;
			}
			break;
		default:
			fprintf(stderr,"decode_rle: decoding of depth=%d not supported!\n",depth);
			return -1;
	}


	/* Copy decoded buffer into output buffer */
	bpl = width * oqt_cmodel_calculate_pixelsize(color_model);
	src = p->mTempData;
	for (y=0;y<height;y++,src+=bpl) {
		 memcpy(output[y],src,bpl);
	}

	return 0;
}




/*************************************************************/
/* Encoding function                                         */
/*************************************************************/
/*static int encode_rle(oqt_t *file, 
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
  int width 	= (int)vtrack->track->tkhd.track_width;
  int height 	= (int)vtrack->track->tkhd.track_height;
	int depth   = oqt_get_video_depth(file, track);
	int bpp     = (depth+7)/8;		
	int useTemp = !rows_consecutive_rle(input,width,height,bpp);
	int encode  = 0;
	int framesize = width*height*bpp;
	
	switch(color_model) {
  case BC_ARGB8888:
    break;
	case BC_RGB888:
		break;
	case BC_NONE:
		break;
	default:
		fprintf(stderr,"encode_rle: encoding of '%d' not supported!\n",color_model);
		return -1;
  }

	*IsAKeyFrame = 1;
			
	if (useTemp) {
		int y;
		int bpl = width*bpp;
		unsigned char* d = output;
		for (y=0;y<height;y++,d+=bpl) {
			memcpy(d,input[y],bpl);
		}
	} else {
		memcpy(output,input[0],framesize);
	}
  return framesize;
}*/

static int reads_colormodel_rle(oqt_t *file, 
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	
	if (depth==16) {
		return (colormodel == BC_RGB888 || colormodel == BC_RGB565);
	} else if (depth==24) {
		return (colormodel == BC_RGB888 || colormodel == BC_BGR888);
	} else if (depth==32) {
		return (colormodel == BC_RGBA8888 || colormodel ==  BC_ARGB8888);
	} else
		return 0;
}

static int preferred_colormodel_rle(oqt_t *file, 
		int track)
{
	int depth = oqt_get_video_depth(file, track);
	
	if (depth==16) 			return BC_RGB888;
	else if (depth==24) 	return BC_RGB888;
	else if (depth==32) 	return BC_RGBA8888;
	else					return BC_UNSUPPORTED;
}

/* no encoding yet
static int writes_colormodel_rle(oqt_t *file, 
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	if (depth==24) {
	return (colormodel == BC_RGB888);
	} else if (depth==32) {
	return (colormodel == BC_RGBA8888 || colormodel ==  BC_ARGB8888);
	} else
	return 0;
}
*/

static int init_rle(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	if (!p) {
		fprintf(stderr,"init_rle: not enough memory!\n");
		return 0;
	}
	memset(p, 0, sizeof(Private));
	vtrack->codec_private = p;
	p->mWarnConversion=1;
	
  return ++Initialised;
}

static int delete_rle(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = (Private*)(vtrack->codec_private);

	if(p) {
		if (p->mTempData) free(p->mTempData);
		free(p);
		vtrack->codec_private = NULL;
	}
	//fprintf(stderr,"delete_rle: External video codec %s deleted\n", Signature);

	return --Initialised;
}


oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;
	
	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4);
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_rle;
	codec->list_params = list_params_rle;
	//codec->encode = encode_rle;
	codec->decode = decode_rle;
	codec->delete_codec = delete_rle;
	codec->reads_colormodel = reads_colormodel_rle;
	// codec->writes_colormodel = writes_colormodel_rle; // no encoding yet
	codec->preferred_colormodel = preferred_colormodel_rle;
	codec->set_param = set_param_rle;
	codec->get_param = get_param_rle;
	
	return &codec->info;
}

static void
decodeRLE16_RGB565(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
	register oqt_uint8_t *dp = input;
	register oqt_uint8_t *iptr;      
	register oqt_uint32_t d, d1,d2; 
	
	dp += 4;    /* skip codec size */  
	d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
	if (dsize < 8) { /* NOP */
		return;
	}
	
	if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
	else {
		y = 0;
		lines = imagey;
	}

	while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint8_t*)(output + 2*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += 3*(xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					d1 = *dp++; d2 = *dp++;
					*iptr++ = d1; *iptr++ = d2;
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				d1 = *dp++; d2 = *dp++;
				while(cnt--) {
					*iptr++ = d1; *iptr++ = d2;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}
static void
decodeRLE16_RGB888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
	register oqt_uint8_t *dp = input;
	register oqt_uint8_t *iptr;      
	register oqt_uint32_t r,g,b,d; 
	
	dp += 4;    /* skip codec size */  
	d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
	if (dsize < 8) { /* NOP */
		return;
	}
	
	if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
	else {
		y = 0;
		lines = imagey;
	}

	while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint8_t*)(output + 3*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += 3*(xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					d = (*dp++ << 8); d |= *dp++;
					r = (d >> 10) & 0x1f;	r = (r << 3) | (r >> 2);
					g = (d >>  5) & 0x1f;	g = (g << 3) | (g >> 2);
					b =  d & 0x1f;			b = (b << 3) | (b >> 2);
					*iptr++ = r;
					*iptr++ = g;
					*iptr++ = b;
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				d = (*dp++ << 8); d |= *dp++;
				r = (d >> 10) & 0x1f;	r = (r << 3) | (r >> 2);
				g = (d >>  5) & 0x1f;	g = (g << 3) | (g >> 2);
				b =  d & 0x1f;			b = (b << 3) | (b >> 2);
				while(cnt--) {
					*iptr++ = r;
					*iptr++ = g;
					*iptr++ = b;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}


static void
decodeRLE24_RGB888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
	register oqt_uint8_t *dp = input;
	register oqt_uint8_t *iptr;      
	register oqt_uint32_t r,g,b,d; 
	
	dp += 4;    /* skip codec size */  
  d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
  if (dsize < 8) { /* NOP */
		return;
	}

  if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
  else {
		y = 0;
		lines = imagey;
	}

  while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint8_t*)(output + 3*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += 3*(xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					r = *dp++;
					g = *dp++;
					b = *dp++;
					*iptr++ = r;
					*iptr++ = g;
					*iptr++ = b;
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				r = *dp++;
				g = *dp++;
				b = *dp++;
				while(cnt--) {
					*iptr++ = r;
					*iptr++ = g;
					*iptr++ = b;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}

static void
decodeRLE24_BGR888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
	register oqt_uint8_t *dp = input;
	register oqt_uint8_t *iptr;      
	register oqt_uint32_t r,g,b,a,d; 
	
	dp += 4;    /* skip codec size */  
  d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
  if (dsize < 8) { /* NOP */
		return;
	}

  if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
  else {
		y = 0;
		lines = imagey;
	}

  while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint8_t*)(output + 3*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += 3*(xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					r = *dp++;
					g = *dp++;
					b = *dp++;
					*iptr++ = b;
					*iptr++ = g;
					*iptr++ = r;
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				r = *dp++;
				g = *dp++;
				b = *dp++;
				while(cnt--) {
					*iptr++ = b;
					*iptr++ = g;
					*iptr++ = r;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}


static void
decodeRLE32_ARGB8888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
  register oqt_uint8_t *dp = input;
	register oqt_uint32_t *iptr;      
	register oqt_uint32_t r,g,b,a,d; 
	
	dp += 4;    /* skip codec size */  
  d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
  if (dsize < 8) { /* NOP */
		return;
	}

  if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
  else {
		y = 0;
		lines = imagey;
	}

  while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint32_t*)(output + 4*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += (xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					a = *dp++;
					r = *dp++;
					g = *dp++;
					b = *dp++;
					*iptr++ = RGBA_TO_ARGB_NATIVE(r,g,b,a); // for ARGB
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				a = *dp++;
				r = *dp++;
				g = *dp++;
				b = *dp++;
				d = RGBA_TO_ARGB_NATIVE(r,g,b,a); // for ARGB
				while(cnt--) {
					*iptr++ = d;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}

static void
decodeRLE32_RGBA8888(unsigned char* input,unsigned long inputsize,int width,int height,unsigned char *output)
{
	oqt_uint32_t imagex = width,imagey = height,y,lines,cnt,dsize=inputsize,xskip;
	register oqt_uint8_t *dp = input;
	register oqt_uint32_t *iptr;      
	register oqt_uint32_t r,g,b,a,d; 
	
	dp += 4;    /* skip codec size */  
  d = (*dp++) << 8;  d |= *dp++;   /* read code either 0008 or 0000 */ 
  if (dsize < 8) { /* NOP */
		return;
	}

  if (d & 0x0008) { /* Header present */
		y = (*dp++) << 8;  y |= *dp++;           /* start line */
		dp += 2;                                  /* unknown */
		lines = (*dp++) << 8; lines |= *dp++;   /* number of lines */
		dp += 2;                                  /* unknown */
	}
  else {
		y = 0;
		lines = imagey;
	}

  while(lines--) {			/* loop thru lines */
		xskip = *dp++;				/* skip x pixels */
		if (xskip == 0) break;			/* exit */
		cnt = *dp++;				/* RLE code */
		iptr = ( oqt_uint32_t*)(output + 4*((y * imagex)+(xskip-1)) );
		while(cnt != 0xff) {			/* while not EOL */
			if (cnt == 0x00) {
				xskip = *dp++;
				iptr += (xskip-1);
			}
			else if (cnt < 0x80) {			/* run of data */
				while(cnt--) {
					a = *dp++;
					r = *dp++;
					g = *dp++;
					b = *dp++;
					*iptr++ = RGBA_TO_RGBA_NATIVE(r,g,b,a); // for ARGB
				}
			}
			else	{				// repeat data
				cnt = 0x100 - cnt;
				a = *dp++;
				r = *dp++;
				g = *dp++;
				b = *dp++;
				d = RGBA_TO_RGBA_NATIVE(r,g,b,a); // for ARGB
				while(cnt--) {
					*iptr++ = d;
				}
			}
			cnt = *dp++;				// get new RLE code
		} // end of line 
		y++;
	}
}

