/* codec_raw.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_raw.c,v 1.10 2002/12/15 01:52:00 nj_humfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"

static int          Initialised   = 0;
static const char   Signature[5]  = "raw ";
static const char   *Name = "Uncompressed (RGB)";
static const char   *Version  = "0.1";
static const char   *Copyright  = "2002 OpenQuicktime Team";
static const char   *ModuleAuthor  = "Jozef Hatala and Nicholas Humfrey";
static const char   *CodecAuthor  = "Adam Williams and Nicholas Humfrey";



static int check_ctab_raw( oqt_ctab_t *ctab ) {
    if (!ctab || ctab->size==0) {
    	fprintf(stderr, "decode_smc: can't decode - colour table for track is not available.\n");
    	return 1;
    } else {
    	return 0;
    }
}

static inline void set_pixel_ctab_raw(oqt_uint8_t* row, int x, oqt_uint8_t color_index, int color_model, oqt_ctab_t* ctab)
{
	if (color_index>ctab->size) {
		fprintf(stderr, "set_pixel_ctab_raw: color_index is not in range (0<%d<%d).\n", (int)color_index, (int)ctab->size);
		return;
	}

	if (color_model==BC_ARGB8888) {
		row[(4*x)+0] = ctab->alpha[color_index];
		row[(4*x)+1] = ctab->red[color_index];
		row[(4*x)+2] = ctab->green[color_index];
		row[(4*x)+3] = ctab->blue[color_index];
	} else {
		row[(3*x)+0] = ctab->red[color_index];
		row[(3*x)+1] = ctab->green[color_index];
		row[(3*x)+2] = ctab->blue[color_index];
	}
}


static int decode_raw(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t *input,
		       oqt_uint8_t **output,
		       int color_model)
{
	oqt_ctab_t *ctab = &file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].ctab;
	int width = oqt_get_video_width(file, track);
	int height = oqt_get_video_height(file, track);
	int depth = oqt_get_video_depth(file, track);

	if(color_model == BC_NONE) {
		// drop frame
		return 0;
	
	} else {
		int sample_bytes = depth/8;
		oqt_uint32_t pixel;
		int x,y, alpha;
		
		/*if(inputsize != height * width * sample_bytes) {
			fprintf(stderr, "decode_raw: data size mismatch.\n");
			return 1;
		}*/
	
		switch(depth) {
		
			case 32:
			case 24:
				for (y=0; y< height; ++y) {
					oqt_uint8_t* row = output[y];
					for (x=0; x< width; ++x) {
						if (depth==32) alpha = *input++;
						else alpha = 0;
						if (color_model==BC_ARGB8888) {
							row[(4*x)+0] = alpha;
							row[(4*x)+1] = *input++;
							row[(4*x)+2] = *input++;
							row[(4*x)+3] = *input++;
						} else {
							row[(3*x)+0] = *input++;
							row[(3*x)+1] = *input++;
							row[(3*x)+2] = *input++;
						}
					}
				}
			break;
			
			case 16:
				for (y=0; y< height; ++y) {
					oqt_uint8_t* row = output[y];
					for (x=0; x< width; ++x) {
						oqt_uint8_t r,g,b;
						pixel = *input++ << 8;
						pixel |= *input++;
						r = (pixel >> 10) & 0x1f;
						g = (pixel >> 5) & 0x1f;
						b = pixel & 0x1f;
						
						if (color_model==BC_ARGB8888) {
							row[(4*x)+0] = 0;
							row[(4*x)+1] = (r << 3) | (r >> 2);
							row[(4*x)+2] = (g << 3) | (g >> 2);
							row[(4*x)+3] = (b << 3) | (b >> 2);
						} else {
							row[(3*x)+0] = (r << 3) | (r >> 2);
							row[(3*x)+1] = (g << 3) | (g >> 2);
							row[(3*x)+2] = (b << 3) | (b >> 2);
						}
					}
				}
			break;
			
			case 40:
			case 8:
				if (check_ctab_raw(ctab)) break;
				for (y=0; y< height; ++y) {
					for (x=0; x< width; ++x) {
						set_pixel_ctab_raw(output[y], x, *input++, color_model, ctab); 
					}
				}
			break;
			
			case 36:
			case 4:
				if (check_ctab_raw(ctab)) break;
				for (y=0; y< height; ++y) {
					for (x=0; x< width;) {
						oqt_uint8_t val = *input++;
						set_pixel_ctab_raw( output[y], x++, (val&0xF0)>>4, color_model, ctab);
						if (x<width)
						set_pixel_ctab_raw( output[y], x++, (val&0x0F)>>0, color_model, ctab);
					}
				}
			break;
			
			case 34:
			case 2:
				if (check_ctab_raw(ctab)) break;
				for (y=0; y< height; ++y) {
					for (x=0; x< width;) {
						oqt_uint8_t val = *input++;
						set_pixel_ctab_raw( output[y], x++, (val&0xC0)>>6, color_model, ctab);
						if (x<width)
						set_pixel_ctab_raw( output[y], x++, (val&0x30)>>4, color_model, ctab);
						if (x<width)
						set_pixel_ctab_raw( output[y], x++, (val&0x0C)>>2, color_model, ctab);
						if (x<width)
						set_pixel_ctab_raw( output[y], x++, (val&0x03)>>0, color_model, ctab);
					}
				}
			break;

			case 1:
				if (check_ctab_raw(ctab)) break;
				for (y=0; y< height; ++y) {
					for (x=0; x< width;) {
						oqt_uint8_t i,val = *input++;
						for (i=0x80; i && x< width; i/=2)
							set_pixel_ctab_raw( output[y], x++, (val&i)?1:0, color_model, ctab);
					}
				}
			break;

			default:
				fprintf(stderr, "decode_raw: unsupported colour depth (%d)\n", depth);
			break;
		}
		
	}
	
	return 0;
}


static int encode_raw(oqt_t *file,
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	int width = oqt_get_video_width(file, track);
	int height = oqt_get_video_height(file, track);
	int sample_bytes =  oqt_get_video_depth(file, track)/8;
	int frame_size = height * width * sample_bytes;
	int y;

	if(out_size < frame_size) {
		fprintf(stderr, "encode_raw: output buffer is not big enough (%d<%d).\n",
		out_size, frame_size);
		return -1;
	}

	for(y = 0; y < height; ++y) {
		memcpy(output, input[y], sample_bytes * width);
		output += sample_bytes * width;
	}

	*IsAKeyFrame = 1;
	return frame_size;
}

static int reads_colormodel_raw(oqt_t *file,
		int track,
		int colormodel)
{
	return (( colormodel == BC_RGB888 ) || 
			( colormodel == BC_ARGB8888)) ;
}

static int writes_colormodel_raw(oqt_t *file,
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	if(depth == 24)
		return ( colormodel == BC_RGB888 );
	else if(depth == 32)
		return ( colormodel == BC_ARGB8888);
	else
		return 0;
}

static int preferred_colormodel_raw(oqt_t *file, 
		int track)
{
	int depth = oqt_get_video_depth(file, track);
	if(depth == 32)	return BC_ARGB8888;
	else return BC_RGB888;
}

static int init_raw(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	return ++Initialised;
}

static int delete_raw(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	return --Initialised;
}


oqt_codec_info_t*
oqt_simple_video_plugin_register_raw( oqt_video_codec_t *parent )
{
	oqt_video_codec_t *codec;
	
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	codec->info.parent = &(parent->info);
	parent->info.has_children = 1;
	
	codec->info.name = Name;
	memcpy( &codec->info.fourcc, Signature, 4);
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_raw;
	codec->delete_codec = delete_raw;
	codec->encode = encode_raw;
	codec->decode = decode_raw;
	codec->reads_colormodel = reads_colormodel_raw;
	codec->writes_colormodel = writes_colormodel_raw;
	codec->preferred_colormodel = preferred_colormodel_raw;
	
	if (oqt_register_codec(&codec->info)) {
		fprintf(stderr, "Failed to register RAW video codec.\n");
		return NULL;
	}
	
	return &(codec->info);
}
