/* codec_yuv2.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_yuv2.c,v 1.4 2003/03/03 02:54:53 jhatala Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"

static int          Initialised   = 0;
static const char   Signature[5]  = "yuv2";
static const char   *Name = "Component Video (8-bit Y'CbCr 4:2:2)";
static const char   *Version  = "0.1";
static const char   *Copyright  = "2002 OpenQuicktime Team";
static const char   *ModuleAuthor  = "Nicholas Humfrey";
static const char   *CodecAuthor  = "Adam Williams";


typedef struct {
	int coded_w, coded_h;

/* The YUV2 codec requires a bytes per line that is a multiple of 4 */
	int bytes_per_line;

} Param;  


static int decode_yuv2(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t *input,
		       oqt_uint8_t **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);

	if(color_model == BC_NONE) {
		// drop frame
		return 0;
		
	} else if (color_model != BC_YUV422) {
		fprintf(stderr, "decode_yuv2: unsupported colour model: %d\n", color_model);
		return -1;
		
	} else {
		int x,y;

		/* Sanity check */
		if(inputsize != (p->bytes_per_line*p->coded_h)) {
			fprintf(stderr, "decode_yuv2: data size mismatch (%d!=%d).\n",
							inputsize, (p->bytes_per_line*p->coded_h));
			return -1;
		}
		
		for (y=0; y<p->coded_h; y++) {
			oqt_uint8_t *in_row = input+(y*p->bytes_per_line);
			oqt_uint8_t *out_row = output[y];
			for(x = 0; x < p->bytes_per_line; x += 4)
			{
				out_row[x+0] = in_row[x+0];
				out_row[x+1] = in_row[x+1] + 128;
				out_row[x+2] = in_row[x+2];
				out_row[x+3] = in_row[x+3] + 128;
			}
		}
	}
		
	return 0;
}

static int encode_yuv2(oqt_t *file,
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);

	if(color_model == BC_NONE) {
		// drop frame
		return 0;
		
	} else if (color_model != BC_YUV422) {
		fprintf(stderr, "encode_yuv2: unsupported colour model: %d\n", color_model);
		return -1;
		
	} else {
		int x,y;

		/* Sanity check */
		if(out_size < (p->bytes_per_line*p->coded_h)) {
			fprintf(stderr, "encode_yuv2: output buffer is not big enough (%d<%d).\n",
							out_size, (p->bytes_per_line*p->coded_h));
			return -1;
		}
		
		for (y=0; y<p->coded_h; y++) {
			oqt_uint8_t *in_row = input[y];
			oqt_uint8_t *out_row = output+(y*p->bytes_per_line);
			for(x = 0; x < p->bytes_per_line; x += 4)
			{
				out_row[x+0] = in_row[x+0];
				out_row[x+1] = in_row[x+1] - 128;
				out_row[x+2] = in_row[x+2];
				out_row[x+3] = in_row[x+3] - 128;
			}
		}
		
	}
		
	return p->bytes_per_line*p->coded_h;
}

static int reads_colormodel_yuv2(oqt_t *file,
		int track,
		int colormodel)
{
	return ( colormodel == BC_YUV422 );
}


static int writes_colormodel_yuv2(oqt_t *file,
		int track,
		int colormodel)
{
	return ( colormodel == BC_YUV422 );
}

static int preferred_colormodel_yuv2(oqt_t *file, 
		int track)
{
	return BC_YUV422;
}

static int init_yuv2(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_yuv2: failed to allocate memory for codec.\n");
		return 0;
	}
	
	/* setup yuv2 */
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	p->coded_w = (int)((float)oqt_get_video_width(file, track) / 4 + 0.5) * 4;
	p->coded_h = (int)((float)oqt_get_video_height(file, track) / 4 + 0.5) * 4;
	p->bytes_per_line = p->coded_w * 2;

	return ++Initialised;
}

static int delete_yuv2(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}

oqt_codec_info_t*
oqt_simple_video_plugin_register_yuv2( oqt_video_codec_t *parent )
{
	oqt_video_codec_t *codec;
	
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	codec->info.parent = &(parent->info);
	parent->info.has_children = 1;
	
	codec->info.name = Name;
	memcpy( &codec->info.fourcc, Signature, 4);
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_yuv2;
	codec->encode = encode_yuv2;
	codec->decode = decode_yuv2;
	codec->delete_codec = delete_yuv2;
	codec->reads_colormodel = reads_colormodel_yuv2;
	codec->writes_colormodel = writes_colormodel_yuv2;
	codec->preferred_colormodel = preferred_colormodel_yuv2;
	
	if (oqt_register_codec(&codec->info)) {
		fprintf(stderr, "Failed to register yuv2 video codec\n");
		return NULL;
	}
	
	return &(codec->info);
}
