/* codec_yv12.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_yv12.c,v 1.9 2003/04/29 08:53:34 nhumfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/plugin.h"

#include "codec_simple.h"

static int          Initialised   = 0;
static const char   Signature[5]  = "yv12";
static const char   *Name = "Component Video (8-bit Y'CbCr 4:2:0)";
static const char   *Version  = "0.1";
static const char   *Copyright  = "2002 OpenQuicktime Team";
static const char   *ModuleAuthor  = "Jozef Hatala";
static const char   *CodecAuthor  = "Adam Williams";
static const char   *Description = "Not compatible with standard Apple Quicktime.";

static const oqt_codec_param_t* list_params_yv12(int *o_numparams)
{
  *o_numparams = 0;
  return NULL;
}

static int set_param_yv12(oqt_t* file,
			 int          track,
			 const char*  key,
			 const void*  value)
{
  fprintf(stderr,"set_param_yv12: unknown parameter named '%s'\n",key);
  return OQT_PARAMETER_UNKNOWN;
}


static int get_param_yv12(oqt_t* file,
			 int          track,
			 const char*  key,
			 void*        value)
{
  fprintf(stderr,"get_param_yv12: unknown parameter named '%s'\n",key);
  return OQT_PARAMETER_UNKNOWN;
}


static int decode_yv12(oqt_t *file,
				int track,
				oqt_uint32_t inputsize,
				oqt_uint8_t *input,
				oqt_uint8_t **output,
				int color_model)
{
	if(color_model == BC_NONE) {
		// drop frame
		return 0;
	
	} else if (color_model != BC_YUV420P) {
		fprintf(stderr,"decode_yv12: error, unsupported colour model %d.\n", color_model);
		return -1;
	
	} else {
		int wh = oqt_get_video_width(file, track) * oqt_get_video_height(file, track);
		int wh4 = wh/4;
		int framesize = wh + wh4 + wh4;
		
		if(inputsize !=  framesize) {
			fprintf(stderr, "decode_yv12: data size mismatch (%d != %d)\n", inputsize, framesize);
			return 1;
		}
	
		memcpy(output[0], input, wh);
		memcpy(output[1], input + wh, wh4);
		memcpy(output[2], input + wh + wh4, wh4);
		
		return 0;
	}
}

static int encode_yv12(oqt_t *file,
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	int width = oqt_get_video_width(file, track);
	int height = oqt_get_video_height(file, track);
	int wh = width * height;
	int wh4 = wh/4;
	
	if (color_model != BC_YUV420P) {
		fprintf(stderr,"encode_yv12: error, unsupported colour model %d.\n", color_model);
		return 0;
	} else {
		memcpy(output, input[0], wh);
		memcpy(output + wh, input[1], wh4);
		memcpy(output + wh + wh4, input[2], wh4);
		
		*IsAKeyFrame = 1;
		return wh + wh4 + wh4;
	}
}

static int reads_colormodel_yv12(oqt_t *file,
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	return ( colormodel == BC_YUV420P );
}

static int writes_colormodel_yv12(oqt_t *file,
		int track,
		int colormodel)
{
	int depth = oqt_get_video_depth(file, track);
	return ( colormodel == BC_YUV420P );
}

static int preferred_colormodel_yv12(oqt_t *file,
		int track)
{
	return BC_YUV420P;
}


static int init_yv12(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Initialised++;
	return Initialised;
}

static int delete_yv12(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	return --Initialised;
}

oqt_codec_info_t*
oqt_simple_video_plugin_register_yv12( oqt_video_codec_t *parent )
{
	oqt_video_codec_t *codec;
	
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	codec->info.parent = &(parent->info);
	parent->info.has_children = 1;
	
	codec->info.name = Name;
	memcpy( &codec->info.fourcc, Signature, 4);
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.description = Description;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_yv12;
	codec->list_params = list_params_yv12;
	codec->encode = encode_yv12;
	codec->decode = decode_yv12;
	codec->delete_codec = delete_yv12;
	codec->reads_colormodel = reads_colormodel_yv12;
	codec->writes_colormodel = writes_colormodel_yv12;
	codec->preferred_colormodel = preferred_colormodel_yv12;
	codec->set_param = set_param_yv12;
	codec->get_param = get_param_yv12;
	
	if (oqt_register_codec(&codec->info)) {
		fprintf(stderr, "Failed to register YV12 video codec\n");
		return NULL;
	}
	
	return &(codec->info);
}
