/* codec_smc.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_smc.c,v 1.3 2002/12/30 18:32:35 antoinemine Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "smc.h"


#ifdef __cplusplus
extern "C" {
#endif

  

/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "smc "; 
  /* Codec Signature */
  
static const char   *Name = "Apple Graphics (SMC)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *License  = "GNU General Public License"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Mike Melanson and Roberto Togni"; 
static const char   *Description  =
"The Apple Graphics, a.k.a. SMC, codec is sometimes used to compress frames "
"in Apple Quicktime files. It's a computationally simple video codec that "
"operates on 4x4 pixels of 8-bit palettized pixel values. It may as well be "
"called the copy-and-paste codec since it has a number of codes which allow "
"it to copy blocks that have already been rendered in the current frame. "
"For more information visit the codec authors "
"homepage here: http://www.pcisys.net/~melanson/codecs/";


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	smc_context *context;
	oqt_uint8_t *buffer;
} Param;  


static int init_smc(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));

	/* Initalise the Codec */
	if(!(p->context = qt_init_decode_smc())) {
		fprintf(stderr, "init_smc: Error initalising SMC codec\n");
		free(NULL);
		return Initialised;
	}
	
	//p->buffer = NULL;
	
	return ++Initialised;
}


static int delete_smc(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		if (p->context)	{
			if(p->context->color_pairs)		free(p->context->color_pairs);
			if(p->context->color_quads)		free(p->context->color_quads);
			if(p->context->color_octets)	free(p->context->color_octets);
			free(p->context);
		}
		if (p->buffer)	free(p->buffer);
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}



/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_smc(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	oqt_ctab_t *ctab = &file->vtracks[track].track->mdia.minf.stbl.stsd.table[0].ctab;
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int depth   = oqt_get_video_depth(file, track);
	int i, bpp = 3;
      
    // Drop the frame ?
    if(color_model == BC_NONE)
		return 0;
    else if (color_model != BC_RGB888) {
    	fprintf(stderr, "decode_smc: decoding of colourmodel=%d not supported.\n", 
    					color_model);
    	return -1;
    }
    
    // Check colour table exists
    if (!ctab || ctab->size==0) {
    	fprintf(stderr, "decode_smc: can't decode - colour table for track is not available.\n");
    	return -1;
    }
    
	// Allocate buffer if it hasn't already
    if(!p->buffer)
      p->buffer = malloc(width * height * 4);

	qt_decode_smc(p->context, input, inputsize,
				   p->buffer, width, height, ctab, bpp);
	for(i = 0; i < height; i++)
	{
		memcpy(output[i], &(p->buffer[i*width*bpp]), width*bpp);
	}
 	
	return 0;
}


static int reads_colormodel_smc(oqt_t *file, 
		int track,
		int colormodel)
{
	return (colormodel == BC_RGB888);
}

static int preferred_colormodel_smc(oqt_t *file, 
		int track)
{
	return BC_RGB888;
}

/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;
	
	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.description = Description;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.license = License;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_smc;
	//codec->list_params = list_params_smc;
	//codec->encode = encode_smc;
	codec->decode = decode_smc;
	codec->delete_codec = delete_smc;
	codec->reads_colormodel = reads_colormodel_smc;
	codec->preferred_colormodel = preferred_colormodel_smc;
	//codec->writes_colormodel = writes_colormodel_smc;
	//codec->set_param = set_param_smc;
	//codec->get_param = get_param_smc;
	
	return &codec->info;
}

