/* codec_svq1.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_svq1.c,v 1.4 2002/12/30 18:32:36 antoinemine Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "codec_svq1.h"

#ifdef __cplusplus
extern "C" {
#endif


/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "SVQ1"; 
  /* Codec Signature */
  
static const char   *Name = "Sorenson Video"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *License  = "GNU General Public License"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "xine project <http://xine.sourceforge.net/>"; 
//static const char   *Description  ="";


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	svq1_t svq1;
} Param;  


static int init_svq1(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_svq1: failed to allocate memory for codec.\n");
		return 0;
	}
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	
	return ++Initialised;
}



static int delete_svq1(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		if (p->svq1.current)	free (p->svq1.current);
		if (p->svq1.previous)	free (p->svq1.previous);
		if (p->svq1.motion)		free (p->svq1.motion);
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}



/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_svq1(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int pitches[3];
	int result;
	
	// Check the colour model
	if(color_model == BC_NONE) {
		// Drop the frame
		// **** Could do with doing something more inteligent here ****
#ifdef DEBUG
		fprintf(stderr, "decode_svq1: dropping frame.\n");
#endif
		return 0;
	} else if (color_model != BC_YUV420P) {
		fprintf(stderr, "decode_svq1: error, unsupported colour model %d.\n", color_model);
		return -1;
	}
	
	// Decode the frame
	result = svq1_decode_frame(&p->svq1, input);
	if (result) {
#ifdef DEBUG
		fprintf(stderr, "decode_svq1: failed to decode frame (%d).\n", result);
#endif
		return -1;
	}
		
	// Copy it to the output buffer
	pitches[0] = width;
	pitches[1] = width / 2;
	pitches[2] = width / 2;
	svq1_copy_frame(&p->svq1, output, pitches);
	
	return 0;
}


static int reads_colormodel_svq1(oqt_t *file, 
		int track,
		int colormodel)
{
	return (colormodel == BC_YUV420P);
}

static int preferred_colormodel_svq1(oqt_t *file, 
		int track)
{
	return BC_YUV420P;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;
	
	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;
	
	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	//codec->info.description = Description;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.license = License;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_svq1;
	//codec->list_params = list_params_svq1;
	//codec->encode = encode_svq1;
	codec->decode = decode_svq1;
	codec->delete_codec = delete_svq1;
	codec->reads_colormodel = reads_colormodel_svq1;
	//codec->writes_colormodel = writes_colormodel_svq1;
	codec->preferred_colormodel = preferred_colormodel_svq1;
	//codec->set_param = set_param_svq1;
	//codec->get_param = get_param_svq1;
	
	return &codec->info;
}

