/* codec_tga.c
 * Copyright (C) 2002 Nicholas Humfrey
 * Copyright (C) 1989 by Jef Poskanzer
 *
 * Partially based on tgatoppm by Jef Poskanzer
 * Which was based on tga2rast, version 1.0, by Ian MacPhedran.
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_tga.c,v 1.4 2002/12/15 01:52:00 nj_humfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include "tga.h"

#ifdef __cplusplus
extern "C" {
#endif


/*************************************************************/
/* Some Global Variables */
/*************************************************************/
static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "tga "; 
  /* Codec Signature */
  
static const char   *Name = "TrueVision Targa (TGA)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 Nicholas Humfrey and 1989 Jef Poskanzer"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "Jef Poskanzer"; 
  
#define MAXCOLORS 16384


#define PARAM_COUNT 1
 
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "norle", OQT_PARAMETER_BOOLEAN, "Disable RLE encoding, Default False"}
};


/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
	int no_rle;
	
	struct ImageHeader	tga;
	int mapped, rlencoded;
	pixel ColorMap[MAXCOLORS];
	gray AlphaMap[MAXCOLORS];
	int RLE_count, RLE_flag;
	
} Param;  

static const oqt_codec_param_t* list_params_tga(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return ListParams;
}


static int set_param_tga (oqt_t* file,
		    int          track,
		    const char*  param,
		    const void*  data)
{
  Param *p = (Param*)(file->vtracks[track].codec_private);

  if (!strcmp(param,"norle")) {
    p->no_rle = *(( int*)data);

    return 0; 
  }

  fprintf(stderr,"set_param_tga: unknown parameter named '%s'\n",param);
  return OQT_PARAMETER_UNKNOWN;
}

static int get_param_tga (oqt_t* file,
		    int          track,
		    const char*  param,
		    void*        data)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if (!strcmp(param,"norle")) {
		*((int*)data) = p->no_rle;
		return OQT_PARAMETER_BOOLEAN;
	}
	
	fprintf(stderr,"set_param_tga: unknown parameter named '%s'\n",param);
	return OQT_PARAMETER_UNKNOWN;
}




/*************************************************************/
/* Callbacks                                                 */
/*************************************************************/

static oqt_uint8_t*
read_tga( oqt_uint8_t* buf, struct ImageHeader *tgaP )
    {
	unsigned char flags;
	ImageIDField junk;
	
	tgaP->IDLength = *buf++;
	tgaP->CoMapType = *buf++;
	tgaP->ImgType = *buf++;
	tgaP->Index_lo = *buf++;
	tgaP->Index_hi = *buf++;
	tgaP->Length_lo = *buf++;
	tgaP->Length_hi = *buf++;
	tgaP->CoSize = *buf++;
	tgaP->X_org_lo = *buf++;
	tgaP->X_org_hi = *buf++;
	tgaP->Y_org_lo = *buf++;
	tgaP->Y_org_hi = *buf++;
	tgaP->Width_lo = *buf++;
	tgaP->Width_hi = *buf++;
	tgaP->Height_lo = *buf++;
	tgaP->Height_hi = *buf++;
	tgaP->PixelSize = *buf++;
	
	flags = *buf++;
	tgaP->AttBits = flags & 0xf;
	tgaP->Rsrvd = ( flags & 0x10 ) >> 4;
	tgaP->OrgBit = ( flags & 0x20 ) >> 5;
	tgaP->IntrLve = ( flags & 0xc0 ) >> 6;
	
	return buf+tgaP->IDLength;
}

static oqt_uint8_t *
get_map_entry_tga( oqt_uint8_t *buf, pixel *Value, int Size, gray *Alpha )
{
    oqt_uint8_t j=0, k=0, r=0, g=0, b=0, a=0;

    /* Read appropriate number of bytes, break into rgb & put in map. */
    switch ( Size )
	{
		case 8:				/* Grey scale, read and triplicate. */
		r = g = b = *buf++;
		a = 0;
		break;
	
		case 16:			/* 5 bits each of red green and blue. */
		case 15:			/* Watch for byte order. */
		j = *buf++;
		k =*buf++;
		r = ( k & 0x7C ) >> 2;
		g = ( ( k & 0x03 ) << 3 ) + ( ( j & 0xE0 ) >> 5 );
		b = j & 0x1F;
		a = 0;
		break;
	
		case 32:            /* 8 bits each of blue, green, red, and alpha */
		case 24:			/* 8 bits each of blue green and red. */
		b = *buf++;
		g = *buf++;
		r = *buf++;
		if ( Size == 32 )
			a = *buf++;
		else 
			a = 0;
		break;
	
		default:
		fprintf( stderr,  "get_map_entry_tga: unknown colormap pixel size (#2) - %d", Size );
	}
    PPM_ASSIGN( *Value, r, g, b );
    *Alpha = a;
    
    return buf;
}

static oqt_uint8_t *
get_pixel_tga( oqt_uint8_t *buf, Param *p, int maxval, pixel* dest, gray* alpha_p )
{
    static pixval Red, Grn, Blu;
    static pixval Alpha;
    unsigned char j, k;
    static unsigned int l;

    /* Check if run length encoded. */
    if ( p->rlencoded )
	{
		if ( p->RLE_count == 0 )
			{ /* Have to restart run. */
			unsigned char i;
			i = *buf++;
			p->RLE_flag = ( i & 0x80 );
			if ( p->RLE_flag == 0 )
			/* Stream of unencoded pixels. */
			p->RLE_count = i + 1;
			else
			/* Single pixel replicated. */
			p->RLE_count = i - 127;
			/* Decrement count & get pixel. */
			--p->RLE_count;
		}
		else 
		{ /* Have already read count & (at least) first pixel. */
			--p->RLE_count;
			if ( p->RLE_flag != 0 )
			/* Replicated pixels. */
			goto PixEncode;
		}
	}
	
    /* Read appropriate number of bytes, break into RGB. */
    switch ( p->tga.PixelSize )
	{
		case 8:				/* Grey scale, read and triplicate. */
			Red = Grn = Blu = l = *buf++;
			Alpha = 0;
		break;
	
		case 16:			/* 5 bits each of red green and blue. */
		case 15:			/* Watch byte order. */
			j = *buf++;
			k = *buf++;
			l = ( (unsigned int) k << 8 ) + j;
			Red = ( k & 0x7C ) >> 2;
			Grn = ( ( k & 0x03 ) << 3 ) + ( ( j & 0xE0 ) >> 5 );
			Blu = j & 0x1F;
			Alpha = 0;
		break;
	
		case 32:            /* 8 bits each of blue, green, red, and alpha */
		case 24:			/* 8 bits each of blue, green, and red. */
			Blu = *buf++;
			Grn = *buf++;
			Red = *buf++;
			if ( p->tga.PixelSize == 32 )
				Alpha = *buf++;
			else
				Alpha = 0;
			l = 0;
		break;
	
		default:
			fprintf(stderr, "get_pixel_tga: unknown pixel size (#2) - %d\n", p->tga.PixelSize );
	}

PixEncode:
    if ( p->mapped ) {
        *dest = p->ColorMap[l];
        *alpha_p = p->AlphaMap[l];
    } else {
        PPM_ASSIGN( *dest, Red, Grn, Blu );
        *alpha_p = Alpha;
    }
    
    // Scale if samples are not 8bit
    if (maxval != 255) {
		PPM_PUTR( *dest, ((float)PPM_GETR(*dest)/maxval)*255);
		PPM_PUTG( *dest, ((float)PPM_GETG(*dest)/maxval)*255);
		PPM_PUTB( *dest, ((float)PPM_GETB(*dest)/maxval)*255);
		*alpha_p = ((float)*alpha_p/maxval)*0xFF;
    }
    
    return buf;
}

static void reset_param_tga( Param *p ) {
	p->RLE_count = 0;
	p->RLE_flag = 0;
	p->rlencoded = 0;
	p->mapped = 0;
}

static int calc_maxval_tga( Param *p ) {
	int maxval = 0;

 	switch ( p->tga.ImgType )
	{
		case TGA_Map:
		case TGA_RGB:
		case TGA_Mono:
		case TGA_RLEMap:
		case TGA_RLERGB:
		case TGA_RLEMono:
		break;

		default:
		fprintf(stderr, "decode_tga: unknown Targa image type %d.\n", p->tga.ImgType );
		return -1;
	}


   if ( p->tga.ImgType == TGA_Map ||
	 p->tga.ImgType == TGA_RLEMap ||
	 p->tga.ImgType == TGA_CompMap ||
	 p->tga.ImgType == TGA_CompMap4 )
	{ /* Color-mapped image */
		p->tga.CoMapType = 1;	/* Bug fix for Quicktime which does not generate headers correctly */
		if ( p->tga.CoMapType != 1 )
			fprintf(stderr, "calc_maxval_tga: mapped image (type %d) with color map type != 1\n",
			p->tga.ImgType );
		p->mapped = 1;
		/* Figure maxval from CoSize. */
		switch ( p->tga.CoSize )
		{
			case 8:
			case 24:
			case 32:
			maxval = 255;
			break;
	
			case 15:
			case 16:
			maxval = 31;
			break;
	
			default:
			fprintf(stderr, "calc_maxval_tga: unknown colormap pixel size - %d.\n", p->tga.CoSize );
			return -1;
		}
	}
    else
	{ /* Not colormap, so figure maxval from PixelSize. */
		p->mapped = 0;
		switch ( p->tga.PixelSize )
			{
			case 8:
			case 24:
			case 32:
			maxval = 255;
			break;
	
			case 15:
			case 16:
			maxval = 31;
			break;
	
			default:
			fprintf(stderr, "calc_maxval_tga: unknown pixel size - %d.\n", p->tga.PixelSize );
			return -1;
		}
	}

    /* Check run-length encoding. */
    if ( p->tga.ImgType == TGA_RLEMap ||
		p->tga.ImgType == TGA_RLERGB ||
		p->tga.ImgType == TGA_RLEMono )
	p->rlencoded = 1;
    else
	p->rlencoded = 0;


	return maxval;
}

static void dump_tga(struct ImageHeader *tga) {

	fprintf( stderr,  "IDLength = %d\n", (int) tga->IDLength );
	fprintf( stderr,  "CoMapType = %d\n", (int) tga->CoMapType );
	fprintf( stderr,  "ImgType = %d\n", (int) tga->ImgType );
	fprintf( stderr,  "Index_lo = %d\n", (int) tga->Index_lo );
	fprintf( stderr,  "Index_hi = %d\n", (int) tga->Index_hi );
	fprintf( stderr,  "Length_lo = %d\n", (int) tga->Length_lo );
	fprintf( stderr,  "Length_hi = %d\n", (int) tga->Length_hi );
	fprintf( stderr,  "CoSize = %d\n", (int) tga->CoSize );
	fprintf( stderr,  "X_org_lo = %d\n", (int) tga->X_org_lo );
	fprintf( stderr,  "X_org_hi = %d\n", (int) tga->X_org_hi );
	fprintf( stderr,  "Y_org_lo = %d\n", (int) tga->Y_org_lo );
	fprintf( stderr,  "Y_org_hi = %d\n", (int) tga->Y_org_hi );
	fprintf( stderr,  "Width_lo = %d\n", (int) tga->Width_lo );
	fprintf( stderr,  "Width_hi = %d\n", (int) tga->Width_hi );
	fprintf( stderr,  "Height_lo = %d\n", (int) tga->Height_lo );
	fprintf( stderr,  "Height_hi = %d\n", (int) tga->Height_hi );
	fprintf( stderr,  "PixelSize = %d\n", (int) tga->PixelSize );
	fprintf( stderr,  "AttBits = %d\n", (int) tga->AttBits );
	fprintf( stderr,  "Rsrvd = %d\n", (int) tga->Rsrvd );
	fprintf( stderr,  "OrgBit = %d\n", (int) tga->OrgBit );
	fprintf( stderr,  "IntrLve = %d\n", (int) tga->IntrLve );
}


/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_tga(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
   	unsigned int i, temp1, temp2;
	int realrow=0, truerow=0, baserow=0, maxval;
    int rows, cols, row, col;
	oqt_uint8_t* buf;
	
	reset_param_tga( p );

	// Drop frame
	if(color_model == BC_NONE) {
		return 0;
	}
	
	buf = read_tga( input, &p->tga );
	//dump_tga( &p->tga );
	

	/* Make sure it is valid */
	if (p->tga.Width_lo != oqt_get_video_width(file, track)) {
		fprintf(stderr, "decode_tga: width of tga does not match that of movie.\n");
		return -1;
	} else if (p->tga.Height_lo != oqt_get_video_height(file, track)) {
		fprintf(stderr, "decode_tga: height of tga does not match that of movie.\n");
		return -1;
	}
	
	/* Calculate rows, cols and more image validation */
	rows = ( (int) p->tga.Height_lo ) + ( (int) p->tga.Height_hi ) * 256;
    cols = ( (int) p->tga.Width_lo ) + ( (int) p->tga.Width_hi ) * 256;
	maxval = calc_maxval_tga( p );
	if (maxval<0) return maxval;

    /* If required, read the color map information. */
    if ( p->tga.CoMapType != 0 )
	{
		unsigned int i, temp1, temp2;
		temp1 = p->tga.Index_lo + p->tga.Index_hi * 256;
		temp2 = p->tga.Length_lo + p->tga.Length_hi * 256;
		if ( ( temp1 + temp2 + 1 ) >= MAXCOLORS )
			fprintf(stderr, "decode_tga: too many colors - %d.\n", ( temp1 + temp2 + 1 ) );
		for ( i = temp1; i < ( temp1 + temp2 ); ++i )
			buf = get_map_entry_tga( buf, &p->ColorMap[i], (int) p->tga.CoSize,
						   &p->AlphaMap[i]);
	}
	
	/* Decode into the output buffer */
    for ( row = 0; row < rows; ++row )
	{
		oqt_uint8_t *out_buf = output[row];
		realrow = truerow;
		if ( p->tga.OrgBit == 0 )
			realrow = rows - realrow - 1;
	
		for ( col = 0; col < cols; ++col ) {
			pixel rgb;
			gray alpha;
			buf = get_pixel_tga( buf, p, maxval, &rgb, &alpha );
			if (color_model == BC_ARGB8888)
				*out_buf++ = alpha;
			*out_buf++ = PPM_GETR(rgb);
			*out_buf++ = PPM_GETG(rgb);
			*out_buf++ = PPM_GETB(rgb);
			if (color_model == BC_RGBA8888)
				*out_buf++ = alpha;
		}
		
		if ( p->tga.IntrLve == TGA_IL_Four )
			truerow += 4;
		else if ( p->tga.IntrLve == TGA_IL_Two )
			truerow += 2;
		else
			++truerow;
		if ( truerow >= rows )
			truerow = ++baserow;
	}

	
	return 0;
}


/*************************************************************/
/* Encoding function                                         */
/*************************************************************/
static int encode_tga(oqt_t *file, 
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
	Param *p = (Param*)(vtrack->codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int depth 	= oqt_get_video_depth(file, track);
	int frame_size = 0;

	
	return frame_size;
}

static int writes_colormodel_tga(oqt_t *file, 
		int track,
		int colormodel)
{ 
	return ( colormodel == BC_RGB888 ||
			 colormodel == BC_RGBA8888 );
}

static int reads_colormodel_tga(oqt_t *file, 
		int track,
		int colormodel)
{
	return ( colormodel == BC_RGB888 ||
			 colormodel == BC_RGBA8888 ||
			 colormodel == BC_ARGB8888 );
}

static int preferred_colormodel_tga(oqt_t *file, 
		int track)
{
	int depth = oqt_get_video_depth(file, track);
	if (depth == 32)	return BC_RGBA8888;
	else 				return BC_RGB888;
}



static int init_tga(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_tga: failed to allocate memory for codec.\n");
		return 0;
	}
	
	/* setup tga */
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	p->no_rle = 0;
	
	return ++Initialised;
}


static int delete_tga(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;

	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;

	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_tga;
	codec->list_params = list_params_tga;
	//codec->encode = encode_tga;
	codec->decode = decode_tga;
	codec->delete_codec = delete_tga;
	codec->reads_colormodel = reads_colormodel_tga;
	//codec->writes_colormodel = writes_colormodel_tga;
	codec->preferred_colormodel = preferred_colormodel_tga;
	codec->set_param = set_param_tga;
	codec->get_param = get_param_tga;
	
	return &codec->info;
}


