/* codec_tiff.c
 * Copyright (C) 2002 OpenQuicktime Team and University of Southampton
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_tiff.c,v 1.4 2002/12/15 01:52:00 nj_humfrey Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#include <time.h>
#include <sys/utsname.h>


#include <tiffio.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#ifdef __cplusplus
extern "C" {
#endif


/*************************************************************/
/* Some Global Variables */
/*************************************************************/
//#define	DEBUG_TIFF		0

static int              Initialised   = 0;      
  /* Initialising counter */

static const char   Signature[5]  = "tiff"; 
  /* Codec Signature */
  
static const char   *Name = "Tag Image File Format (TIFF)"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Nicholas Humfrey"; 
static const char   *CodecAuthor  = "libtiff"; 
  
#define PARAM_COUNT 2
 
static const oqt_codec_param_t ListParams[PARAM_COUNT] =
{
	{ "packbits", OQT_PARAMETER_BOOLEAN, "Use Packbits Compression [Default True]"},
	{ "lendian", OQT_PARAMETER_BOOLEAN, "Use little endian [Default False]"}
};


// Each instance of the codec should have its own Param structure allocated
typedef struct {
	int packbits;
	int lendian;
	
	int reading;
	
	oqt_uint32_t buffer_size;
	oqt_uint32_t buffer_max;
	oqt_uint32_t buffer_position;
	oqt_uint32_t buffer_allocated;
	oqt_uint8_t  *buffer;
	
} Param;  


static const oqt_codec_param_t* list_params_tiff(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	return ListParams;
}


static int set_param_tiff (oqt_t* file,
		    int          track,
		    const char*  param,
		    const void*  data)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if (!strcmp(param,"packbits")) {
		p->packbits = *(( int*)data);
	} else if (!strcmp(param,"lendian")) {
		p->lendian = *(( int*)data);
	} else {
		fprintf(stderr,"set_param_tiff: unknown parameter named '%s'\n",param);
		return OQT_PARAMETER_UNKNOWN;
	}
	
	return 0;
}

static int get_param_tiff (oqt_t* file,
		    int          track,
		    const char*  param,
		    void*        data)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	
	if (!strcmp(param,"packbits")) {
		*((int*)data) = p->packbits;
		return OQT_PARAMETER_BOOLEAN;
	}
	if (!strcmp(param,"lendian")) {
		*((int*)data) = p->lendian;
		return OQT_PARAMETER_BOOLEAN;
	}
	
	fprintf(stderr,"get_param_tiff: unknown parameter named '%s'\n",param);
	return OQT_PARAMETER_UNKNOWN;
}



/*************************************************************/
/* libtiff interfacing                                       */
/*************************************************************/

static tsize_t
read_function(thandle_t fd, tdata_t data, tsize_t length)
{
	Param *p = (Param*)(fd);
#ifdef DEBUG_TIFF
	fprintf(stderr, "read_function: read %d bytes.\n", (int)length);
#endif	
	if (!p->reading) return -1;

	if(length + p->buffer_position <= p->buffer_size)
	{
		memcpy(data, p->buffer + p->buffer_position, length);
		p->buffer_position += length;
	} else {
		fprintf(stderr, "codec_tiff: ran out of buffer space in read_function.\n");
		return -1;
	}
	
	if (p->buffer_position > p->buffer_max)
		p->buffer_max = p->buffer_position;
	
	return length;
}

static tsize_t
write_function(thandle_t fd, tdata_t data, tsize_t length)
{
	Param *p = (Param*)(fd);
#ifdef DEBUG_TIFF
	fprintf(stderr, "write_function: wrote %d bytes.\n", (int)length);
#endif	
	if (p->reading) return -1;

	if(length + p->buffer_position <= p->buffer_size)
	{
		memcpy(p->buffer + p->buffer_position, data, length);
		p->buffer_position += length;
	} else {
		fprintf(stderr, "codec_tiff: ran out of buffer space in write_function.\n");
		return -1;
	}
	
	if (p->buffer_position > p->buffer_max)
		p->buffer_max = p->buffer_position;
	
	return (length);
}

static toff_t
seek_function(thandle_t fd, toff_t off, int whence)
{
	Param *p = (Param*)(fd);
	
	if (whence == SEEK_SET) p->buffer_position = off;
	else if (whence == SEEK_CUR) p->buffer_position += off;
	else if (whence == SEEK_END) p->buffer_max += off;
	else {
		fprintf(stderr, "codec_tiff: unknown whence: %d\n", (int)whence);
	}
	
#ifdef DEBUG_TIFF
	fprintf(stderr, "seek_function: seeking to %d (%d).\n",p->buffer_position, whence);
#endif

	if (p->buffer_position > p->buffer_size) {
		fprintf(stderr, "codec_tiff: warning, seek overshot buffer.\n");
		return -1;
	}

	if (p->buffer_position > p->buffer_max)
		p->buffer_max = p->buffer_position;
	
	return (p->buffer_position);
}

static toff_t
size_function(thandle_t fd)
{
	Param *p = (Param*)(fd);
	
#ifdef DEBUG_TIFF
	fprintf(stderr, "size_function: size is %d.\n",p->buffer_max);
#endif

	return (p->buffer_max);
}

static void reset_buffer_tiff(Param *p) {
	p->buffer_size = 0;
	p->buffer_max = 0;
	p->buffer_position = 0;
	p->buffer_allocated = 0;
	p->buffer = NULL;
}

static int
close_function(thandle_t fd)
{
	Param *p = (Param*)(fd);
	reset_buffer_tiff(p);
	return (0);
}

static TIFF* open_tiff_mem(char *mode, Param* p) {
	return TIFFClientOpen("(oqt_codec_tiff)", mode, (thandle_t)p,
		 read_function, write_function,
		 seek_function, close_function, size_function, NULL, NULL);
}


/*************************************************************/
/* Decoding function                                         */
/*************************************************************/
static int decode_tiff(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
	Param *p = (Param*)(file->vtracks[track].codec_private);
	oqt_uint32_t t_width, t_height;
	oqt_uint32_t *raster = NULL;
	int samples_per_pixel=0, result = 0;
	TIFF *tiff;
	
	/* Check the colour model */
	if (color_model == BC_RGB888) samples_per_pixel = 3;
	//else if (file->color_model == BC_ARGB8888) samples_per_pixel = 4;
	else {
		fprintf(stderr, "decode_tiff: unsupported colour model %d.\n", color_model);
		return -1;
	}
	fprintf(stderr, "decode_tiff: samples_per_pixel=%d\n", samples_per_pixel);
	
	/* Setup buffers */
	reset_buffer_tiff( p );
	p->reading = 1;
	p->buffer = input;
	p->buffer_size = inputsize;
	p->buffer_max = inputsize;

	
	/* 'Open' the pseudo tiff file */
	tiff = open_tiff_mem("rm", p);
	if (tiff == NULL) {
		fprintf(stderr, "decode_tiff: failed to open tiff for reading.\n");
		return -1;
	}

	/* Check that the format is right */
	TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &t_width);
	TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &t_height);
	
	if (t_width != oqt_get_video_width(file, track)) {
		fprintf(stderr, "decode_tiff: width of tiff does not match that of movie.\n");
	} else if (t_height != oqt_get_video_height(file, track)) {
		fprintf(stderr, "decode_tiff: height of tiff does not match that of movie.\n");
	} else {
		raster = (oqt_uint32_t*)_TIFFmalloc(t_height * t_width * sizeof(oqt_uint32_t));
	
		/* go ahead and read the image */
		if (TIFFReadRGBAImage(tiff, t_width, t_height, (uint32*)raster, 0)) {
			int i,j;

			/* *** Could/should be optimised using TIFFRGBAImageBegin/TIFFRGBAImageGet
		 	       and put callbacks.... ***
			*/
			for (i=0;i<t_height; i++) {
				oqt_uint8_t *row = output[t_height-(i+1)];
				//if (samples_per_pixel==4) {
				//	memcpy(row, &raster[i*t_width], t_width*4);
				//} else {
					for(j=0;j<t_width; j++) {
						oqt_uint32_t colour = raster[(i*t_width)+j];
						row[(j*3)+0] = TIFFGetR(colour);
						row[(j*3)+1] = TIFFGetG(colour);
						row[(j*3)+2] = TIFFGetB(colour);
					}
				//} 
			}
		} else {
			fprintf(stderr, "decode_tiff: Failed to decode frame of tiff.\n");
			result = -1;
		}
		
		if (raster) _TIFFfree(raster);
	}
	
	/* Close the 'file' */
	TIFFClose( tiff );
	
	return result;
}

#define STR_BUF_SIZE	255

static void set_info_strings(TIFF* tiff) {
	char *c, *ver_str, buf[STR_BUF_SIZE];
	struct utsname machine;
	time_t curtime;	

	// Create  a host string
	if (uname(&machine)==0) {
		sprintf(buf, "%s %s", machine.sysname, machine.release);
		TIFFSetField(tiff, TIFFTAG_HOSTCOMPUTER, buf);
	} else {
		perror("codec_tiff: uname in set_info_strings failed");
	}

	// Current local time
	curtime = time(NULL);
	strftime(buf, STR_BUF_SIZE, "%Y:%m:%d %H:%M:%S", localtime(&curtime));
  	TIFFSetField(tiff, TIFFTAG_DATETIME, buf);
	
	// Terminate the version string at the end of the first line
	ver_str = strdup(TIFFGetVersion());
	for(c=ver_str;c[0];c++) {
		if (c[0] == '\n' || c[0] == '\r') {
			c[0] = 0;
			break;
		}
	}
	TIFFSetField(tiff, TIFFTAG_SOFTWARE, ver_str);
	free(ver_str);
}

/*************************************************************/
/* Encoding function                                         */
/*************************************************************/
static int encode_tiff(oqt_t *file, 
		       int track,
		       oqt_uint8_t **input,
		       oqt_uint32_t out_size,
		       oqt_uint8_t *output,
		       int *IsAKeyFrame,
		       int color_model)
{
	oqt_video_map_t *vtrack = &(file->vtracks[track]);
	Param *p = (Param*)(vtrack->codec_private);
	int width 	= oqt_get_video_width(file, track);
	int height 	= oqt_get_video_height(file, track);
	int samples_per_pixel = 0, row, frame_size = 0;
	oqt_uint8_t *raster=NULL;
	
	TIFF *tiff;
	
	/* Setup buffers */
	reset_buffer_tiff( p );
	p->reading = 0;
	p->buffer = output;
	p->buffer_size = out_size;
	

	/* Check the colour model */
	if (color_model == BC_RGB888) samples_per_pixel = 3;
	//else if (color_model == BC_ARGB8888) samples_per_pixel = 4;
	else {
		fprintf(stderr, "decode_tiff: unsupported colour model %d.\n", color_model);
		return -1;
	}

	/* Copy row buffer into framebuffer */
	raster = _TIFFmalloc(height * width * samples_per_pixel);
	for (row=0;row<height; row++) {
		memcpy(raster+(row*width*samples_per_pixel), input[row], width*samples_per_pixel);
	}
	

	/* 'Open' the pseudo tiff file */
	if (p->lendian)
			tiff = open_tiff_mem("wml", p);
	else	tiff = open_tiff_mem("wmb", p);
	if (tiff == NULL) {
		fprintf(stderr, "encode_tiff: failed to open tiff for writing.\n");
		return -1;
	}
	

	/* Setup the output format */
	TIFFSetField(tiff, TIFFTAG_SUBFILETYPE, 0x0);
	TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, width);
	TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, height);
	TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, samples_per_pixel);
	TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, height);
	//rowsperstrip = TIFFDefaultStripSize(tiff, 15));
	
	if (p->packbits)
			TIFFSetField(tiff, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
	else	TIFFSetField(tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE);

	TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
	TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	
	set_info_strings( tiff );
	
	TIFFSetField(tiff, TIFFTAG_XRESOLUTION, 72.0);
	TIFFSetField(tiff, TIFFTAG_YRESOLUTION, 72.0);
	TIFFSetField(tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
	//TIFFSetField(tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  
  	
    /* Encode the frame as a single strip */
    frame_size = TIFFWriteEncodedStrip(tiff, 0, raster, height*width*samples_per_pixel);
	
	/* Close the 'file' */
	TIFFClose( tiff );
	_TIFFfree(raster);
	
	return frame_size;
}

static int writes_colormodel_tiff(oqt_t *file, 
		int track,
		int colormodel)
{ 
//	return ( colormodel == BC_RGB888 ||
//	         colormodel == BC_ARGB8888);
	return ( colormodel == BC_RGB888);
}

static int reads_colormodel_tiff(oqt_t *file, 
		int track,
		int colormodel)
{
//	return ( colormodel == BC_RGB888 ||
//	         colormodel == BC_ARGB8888);
	return ( colormodel == BC_RGB888);
}

static int preferred_colormodel_tiff(oqt_t *file, 
		int track)
{
	return BC_RGB888;
}



static int init_tiff(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p;
		
	/* Init params, should be called for every instance */
	vtrack->codec_private = malloc(sizeof(Param));
	if (!vtrack->codec_private) {
		fprintf(stderr, "init_tiff: failed to allocate memory for codec.\n");
		return 0;
	}
	
	/* setup libtiff */
	p = (Param*)vtrack->codec_private;
	memset(p, 0, sizeof(Param));
	reset_buffer_tiff( p );
	p->packbits = 1;
	p->lendian = 0;
	
	return ++Initialised;
}


static int delete_tiff(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Param *p = (Param*)(vtrack->codec_private);

	if(p) {
		free(p);
		vtrack->codec_private=NULL;
	}

	return --Initialised;
}


/*************************************************************/
/* Registering function                                      */
/*************************************************************/
oqt_codec_info_t*
oqt_plugin_register( void )
{
	oqt_video_codec_t *codec;

	// Allocate the memory for the codec record
	codec = oqt_allocate_video_codec();
	if (!codec) return NULL;

	// Information about the codec
	memcpy( &codec->info.fourcc, Signature, 4); 
	codec->info.name = Name;
	codec->info.version = Version;
	codec->info.copyright = Copyright;
	codec->info.module_author = ModuleAuthor;
	codec->info.codec_author = CodecAuthor;
	codec->info.interface_age = OQT_INTERFACE_AGE;
	
	codec->init_codec = init_tiff;
	codec->encode = encode_tiff;
	codec->decode = decode_tiff;
	codec->delete_codec = delete_tiff;
	codec->reads_colormodel = reads_colormodel_tiff;
	codec->writes_colormodel = writes_colormodel_tiff;
	codec->preferred_colormodel = preferred_colormodel_tiff;
	codec->list_params = list_params_tiff;
	codec->set_param = set_param_tiff;
	codec->get_param = get_param_tiff;
	
	return &codec->info;
}


